(** An implementation of arrays in Coq.
 * @author Patryk Czarnik
 *)

Require Import List.

(** Arrays of elements of a fixed type. *)
Module Type ARRAY_TYPE.

  (** The type of array elements. *)
  Parameter carrier_t : Set.

  (** The type of array itself. *)
  Parameter array_t : Set.

  (** Create an empty array of the given length. The value of  every of its element is [None]. *)
  Parameter arr_create: nat -> array_t.

  (** Returns the length of the array. *)
  Parameter arr_length: array_t -> nat.

  (** If [val] is the [i]th element of the array [arr], then [arr_get arr i] is equal to [Some val],
       if the [i]th cell is uninitialized, then [arr_get arr i] is equal to [None]. *)
  Parameter arr_get : array_t-> nat -> option carrier_t.

  (** [arr_set arr i val] returns [arr] with the [i]th element set to [val]. *)
  Parameter arr_set : array_t -> nat -> carrier_t -> array_t.

  (** [arr_from_list l] returns an array containing exactly the elements,
     * that [l] has contained *)
  Parameter arr_from_list : list carrier_t -> array_t.

  (** [arr_from_list_ext l n] returns an array [A] of size [n] form list [l].
     * For a given [i]: if [i < length l] then [arr_get A i = nth l i], else [arr_get A i = None].*)
  Parameter arr_from_list_ext : list carrier_t -> nat -> array_t.

  (* properties *)
  (** [get arr i] returns [None] when [i >= arr_length arr].*)
  Axiom arrprop_get1 : forall arr i,
    i >= arr_length arr ->
    arr_get arr i = None.

  (** [arr_create n] is of length [n] *)
  Axiom arrprop_create_length : forall n, arr_length (arr_create n) = n.

  (** Newly created array has uninitialised cells. *)
  Axiom arrprop_create_get : forall n i, arr_get (arr_create n) i = None.

  (** [arr_set] does not change the length of the array. *)
  Axiom arrprop_set_length : forall arr i el,
    arr_length (arr_set arr i el) = arr_length arr.

  (** [arr_set arr i el] sets the [i]th element to [el] . *)
  Axiom arrprop_set_get1 : forall arr i el,
    i < arr_length arr ->
    arr_get (arr_set arr i el) i = Some el.

  (** [arr_set arr i el] does not change elements other than the [i]th. *)
  Axiom arrprop_set_get2 : forall arr i j el,
    i <> j ->
    arr_get (arr_set arr i el) j = arr_get arr j.

  Axiom arrprop_arr_from_list_length : forall l,
    arr_length (arr_from_list l) = length l.

  Axiom arrprop_arr_from_list_get :
    let f := fun x => Some x in
    forall l i,
    arr_get (arr_from_list l) i = nth i (map f l) None.

  Axiom arrprop_arr_from_list_ext_length : forall n l,
    arr_length (arr_from_list_ext l n) = n.

  Axiom arrprop_arr_from_list_ext_get1 :
    let f := fun x => Some x in
    forall n l i y,
    (i < n /\ i < length l) ->
    arr_get (arr_from_list_ext l n) i = nth i (map f l) y.

  Axiom arrprop_arr_from_list_ext_get2 :
    forall n l i,
    i >= n \/ i >= length l->
    arr_get (arr_from_list_ext l n) i = None.
End ARRAY_TYPE.

Module Type ELEMENT_TYPE.
  (** The type of array elements. *)
  Parameter carrier_t : Set.
End ELEMENT_TYPE.

Module ArrayAsList (Element: ELEMENT_TYPE) : ARRAY_TYPE with Definition carrier_t := Element.carrier_t.
Require Import Omega.

  Definition carrier_t := Element.carrier_t.

  Definition array_t := list (option carrier_t).

  Fixpoint arr_create (n:nat) {struct n} : array_t :=
  match n with
  | O => nil
  | S n' => None::(arr_create n')
  end.

  Definition arr_length arr := @length (option carrier_t) arr.

  Fixpoint arr_get (arr: array_t) (i:nat) {struct arr} : option carrier_t :=
    match i, arr with
    | O, nil => None
    | O, (hd::tl) => hd
    | S i', nil => None
    | S i', (hd::tl) => arr_get tl i'
    end.

  Fixpoint arr_set (arr: array_t) (i: nat) (el: carrier_t) {struct arr}: array_t :=
    match arr, i with
    | nil, _ => nil (* index out of bounds *)
    | (hd::tl), O => ((Some el)::tl) (* this is the cell to replace *)
    | (hd::tl), (S i') => hd::(arr_set tl i' el)
    end.

  Definition arr_from_list (l: list carrier_t) : array_t :=
    map (fun el => Some el) l.

  Fixpoint arr_from_list_ext (l: list carrier_t) (n:nat) {struct n} : array_t :=
  match n with
  | O => nil
  | S n' =>
    match l with
    | nil =>  None::(arr_from_list_ext nil n')
    | (hd::tl) => (Some hd)::(arr_from_list_ext tl n')
    end
  end.
  

  Theorem arrprop_create_length : forall n, arr_length (arr_create n) = n.
  Proof.
    intro n.
    induction n.
    reflexivity.
    simpl.
    auto.
  Qed.

  Lemma length_le_0_nil : forall A l,
    @length A l <= O -> l = nil.
  Proof.
    intros A l H.
    destruct l.
    reflexivity.
    simpl in H.
    assert False.
    omega.
    contradiction.
  Qed.

  Theorem arrprop_get1 : forall arr i,
    i >= arr_length arr -> arr_get arr i = None.
  Proof.
    unfold arr_length.
    induction arr.

      intros i H.
      destruct i; simpl;reflexivity.

      intros i H.
      induction i.
        rewrite ((length_le_0_nil (option carrier_t) (a::arr)) H).
        simpl; reflexivity.

        simpl.
        apply IHarr.
        simpl in H.
        omega.
  Qed.

  (** Newly created array has uninitialised cells. *)
  Theorem arrprop_create_get : forall n i, arr_get (arr_create n) i = None.
  Proof.
    induction n; induction i; simpl; auto.
  Qed.

  Theorem arrprop_set_length : forall arr i el,
    arr_length (arr_set arr i el) = arr_length arr.
  Proof.
    induction arr.
      destruct i; simpl; tauto.

      destruct i.
        simpl; tauto.

        simpl;auto.
  Qed.

  (** [arr_set arr i el] sets the [i]th element to [el] . *)
  Theorem arrprop_set_get1 : forall arr i el,
    i < arr_length arr ->
    arr_get (arr_set arr i el) i = Some el.
  Proof.
    induction arr; intros i el H.
     destruct i; simpl in H.
       assert False; [omega | contradiction].
       assert False; [omega | contradiction].

    destruct i.
      simpl; tauto.

      simpl.
      simpl in H.
      apply IHarr.
      omega.
  Qed.

  (** [arr_set arr i el] does not change elements other than the [i]th. *)
  Theorem arrprop_set_get2 : forall arr i j el,
    i <> j ->
    arr_get (arr_set arr i el) j = arr_get arr j.
  Proof.
    induction arr; intros i j el H.
     destruct i; destruct j; simpl; tauto.

    destruct i; destruct j; simpl.

      assert False.
      omega.
      contradiction.

      reflexivity.

      reflexivity.

      apply IHarr.
      omega.
  Qed.

  Theorem arrprop_arr_from_list_length : forall l,
    arr_length (arr_from_list l) = length l.
  Proof.
    induction l.
      simpl; reflexivity.

      simpl.
      rewrite IHl.
      reflexivity.
  Qed.

  Theorem arrprop_arr_from_list_get :
    let f := fun x => Some x in
    forall l i,
    arr_get (arr_from_list l) i = nth i (map f l) None.
  Proof.
    induction l; intros i.
      destruct i; simpl; reflexivity.

      destruct i; simpl.
        unfold f; reflexivity.

        apply IHl.
  Qed.

  Theorem arrprop_arr_from_list_ext_length : forall n l,
    arr_length (arr_from_list_ext l n) = n.
  Proof.
    induction n.
      destruct l; simpl; reflexivity.
      destruct l; simpl; rewrite IHn; reflexivity.
  Qed.

  Theorem arrprop_arr_from_list_ext_get1 :
    let f := fun x => Some x in
    forall n l i y,
    (i < n /\ i < length l) ->
    arr_get (arr_from_list_ext l n) i = nth i (map f l) y.
  Proof.
    induction n.
      intros l i el H.
      assert False.
        omega.
        tauto.

      destruct l.
        intros i y H.
        simpl in H.
          assert False.
            omega.
            tauto.

        destruct i.
          intros y H.
          simpl.
          trivial.

          intros y H.
          simpl.
          apply IHn.
          split.
            omega.

            simpl in H.
            omega.
  Qed.


  Theorem arrprop_arr_from_list_ext_get2 :
    forall n l i,
    i >= n \/ i >= length l->
    arr_get (arr_from_list_ext l n) i = None.
  Proof.
    intuition.
      apply arrprop_get1.
      rewrite arrprop_arr_from_list_ext_length.
      assumption.

    generalize n l i H0.
    induction n0.
      destruct i0; simpl; reflexivity.

      induction l0.
      destruct i0.
        simpl; reflexivity.

        intro H1.
        simpl.
        apply IHn0.
        simpl.
        omega.

      destruct i0.
        intros H1.
        simpl in H1.
        assert False.
          omega.
          tauto.

        intros H1.
        simpl.
        apply IHn0.

        simpl in H1.
        omega.
  Qed.

End ArrayAsList.
