(** Formalization of the bytecode verification.
 Based on The "Java (TM) Virtual Machine Specification, Second Edition, 
  Tim Lindholm, Frank Yellin",  sections 4.8 and 4.9.

 @author Patryk Czarnik, ...  *)

Require Import Program.
Require Import Domain.
Require Import SmallStep.
Require Import ZArith.
Require Import List.
Open Scope Z_scope.

Module BtcVerif (Prog': PROGRAM).
Import Prog'.

(** Types of constraints. *)
Module ConstraintTypes.
  (** Properties of whole programs have the following type. *)
  Definition program_constraint_type := Program -> Prop.

  (** Properties of single classes have the following type. *)
  Definition class_constraint_type := Class -> Prop.

  (** Properties of single methods have the following type. *)
  Definition method_constraint_type := Method -> Prop.

  (** Properties of single methods code have the following type. *)
  Definition code_constraint_type := BytecodeMethod -> Prop.

  (** The type of a property on a single instruction within a method. *)
  Definition instruction_constraint_type := PC -> Instruction -> Prop.
End ConstraintTypes.

Import ConstraintTypes.

(** Auxiliary operations. *)
Module Auxiliary.

(** [sorted_list E rel l] holds iff [l] is sorted with respect to the [rel].
   [rel] should be a linear order. *)
Inductive sorted_list (E: Set) (rel: (E -> E -> Prop)): list E -> Prop :=
| SortedList_Nil: sorted_list E rel nil
| SortedList_Singleton: forall x, sorted_list E rel (x::nil)
| SortedList_Cons: forall x y tail, sorted_list E rel (y::tail) -> rel x y -> sorted_list E rel (x::y::tail).

(** [for_all_in_list E pred l] holds iff [pred] holds for every member of [l]. *)
Inductive for_all_in_list (E: Set) (pred: E -> Prop): list E -> Prop :=
| ForAllInList_Nil: for_all_in_list E pred nil
| ForAllInList_Cons: forall el tail,
    for_all_in_list E pred tail ->
    pred el ->
    for_all_in_list E pred (el::tail).

(** [for_all_classes prog pred] holds iff [pred] holds for every class in program [prog]. *)
Definition for_all_classes: class_constraint_type -> program_constraint_type :=
  fun pred prog =>
  forall c, PROG.defined_Class prog c -> pred c.

(** [for_all_methods cl pred] holds iff [pred] holds for every method in [cl]. *)
Definition for_all_methods: method_constraint_type -> class_constraint_type :=
  fun pred cl =>
  forall m, CLASS.defined_Method cl m -> pred m.

(** [for_code m pred] holds iff m is not abstract and [pred] holds for the body of [m]. *)
Definition for_body: code_constraint_type -> method_constraint_type :=
  fun pred m =>
  match METHOD.body m with
  | None => False
  | Some bm => pred bm
  end.

(** [for_all_instructions m pred] holds iff [pred] holds for every instruction in [bm]. *)
(*? Shouldn't it handle only instructions reachable wrt [firstAddress] and [nextAddress]? *)
Definition for_all_instructions: instruction_constraint_type -> code_constraint_type :=
  fun pred bm =>
  forall pc ins,
  BYTECODEMETHOD.instructionAt bm pc = Some ins ->
  pred pc ins.

Inductive reachable_pc (bm: BytecodeMethod) : PC -> Prop :=
| ReachablePC_First: reachable_pc bm (BYTECODEMETHOD.firstAddress bm)
| ReachablePC_Handler: forall exn, In exn (BYTECODEMETHOD.exceptionHandlers bm) ->
  reachable_pc bm (EXCEPTIONHANDLER.handler exn)
| ReachablePC_Next: forall pc pc', reachable_pc bm pc ->
  BYTECODEMETHOD.nextAddress bm pc = Some pc' ->
  reachable_pc bm pc'.

Definition for_all_instructions1: instruction_constraint_type -> code_constraint_type :=
  fun pred bm =>
  forall pc ins, reachable_pc bm pc ->
  BYTECODEMETHOD.instructionAt bm pc = Some ins ->
  pred pc ins.


(** [target_ok bm pc off] holds iff [off] is a proper offset from [pc] in the method [bm]. *)
(*? Does it really mean what it ought to mean? ?*)
Definition target_ok: BytecodeMethod -> PC -> OFFSET.t -> Prop :=
fun bm pc offset =>
  let pc' := OFFSET.jump pc offset in
  exists ins',
  BYTECODEMETHOD.instructionAt bm pc' = Some ins'.

(** [array_dimensions_eq t d] holds iff
  [t] is an array type and the array has exactly [d] dimensions
  or [d]=0 and [t] is not an array type (it is a class reference type, an interface reference type or a primitive type)
*) 
Inductive array_dimensions_eq: type -> Z -> Prop :=
| ArrayDimensionEq_Primitive:
  forall pt, array_dimensions_eq (PrimitiveType pt) 0
| ArrayDimensionEq_Class:
  forall ct, array_dimensions_eq (ReferenceType (ClassType ct)) 0
| ArrayDimensionEq_Interface:
  forall it, array_dimensions_eq (ReferenceType (InterfaceType it)) 0
| ArrayDimensionEq_Array:
  forall t d, array_dimensions_eq t d ->
  array_dimensions_eq (ReferenceType (ArrayType t)) (d+1)
.
End Auxiliary.
Import Auxiliary.

(** Module contains formalizations of static constraints (4.8.1).*)
Module StaticConstraints.

(** The code array must not be empty, so the code_length item cannot have the value 0. *)
Definition constraint_1_1: code_constraint_type :=
  fun bm =>
    let pc0 := BYTECODEMETHOD.firstAddress bm in
    exists ins, BYTECODEMETHOD.instructionAt bm pc0 = Some ins.

(* The value of the code_length item must be less than 65536. *)
(* constraint_1_2 - to count intructions? *)

(* The opcode of the first instruction in the code array begins at index 0. *)
(* constraint_1_3 - has no sense in Bicolano *)

(* Only instances of the instructions documented in Section 6.4 may appear in the code array.
    Instances of instructions using the reserved opcodes (§6.2) or any opcodes not documented
    in this specification may not appear in the code array. *)
(* constraint_1_4 - guaranteed by the inductive definition of instructions *)

(* For each instruction in the code array except the last, the index of the opcode of the next instruction
    equals the index of the opcode of the current instruction plus the length of that instruction, including
    all its operands. The wide instruction is treated like any other instruction for these purposes; the opcode
    specifying the operation that a wide instruction is to modify is treated as one of the operands of that wide
    instruction. That opcode must never be directly reachable by the computation. *)
(* constraint_1_5 - has no sense in Bicolano *)

(* The last byte of the last instruction in the code array must be the byte at index code_length-1. *)
(* constraint_1_6 - has no sense in Bicolano *)

(** The target of each jump and branch instruction (jsr, jsr_w, goto, goto_w, ifeq, ifne, ifle, iflt, ifge, ifgt, ifnull,
     ifnonnull, if_icmpeq, if_icmpne, if_icmple, if_icmplt, if_icmpge, if_icmpgt, if_acmpeq, if_acmpne) must be
     the opcode of an instruction within this method. The target of a jump or branch instruction must never be
     the opcode used to specify the operation to be modified by a wide instruction; a jump or branch target
     may be the wide instruction itself. *)
Definition constraint_1_7: code_constraint_type :=
  fun bm =>
    let instruction_valid pc ins :=
      match ins with
      | Goto off => target_ok bm pc off
      | If_acmp _ off => target_ok bm pc off
      | If_icmp _ off => target_ok bm pc off
      | If0 _ off => target_ok bm pc off
      | Ifnull _ off => target_ok bm pc off
      (*| jsr | jsr_w | goto_w *)
      | _ => True
      end
    in for_all_instructions instruction_valid bm.

(** Each target, including the default, of each tableswitch instruction must be the opcode of an instruction within
     this method. Each tableswitch instruction must have a number of entries in its jump table that is consistent
     with the value of its low and high jump table operands, and its low value must be less than or equal to its high
     value. No target of a tableswitch instruction may be the opcode used to specify the operation to be modified
     by a wide instruction; a tableswitch target may be a wide instruction itself. *)
Definition constraint_1_8: code_constraint_type :=
  fun bm =>
    let instruction_valid pc ins :=
      match ins with
      | Tableswitch def low high l =>
        (low <= high) /\ Z_of_nat (length l) = high - low + 1 /\
        target_ok bm pc def /\
        for_all_in_list OFFSET.t (fun el => target_ok bm pc el) l
      | _ => True
      end
    in for_all_instructions instruction_valid bm.

(** Each target, including the default, of each lookupswitch instruction must be the opcode of an instruction
     within this method. Each lookupswitch instruction must have a number of match-offset pairs that is consistent
     with the value of its npairs operand. The match-offset pairs must be sorted in increasing numerical order
     by signed match value. No target of a lookupswitch instruction may be the opcode used to specify the
     operation to be modified by a wide instruction; a lookupswitch target may be a wide instruction itself. *)
Definition constraint_1_9: code_constraint_type :=
  fun bm =>
    let instruction_valid pc ins :=
      match ins with
      | Lookupswitch def l =>
        sorted_list Z Zlt (map (@fst Z OFFSET.t) l) /\
        target_ok bm pc def /\
        for_all_in_list (Z*OFFSET.t) (fun el => target_ok bm pc (snd el)) l
      | _ => True
      end
    in for_all_instructions instruction_valid bm.

(* The operand of each ldc instruction must be a valid index into the constant_pool table. The operands
   of each ldc_w instruction must represent a valid index into the constant_pool table. In both cases
   the constant pool entry referenced by that index must be of type CONSTANT_Integer, CONSTANT_Float,
   or CONSTANT_String. *)
(* constraint_1_10 - ldc; not supported by Bicolano *)

(* The operands of each ldc2_w instruction must represent a valid index into the constant_pool table.
   The constant pool entry referenced by that index must be of type CONSTANT_Long or CONSTANT_Double.
   In addition, the subsequent constant pool index must also be a valid index into the constant pool, and the constant
   pool entry at that index must not be used. *)
(* constraint_1_11 - ldc_2; not supported by Bicolano *)

(* The operands of each getfield, putfield, getstatic, and putstatic instruction must represent a valid index into
    the constant_pool table. The constant pool entry referenced by that index must be of type CONSTANT_Fieldref. *)
(* constraint_1_12 - No constant pool in Bicolano. I guess it is guaranteed by the structure of Bicolano *)

(* The indexbyte operands of each invokevirtual, invokespecial, and invokestatic instruction must represent
   a valid index into the constant_pool table. The constant pool entry referenced by that index must be of type
   CONSTANT_Methodref. *)
(* constraint_1_13 - No constant pool in Bicolano. I guess it is guaranteed by the structure of Bicolano *)

(* Only the invokespecial instruction is allowed to invoke an instance initialization method (§3.9). No other method
    whose name begins with the character '<' ('\u003c') may be called by the method invocation instructions. In particular,
    the class or interface initialization method specially named <clinit> is never called explicitly from Java virtual machine
    instructions, but only implicitly by the Java virtual machine itself. *)
(* constraint 1_14: TODO  Don't know how to check method names *)

(* The indexbyte operands of each invokeinterface instruction must represent a valid index into the constant_pool table.
    The constant pool entry referenced by that index must be of type CONSTANT_InterfaceMethodref. The value
    of the count operand of each invokeinterface instruction must reflect the number of local variables necessary
    to store the arguments to be passed to the interface method, as implied by the descriptor of the
    CONSTANT_NameAndType_info structure referenced by the CONSTANT_InterfaceMethodref constant
    pool entry. The fourth operand byte of each invokeinterface instruction must have the value zero. *)
(* constraint_1_15 - No constant pool in Bicolano. I guess it is guaranteed by the structure of Bicolano *)


(* The operands of each instanceof, checkcast, new, and anewarray instruction and the indexbyte operands
    of each multianewarray instruction must represent a valid index into the constant_pool table. The constant
    pool entry referenced by that index must be of type CONSTANT_Class. *)
(* constraint_1_16 - No constant pool in Bicolano. I guess it is guaranteed by the structure of Bicolano *)

(** No anewarray instruction may be used to create an array of more than 255 dimensions. *)
Definition constraint_1_17: code_constraint_type :=
  fun bm =>
    let instruction_valid pc ins :=
      match ins with
      | Newarray t =>
        exists d, d <= 255 /\
        array_dimensions_eq t d
      | _ => True
      end
    in for_all_instructions instruction_valid bm.

(** No new instruction may reference a CONSTANT_Class constant_pool table entry representing an array class.
     The new instruction cannot be used to create an array. The new instruction also cannot be used to create
      an instance of an interface or an instance of an abstract class. *)
Definition constraint_1_18: program_constraint_type :=
  fun prog =>
  let instruction_valid pc ins :=
    match ins with
    | New cn =>
      match (PROG.class prog cn) with
      | Some cl =>
        ~ CLASS.isAbstract cl = true
      | None => False
      end
    | _ => True
    end
  in for_all_classes (for_all_methods (for_body (for_all_instructions instruction_valid))) prog.

(** A multianewarray instruction must be used only to create an array of a type that has at least as many
    dimensions as the value of its dimensions operand. That is, while a multianewarray instruction is not required
    to create all of the dimensions of the array type referenced by its indexbyte operands, it must not attempt
    to create more dimensions than are in the array type. The dimensions operand of each multianewarray instruction
    must not be zero.*)
(* Multianewarray is not currently supported 
Definition constraint_1_19: code_constraint_type :=
  fun bm =>
    let instruction_valid: instruction_constraint_type :=
      fun pc ins =>
      match ins with
      | Multianewarray t dims =>
        exists d, dims <= d /\
        array_dimensions_eq (ReferenceType t) d
      | _ => True
      end
    in for_all_instructions instruction_valid bm.
*)

(* The atype operand of each newarray instruction must take one of the values T_BOOLEAN (4), T_CHAR (5),
    T_FLOAT (6), T_DOUBLE (7), T_BYTE (8), T_SHORT (9), T_INT (10), or T_LONG (11). *)
(* constraint_1_20 newarray - guaranteed by Bicolano structure. *)

(** The index operand of each iload, fload, aload, istore, fstore, astore, iinc, and ret instruction must be
     a nonnegative integer no greater than max_locals-1 *)
Definition constraint_1_21: method_constraint_type :=
  fun m => forall bm, METHOD.body m = Some bm ->
    let var_ok: Var -> Prop :=
      fun v =>
        let i := Var_toN v in
        lt i (BYTECODEMETHOD.max_locals bm)
    in
    let instruction_valid: instruction_constraint_type :=
      fun pc ins =>
      match ins with
      | Vload k x => var_ok x
      | Vstore k x => var_ok x
      | Iinc x z => var_ok x
(* fload | fstore | ret *)
      | _ => True
      end
    in for_body (for_all_instructions instruction_valid) m.

(* The implicit index of each iload_<n>, fload_<n>, aload_<n>, istore_<n>, fstore_<n>, and astore_<n>
    instruction must be no greater than the value of max_locals-1. *)
(* constraint_1_22 - in Bicolano equivalent to constraint_1_21. *)

(* The index operand of each lload, dload, lstore, and dstore instruction must be no greater than the value of max_locals-2. *)
(* constraint_1_23 - no double values in Bicolano. *)

(* The implicit index of each lload_<n>, dload_<n>, lstore_<n>, and dstore_<n> instruction must be
    no greater than the value of max_locals-2. *)
(* constraint_1_24 - no double values in Bicolano. *)

(* The indexbyte operands of each wide instruction modifying an iload, fload, aload, istore, fstore, astore, ret, or iinc
     instruction must represent a nonnegative integer no greater than max_locals-1. The indexbyte operands of each
    wide instruction modifying an lload, dload, lstore, or dstore instruction must represent a nonnegative integer no greater
    than max_locals-2. *)
(* constraint_1_24 - no wide in Bicolano. *)

End StaticConstraints.
Close Scope Z_scope.

(** Module implements "state shadow", which consists of number and type of values
on the stack and in the local variables array. *)
Module StateShadow.
Require Import Array.

Definition ValueShadow := type.

Definition type_to_value_shadow: type -> ValueShadow :=
  fun typ => typ
.

  (* This is neccesary to make type of arrays *)
  Module ShadowElement : ELEMENT_TYPE with Definition carrier_t := ValueShadow.
    Definition carrier_t := ValueShadow.
  End ShadowElement.

  Module ArrayShadowM : ARRAY_TYPE with Definition carrier_t := ValueShadow := ArrayAsList (ShadowElement).
  Import ArrayShadowM.
  Definition StackShadow := list ValueShadow.
  Definition ArrayShadow := array_t.
  
  Record StateShadow: Set := ShMake {
    get_pc : PC;
    get_array : ArrayShadow;
    get_stack : StackShadow
  }.

Definition is_initial_state_shadow : Method -> StateShadow -> Prop :=
  fun m stsh => forall bm,
  METHOD.body m = Some bm ->
  let pc := BYTECODEMETHOD.firstAddress bm in
  let init_arr_head := (METHODSIGNATURE.parameters (METHOD.signature m)) in
  let arr := arr_from_list_ext (map type_to_value_shadow init_arr_head) (BYTECODEMETHOD.max_locals bm) in
  stsh = ShMake pc arr nil.


Definition is_initial_exception_state_shadow : Method -> StateShadow -> Prop :=
  fun m stsh => forall bm h,
  METHOD.body m = Some bm ->
  In h (BYTECODEMETHOD.exceptionHandlers bm) ->
  let pc := EXCEPTIONHANDLER.handler h in
  let init_arr_head := (METHODSIGNATURE.parameters (METHOD.signature m)) in
  let arr := arr_from_list_ext (map type_to_value_shadow init_arr_head) (BYTECODEMETHOD.max_locals bm) in
  stsh = ShMake pc arr nil.


  (** Copy and Paste from Domain, to be moved to Program,
   compatibility between ArrayKind and type *) 
  Inductive compat_ArrayKind_type : ArrayKind -> type -> Prop :=
    | compat_ArrayKind_type_ref : forall rt,
        compat_ArrayKind_type Aarray (ReferenceType rt)
    | compat_ArrayKind_type_int : 
        compat_ArrayKind_type Iarray (PrimitiveType INT)
    | compat_ArrayKind_type_byte : 
        compat_ArrayKind_type Barray (PrimitiveType BYTE)
    | compat_ArrayKind_type_bool : 
        compat_ArrayKind_type Barray (PrimitiveType BOOLEAN)
    | compat_ArrayKind_type_short : 
        compat_ArrayKind_type Sarray (PrimitiveType SHORT).

Inductive possible_step (m:Method) : StateShadow -> StateShadow -> Prop :=
| PossibleStep_Vaload: forall k pc1 arr st pc2 t,
  instructionAt m pc1 = Some (Vaload k) ->
  next m pc1 = Some pc2 ->
  compat_ArrayKind_type k t ->
  possible_step m
    (ShMake pc1 arr ((PrimitiveType INT)::(ReferenceType (ArrayType t))::st))
    (ShMake pc2 arr (t::st)).

Inductive possible_state_shadow : Method -> StateShadow -> Prop :=
| PossibleStSh_Init :  forall m sh,
  is_initial_state_shadow m sh ->
  possible_state_shadow m sh
| PossibleStSh_Step : forall m sh1 sh2,
  possible_state_shadow m sh1 ->
  possible_step m sh1 sh2 ->
  possible_state_shadow m sh2
(*TODO exceptions *)
.
End StateShadow.

(** Module contains formalizations of dynamic constraints (4.8.2).*)
Module DynamicConstraints (D:SEMANTIC_DOMAIN with Module Prog := Prog').

  Module SmallStepImpl <: SMALLSTEP with Module Dom := D := MakeSmallStep (D).

  Import D.
  Import SmallStepImpl.

  Definition kind_of_value: value -> ValKind :=
    fun v =>
    match v with
    | Num (I _) => Ival
    | Num (B _) => Ival
    | Num (Sh _) => Ival
    | Ref _ => Aval
    | Null => Aval
    end.

  Definition kind_of_type: type -> ValKind :=
    fun t =>
    match t with
      | ReferenceType _ => Aval
      | PrimitiveType _ => Ival
    end.

   Inductive parameters_kind_ok: list type -> list value -> Prop :=
   | parameters_kind_ok_nil:
     forall lv,
     parameters_kind_ok nil lv
   | parameters_kind_ok_cons:
     forall lv lt v t,
     parameters_kind_ok lt lv ->
     kind_of_type t = kind_of_value v ->
     parameters_kind_ok (t::lt) (v::lv).


(* TODO program initial state *)
Definition is_initial_state: Program -> State.t -> Prop :=
fun p st => True.

Inductive step_closure (p:Program) : State.t -> State.t -> Prop :=
 | step_closure_refl : forall st,
    step_closure p st st
 | step_closure_step : forall st1 st2 st3,
    step_closure p st1 st2 -> 
    step p st2 st3 -> 
    step_closure p st1 st3.

Inductive is_possible_state (p:Program): State.t -> Prop :=
| is_possible_state_init: forall st,
  is_initial_state p st ->
  is_possible_state p st
| is_possible_state_step: forall st1 st2,
  is_possible_state p st1 ->
  step p st1 st2 ->
  is_possible_state p st2.

(* Each instruction must only be executed with the appropriate type and number of arguments
    in the operand stack and local variable array, regardless of the execution path that leads
    to its invocation. An instruction operating on values of type int is also permitted to operate
    on values of type boolean, byte, char, and short. *)
Definition constraint_2_1: program_constraint_type :=
  fun p =>
  forall m pc h st arr sf bm,
  is_possible_state p (St h (Fr m pc st arr) sf) ->
  METHOD.body m = Some bm ->
  match   BYTECODEMETHOD.instructionAt bm pc with
  | Some Aconst_null => True
  | Some Arraylength  =>
    match st with
    | ((Ref loc)::tl) => True
      (*? typeof h loc = Some (LocationArray _ _) , I assume it is not needed, and don't check the heap in the following cases ?*)
    | (Null::tl) => True (* however, an exception will be thrown in this case *)
    | _ => False
    end
  | Some Athrow =>
    match st with
    | (v::tl) => kind_of_value v = Aval
    | _ => False
    end
  | Some (Checkcast t) =>
    match st with
    | (v::tl) => kind_of_value v = Aval
    | _ => False
    end
  | Some (Const t z) => True
  | Some Pop =>
     (* all values in Bicolano are the category 1. computational type, so the verification of Pop, Dup etc. is easy;
         to be changed when double and long values come *)
    match st with
    | (v::tl) => True
    | _ => False
    end
  | Some Pop2 =>
    match st with
    | (v1::v2::tl) => True
    | _ => False
    end
  | Some Swap => 
    match st with
    | (v1::v2::tl) => True
    | _ => False
    end
  | Some Dup_x1 =>
    match st with
    | (v1::v2::tl) => True
    | _ => False
    end
  | Some Dup =>
    match st with
    | (v::tl) => True
    | _ => False
    end
  | Some Dup_x2 =>
    match st with
    | (v1::v2::v3::tl) => True
    | _ => False
    end
  | Some Dup2 =>
    match st with
    | (v1::v2::tl) => True
    | _ => False
    end
  | Some Dup2_x1 =>
    match st with
    | (v1::v2::v3::tl) => True
    | _ => False
    end
  | Some Dup2_x2 =>
    match st with
    | (v1::v2::v3::v4::tl) => True
    | _ => False
    end
  | Some (Getfield f) =>
    match st with
    | (v::tl) => kind_of_value v = Aval
    | _ => False
    end
  | Some (Getstatic f) => True
  | Some (Goto o) => True
  | Some I2b =>
    match st with
    | (v::tl) => kind_of_value v = Ival
    | _ => False
    end
  | Some I2s =>
    match st with
    | (v::tl) => kind_of_value v = Ival
    | _ => False
    end
  | Some (Ibinop op) =>
    match st with
    | (v1::v2::tl) => kind_of_value v1 = Ival /\ kind_of_value v2 = Ival
    | _ => False
    end
  | Some (If_acmp cmp o) =>
    match st with
    | (v1::v2::tl) => kind_of_value v1 = Aval /\ kind_of_value v2 = Aval
    | _ => False
    end
  | Some (If_icmp cmp o) => 
    match st with
    | (v1::v2::tl) => kind_of_value v1 = Ival /\ kind_of_value v2 = Ival
    | _ => False
    end
  | Some (If0 cmp o) =>
    match st with
    | (v::tl) => kind_of_value v = Ival
    | _ => False
    end
  | Some (Ifnull cmp o) =>
    match st with
    | (v::tl) => kind_of_value v = Aval
    | _ => False
    end
  | Some (Iinc x z) =>
    exists v,
    LocalVar.get arr x = Some v /\ 
    kind_of_value v = Ival
  | Some Ineg =>
    match st with
    | (v::tl) => kind_of_value v = Ival
    | _ => False
    end
  | Some (Instanceof t) =>
    match st with
    | (v::tl) => kind_of_value v = Aval
    | _ => False
    end
  | Some (Invokeinterface m) =>
    match st with
    | (v::tl) => kind_of_value v = Aval
      /\ parameters_kind_ok (METHODSIGNATURE.parameters (snd m)) tl
    | _ => False
    end    
  | Some (Invokespecial m) =>
    match st with
    | (v::tl) => kind_of_value v = Aval
      /\ parameters_kind_ok (METHODSIGNATURE.parameters (snd m)) tl
    | _ => False
    end    
  | Some (Invokestatic m) =>
    parameters_kind_ok (METHODSIGNATURE.parameters (snd m)) st
  | Some (Invokevirtual m) =>
    match st with
    | (v::tl) => kind_of_value v = Aval
      /\ parameters_kind_ok (METHODSIGNATURE.parameters (snd m)) tl
    | _ => False
    end    
  | Some (Lookupswitch def l) =>
    match st with
    | (v::tl) => kind_of_value v = Ival
    | _ => False
    end
  | Some (New c) => True
  | Some (Newarray t) =>
    match st with
    | (v::tl) => kind_of_value v = Ival
    | _ => False
    end
  | Some Nop => True
  | Some (Putfield f) =>
    match st with
    | (v::ref::tl) => kind_of_value ref = Aval /\
      kind_of_type (FIELDSIGNATURE.type (snd f)) = kind_of_value v
    | _ => False
    end
  | Some (Putstatic f) =>
    match st with
    | (v::tl) => kind_of_type (FIELDSIGNATURE.type (snd f)) = kind_of_value v
    | _ => False
    end
  | Some Return => True
  | Some (Tableswitch def low high l) =>
    match st with
    | (v::tl) => kind_of_value v = Ival
    | _ => False
    end    
  | Some (Vaload k) =>
    match st with
    | (ind::ref::tl) => kind_of_value ind = Ival /\ kind_of_value ref = Aval
    | _ => False
    end
  | Some (Vastore k) =>
    match st with
    | (v::ind::ref::tl) => kind_of_value ind = Ival /\ kind_of_value ref = Aval
    | _ => False
    end
  | Some (Vload k x) =>
    exists v,
    LocalVar.get arr x = Some v /\ 
    kind_of_value v = k
  | Some (Vreturn k) =>
    match st with
    | (v::tl) => kind_of_value v = k
    | nil => False
    end
  | Some (Vstore k x) =>
    match st with
    | (v::tl) => kind_of_value v = k
    | nil => False
    end
  | None => True
  end
.

(** If an instruction can be executed along several different execution paths, the operand stack must have
     the same depth (§3.6.2) prior to the execution of the instruction, regardless of the path taken. *)
Definition constraint_2_2: program_constraint_type :=
  fun p =>
  forall m pc h1 h2 st1 st2 arr1 arr2 sf1 sf2,
  is_possible_state p (St h1 (Fr m pc st1 arr1) sf1) ->
  is_possible_state p (St h2 (Fr m pc st2 arr2) sf2) ->
  length st1 = length st2.

(* At no point during execution can the order of the local variable pair holding a value of type long
   or double be reversed or the pair split up. At no point can the local variables of such a pair be operated
   on individually. *)
(* Definition constraint_2_3 - no long or double types in Bicolano *)

(** No local variable (or local variable pair, in the case of a value of type long or double)
   can be accessed before it is assigned a value. *)
(* TODO Assert that uninitialized array cell has None value *)
Definition constraint_2_4: program_constraint_type :=
  fun p =>
  forall m pc h st arr sf bm k x,
  is_possible_state p (St h (Fr m pc st arr) sf) ->
  METHOD.body m = Some bm ->
  BYTECODEMETHOD.instructionAt bm pc = Some (Vload k x) ->
  exists v,
  LocalVar.get arr x = Some v.

(** At no point during execution can the operand stack grow to a depth (§3.6.2)
     greater than that implied by the max_stack item. *)
Definition constraint_2_5: program_constraint_type :=
  fun p =>
  forall m pc h st arr sf,
  is_possible_state p (St h (Fr m pc st arr) sf) ->
  METHOD.valid_stack_size m (length st).

(** At no point during execution can more values be popped from the operand stack
     than it contains. *)
Definition constraint_2_6: program_constraint_type :=
  fun p =>
  forall m pc h st arr sf bm,
  is_possible_state p (St h (Fr m pc st arr) sf) ->
  METHOD.body m = Some bm ->
  let len := (length st) in
  match   BYTECODEMETHOD.instructionAt bm pc with
  | Some Arraylength => len >= 1
  | Some Athrow => len >= 1
  | Some (Checkcast _) => len >= 1
  | Some Dup => len >= 1
  | Some Dup_x1 => len >= 2
  | Some Dup_x2 => len >= 3
  | Some Dup2 => len >= 2
  | Some Dup2_x1 => len >= 3
  | Some Dup2_x2 => len >= 4
  | Some (Getfield _) => len >= 1
  | Some I2b => len >= 1
  | Some I2s => len >= 1
  | Some (Ibinop op) => len >= 2
  | Some (If_acmp _ _) => len >= 2
  | Some (If_icmp _ _)  => len >= 2
  | Some (If0 _ _)  => len >= 1
  | Some (Ifnull _ _)  => len >= 1
  | Some Ineg => len >= 1
  | Some (Instanceof _) => len >= 1
  | Some (Invokeinterface ms) => len >= 1 + (length (METHODSIGNATURE.parameters (snd ms)))
  | Some (Invokespecial ms) => len >= 1 + (length (METHODSIGNATURE.parameters (snd ms)))
  | Some (Invokestatic ms) => len >= (length (METHODSIGNATURE.parameters (snd ms)))
  | Some (Invokevirtual ms) => len >= 1 + (length (METHODSIGNATURE.parameters (snd ms)))
  | Some (Lookupswitch _ _) => len >= 1
  | Some (Newarray _) => len >= 1
  | Some Pop => len >= 1
  | Some Pop2 => len >= 2
  | Some (Putfield _) => len >= 2
  | Some (Putstatic _) => len >= 1
  | Some Swap => len >= 2
  | Some (Tableswitch _ _ _ _) => len >= 1
  | Some (Vaload k) => len >= 2
  | Some (Vastore k) => len >= 3 (* to be changed when double values comes *)
  | Some (Vload k _) => True
  | Some (Vreturn k) => len >= 1
  | Some (Vstore k x) => len >= 1
  | Some _ => True
  | None => True
  end
.

(* Each invokespecial instruction must name an instance initialization method (§3.9),
    a method in the current class, or a method in a superclass of the current class. *)
(* Definition constraint_2_7 - TODO *)

(* When the instance initialization method (§3.9) is invoked, an uninitialized class instance must be
    in an appropriate position on the operand stack. An instance initialization method must never be
    invoked on an initialized class instance. *)
(* ??? *)

(* When any instance method is invoked or when any instance variable is accessed, the class instance
    that contains the instance method or instance variable must already be initialized. *)
(* ??? *)

(* There must never be an uninitialized class instance on the operand stack or in a local variable when
    any backwards branch is taken. There must never be an uninitialized class instance in a local variable
    in code protected by an exception handler. However, an uninitialized class instance may be on the operand
    stack in code protected by an exception handler. When an exception is thrown, the contents of the operand stack
    are discarded. *)
(* ??? *)

(* Each instance initialization method (§3.9), except for the instance initialization method derived from the constructor
   of class Object, must call either another instance initialization method of this or an instance initialization method
   of its direct superclass super before its instance members are accessed. However, instance fields of this that are
   declared in the current class may be assigned before calling any instance initialization method. *)

(* The arguments to each method invocation must be method invocation compatible (§2.6.8) with the method
   descriptor (§4.3.3). *)

(* The type of every class instance that is the target of a method invocation instruction must be assignment compatible
   (§2.6.7) with the class or interface type specified in the instruction. *)

(* Each return instruction must match its method's return type. If the method returns a boolean, byte, char, short, or int,
    only the ireturn instruction may be used. If the method returns a float, long, or double, only an freturn, lreturn,
    or dreturn instruction, respectively, may be used. If the method returns a reference type, it must do so using
    an areturn instruction, and the type of the returned value must be assignment compatible (§2.6.7) with the return
    descriptor (§4.3.3) of the method. All instance initialization methods, class or interface initialization methods,
    and methods declared to return void must use only the return instruction. *)

(* If getfield or putfield is used to access a protected field of a superclass, then the type of the class instance
    being accessed must be the same as or a subclass of the current class. If invokevirtual or invokespecial
    is used to access a protected method of a superclass, then the type of the class instance being accessed
    must be the same as or a subclass of the current class. *)

(* The type of every class instance accessed by a getfield instruction or modified by a putfield instruction must be
    assignment compatible (§2.6.7) with the class type specified in the instruction. *)

(* The type of every value stored by a putfield or putstatic instruction must be compatible with the descriptor
    of the field (§4.3.2) of the class instance or class being stored into. If the descriptor type is boolean, byte, char,
    short, or int, then the value must be an int. If the descriptor type is float, long, or double, then the value must be
    a float, long, or double, respectively. If the descriptor type is a reference type, then the value must be of a type
    that is assignment compatible (§2.6.7) with the descriptor type. *)

(* The type of every value stored into an array of type reference by an aastore instruction must be assignment
    compatible (§2.6.7) with the component type of the array. *)

(* Each athrow instruction must throw only values that are instances of class Throwable or of subclasses of Throwable. *)

(* Execution never falls off the bottom of the code array. *)

(* No return address (a value of type returnAddress) may be loaded from a local variable. *)

(* The instruction following each jsr or jsr_w instruction may be returned to only by a single ret instruction. *)

(* No jsr or jsr_w instruction may be used to recursively call a subroutine if that subroutine is already present 
    in the subroutine call chain. (Subroutines can be nested when using try-finally constructs from within a finally clause.
    For more information on Java virtual machine subroutines, see §4.9.6.) *)

(* Each instance of type returnAddress can be returned to at most once. If a ret instruction returns to a point in the
   subroutine call chain above the ret instruction corresponding to a given instance of type returnAddress, then that
   instance can never be used as a return address.  *)

End DynamicConstraints.

End BtcVerif.
