(* <Insert License Here>

    $Id: BigStep.v 68 2006-02-02 15:06:27Z davidpichardie $ *)

(** Equivalence between Smal and Big step semantics.
    The main theorem is Equiv_SmallStep_BigStep at the end of the file.

 @author David Pichardie *)

Require Export BigStep.
Require Export SmallStep.
Require Import Relation_Operators.
Require Import Omega.

Module Type ABSTRACT_SMALLSTEP.

  Parameter program IntraState ReturnState InitCallState ContextCall IntraException: Type.

  Inductive State : Type := 
    mks : ContextCall -> IntraState -> State
  | mke : ContextCall -> IntraException -> State.

  Parameter IntraSmallStep : program -> ContextCall -> IntraState -> IntraState  -> Prop.
  Parameter ReturnSmallStep : program -> ContextCall -> IntraState -> ReturnState -> Prop.
  Parameter IntraSmallStepException : program -> ContextCall -> IntraState -> IntraException -> Prop.
  Parameter CallStep : program -> ContextCall -> IntraState -> InitCallState -> Prop.

  Parameter CaughtLocalException : program -> ContextCall -> IntraException -> IntraState -> Prop.
  Parameter UncaughtLocalException : program -> ContextCall -> IntraException -> ReturnState -> Prop.

  Parameter SmallStep : program -> State -> State -> Prop.

  Parameter init_context : ContextCall -> IntraState -> InitCallState -> ContextCall.
  Parameter init_intra : ContextCall -> IntraState -> InitCallState -> IntraState -> Prop.
  Parameter return_context : ContextCall -> ContextCall -> Prop.
  Parameter return_intra : ContextCall -> ReturnState -> IntraState -> Prop.
  Parameter return_exception : ContextCall -> ReturnState -> IntraException -> Prop.
  Parameter call_and_return : ContextCall -> IntraState -> InitCallState -> IntraState -> ReturnState -> IntraState -> Prop.
  Parameter call_and_return_exception : ContextCall -> IntraState -> InitCallState -> IntraState -> ReturnState -> IntraException -> Prop.

  Parameter UncaughtLocalException_functionnal : forall p ct se ret1 ret2,
    UncaughtLocalException p ct se ret1 ->
    UncaughtLocalException p ct se ret2 -> 
    ret1 = ret2.

  Parameter return_context_functionnal : forall ct ct1 ct2,
    return_context ct ct1 ->
    return_context ct ct2 ->
    ct1=ct2.
  Parameter return_context_init_context_prop2 : forall ct s si,
    return_context (init_context ct s si) ct.

  Parameter return_context_no_cycle : forall ct, 
    ~ clos_trans _ return_context ct ct.

  Parameter ReturnSmallStep_not_return_exception : forall P ct s sr se,
   ReturnSmallStep P ct s sr -> return_exception ct sr se -> False.

  Parameter return_intra_not_UncaughtLocalException : forall P ct se sr s,
    UncaughtLocalException P ct se sr ->
    return_intra ct sr s -> False.

  Parameter call_and_return_equiv1 : forall ct s si s0 sr s',
      init_intra ct s si s0 -> 
      return_intra (init_context ct s si) sr s' -> 
      call_and_return ct s si s0 sr s'.

  Parameter call_and_return_equiv2 : forall ct s si s0 sr s',
      call_and_return ct s si s0 sr s' ->
      init_intra ct s si s0 /\ return_intra (init_context ct s si) sr s'.

  Parameter call_and_return_exception_equiv1 : forall ct s si s0 sr s',
      init_intra ct s si s0 -> 
      return_exception (init_context ct s si) sr s' -> 
      call_and_return_exception ct s si s0 sr s'.

  Parameter call_and_return_exception_equiv2 : forall ct s si s0 sr s',
      call_and_return_exception ct s si s0 sr s' ->
      init_intra ct s si s0 /\ return_exception (init_context ct s si) sr s'.

  Inductive SmallStep_case (P:program) : State -> State -> Prop :=
  | step_simpl : forall ct s1 s2,
	IntraSmallStep P ct s1 s2 ->
        SmallStep_case P (mks ct s1) (mks ct s2)
  | step_call : forall ct s si s0,
        CallStep P ct s si ->
        init_intra ct s si s0 ->
        SmallStep_case P (mks ct s) (mks (init_context ct s si) s0)
  | step_return : forall ct s sr ct' s' ,
	ReturnSmallStep P ct s sr ->
	return_context ct ct' ->
	return_intra ct sr s' ->
        SmallStep_case P (mks ct s) (mks ct' s')
  | step_thrown_exception : forall ct s se,
       IntraSmallStepException P ct s se ->
       SmallStep_case P (mks ct s) (mke ct se)
  | step_caught_exception : forall ct s se,
       CaughtLocalException P ct se s ->
       SmallStep_case P (mke ct se) (mks ct s)
  | step_uncaught_exception : forall ct1 se1 ct2 se2 sr,
       UncaughtLocalException P ct1 se1 sr ->
       return_context ct1 ct2 ->
       return_exception ct1 sr se2 ->
       SmallStep_case P (mke ct1 se1) (mke ct2 se2).

  Parameter SmallStep_implies_SmallStep_case : forall P st1 st2,
    SmallStep P st1 st2 ->
    SmallStep_case P st1 st2.

  Parameter SmallStep_case_implies_SmallStep : forall P st1 st2,
    SmallStep_case P st1 st2 ->
    SmallStep P st1 st2.

End ABSTRACT_SMALLSTEP.

Module NewSemantics (S:ABSTRACT_SMALLSTEP).
  Import S.
  Implicit Arguments SmallStep_implies_SmallStep_case.
  Implicit Arguments SmallStep_case_implies_SmallStep.

(** Semantics with intermediate reachables states *)
Inductive IntraNormalStep2 (P:program) : ContextCall -> IntraState -> IntraState -> Prop :=
  | IntraNormalStep2_IntraSmallStep : forall ct s1 s2,
      IntraSmallStep P ct s1 s2 -> 
      IntraNormalStep2 P ct s1 s2
  | IntraNormalStep2_call_return_value' : forall ct s si sr s' s0,
      CallStep P ct s si ->
      call_and_return ct s si s0 sr s' ->
      IntraBigStep2 P (init_context ct s si) s0 sr ->
      IntraNormalStep2 P ct s s'

with IntraExceptionStep2 (P:program) : ContextCall -> IntraState -> IntraException -> Prop :=
  | IntraExceptionStep2_IntraSmallStepException : forall ct s1 se,
      IntraSmallStepException P ct s1 se ->
      IntraExceptionStep2 P ct s1 se
  | IntraExceptionStep2_call_return_value : forall ct s si sr se s0,
      CallStep P ct s si ->
      call_and_return_exception ct s si s0 sr se ->
      IntraBigStep2 P (init_context ct s si) s0 sr ->
      IntraExceptionStep2 P ct s se

with IntraStarStep (P:program) : ContextCall -> IntraState -> IntraState -> Prop :=
  | IntraStarStep_refl : forall ct s,
      IntraStarStep P ct s s
  | IntraStarStep_next : forall ct s1 s2 s3,
      IntraNormalStep2 P ct s1 s2 -> 
      IntraStarStep P ct s2 s3 ->
      IntraStarStep P ct s1 s3
  | IntraStarStep_exception_caught : forall ct s1 se s2 s3,
      IntraExceptionStep2 P ct s1 se ->
      CaughtLocalException P ct se s2 ->
      IntraStarStep P ct s2 s3 ->
      IntraStarStep P ct s1 s3

with IntraBigStep2 (P:program) : ContextCall -> IntraState -> ReturnState -> Prop :=
  | IntraBigStep2_normal : forall ct s1 s2 ret,
      IntraStarStep P ct s1 s2 ->
      ReturnSmallStep P ct s2 ret ->
      IntraBigStep2 P ct s1 ret
  | IntraBigStep2_exception_uncaught : forall ct s1 se s2 sr,
      IntraStarStep P ct s1 s2 ->
      IntraExceptionStep2 P ct s2 se ->
      UncaughtLocalException P ct se sr ->
      IntraBigStep2 P ct s1 sr.
Hint Constructors IntraBigStep2 IntraExceptionStep2 IntraNormalStep2 IntraStarStep.

(** We need strong induction principles *)
Scheme IntraBigStep2_ind_strong := Minimality for IntraBigStep2 Sort Prop
  with IntraStarStep_ind_strong := Minimality for IntraStarStep Sort Prop
  with IntraExceptionStep2_ind_strong := Minimality for IntraExceptionStep2 Sort Prop
  with IntraNormalStep2_ind_strong := Minimality for IntraNormalStep2 Sort Prop.

(****************************************************)
(** Equivalence between IntraBigStep2 and SmallStep *)
(****************************************************)

Definition prefix := (clos_refl_trans _ return_context).
Definition prefix_strict := (clos_trans _ return_context).
Hint Constructors clos_trans clos_refl_trans.
Hint Unfold prefix_strict prefix.

Lemma prefix_strict_prefix : forall ct1 ct2,
  prefix_strict ct1 ct2 -> prefix ct1 ct2.
Proof.
  induction 1; eauto.
Qed.
Hint Resolve prefix_strict_prefix.

Lemma prefix_strict_or_eq : forall ct1 ct2,
  prefix ct1 ct2 -> prefix_strict ct1 ct2 \/ ct1=ct2.
Proof.
  induction 1; auto.
  destruct IHclos_refl_trans2; subst; auto.
  destruct IHclos_refl_trans1; subst; eauto.
Qed.

Lemma return_context_no_cycle2 : forall ct1 ct2,
    prefix_strict ct1 ct2 ->
    prefix ct2 ct1 -> False.
Proof.
  induction 2.
  elim (return_context_no_cycle y); eauto.
  elim (return_context_no_cycle x); eauto.
  destruct (prefix_strict_or_eq _ _ H0_); subst; eauto.
Qed.

Lemma prefix_antysym : forall ct1 ct2,
  prefix ct1 ct2 -> prefix ct2 ct1 -> ct1 = ct2.
Proof.
  intros.
  elim prefix_strict_or_eq with (1:=H); intros.
  elim return_context_no_cycle2 with (1:=H1); auto.
  auto.
Qed.

Lemma return_context_no_cycle3 : forall ct1 ct2,
    return_context ct1 ct2 ->
    prefix ct2 ct1 -> False.
Proof.
  intros until 1; apply return_context_no_cycle2; auto.
Qed.

Lemma return_context_init_context_prop1 : forall ct1 ct2 s si,
    return_context (init_context ct1 s si) ct2 ->
    ct1=ct2.
Proof.
  intros; apply return_context_functionnal with (init_context ct1 s si); auto.
  apply return_context_init_context_prop2.
Qed.

Definition context (st : State) : ContextCall :=
  match st with
    mks ct _ => ct
  | mke ct _ => ct
  end.

Inductive star (A:Type) (R:A->A->Prop) : A -> A -> Prop :=
  star_refl : forall a, star A R a a
| star_R : forall a1 a2, R a1 a2 -> star A R a1 a2
| star_trans : forall a1 a2 a3, R a1 a2 -> star A R a2 a3 -> star A R a1 a3.
Implicit Arguments star [A].

Inductive uncaught (p:program) : (ContextCall*IntraException) -> (ContextCall*IntraException) -> Prop :=
 uncaught_def : forall ct1 ct2 se1 se2 sr, 
       UncaughtLocalException p ct1 se1 sr ->
       return_context ct1 ct2 ->
       return_exception ct1 sr se2 ->
       uncaught p (ct1,se1) (ct2,se2).

(** SmallStep reflexive transitive closure without leaving the current callstack prefix *)
Inductive SmallStepTransitiveClosure  (P:program) : State -> State -> Prop :=
 | SmallStepTransitiveClosure_refl : forall st,
    SmallStepTransitiveClosure P st st
 | SmallStepTransitiveClosure_step : forall st1 st2 st3,
    SmallStepTransitiveClosure P st1 st2 -> 
    prefix (context st3) (context st1) -> 
    SmallStep P st2 st3 -> 
    SmallStepTransitiveClosure P st1 st3.

Inductive SmallStepTransitiveClosure2  (P:program) : State -> State -> Prop :=
 | SmallStepTransitiveClosure_star : forall ct s1 s2,
    IntraStarStep P ct s1 s2 ->
    SmallStepTransitiveClosure2 P (mks ct s1) (mks ct s2)
 | SmallStepTransitiveClosure_call : forall ct s1 s2 si ct3 s3 s0,
    IntraStarStep P ct s1 s2 ->
    CallStep P ct s2 si ->
    init_intra ct s2 si s0 ->
    SmallStepTransitiveClosure2 P (mks (init_context ct s2 si) s0) (mks ct3 s3) ->
    SmallStepTransitiveClosure2 P (mks ct s1) (mks ct3 s3)
 | SmallStepTransitiveClosure_exception :  forall ct1 s1 ct2 s2 se2 ct3 se3,
    SmallStepTransitiveClosure2 P (mks ct1 s1) (mks ct2 s2) ->
    IntraSmallStepException P ct2 s2 se2 ->
    clos_refl_trans _ (uncaught P) (ct2,se2) (ct3,se3) ->
    prefix ct3 ct1 -> 
    SmallStepTransitiveClosure2 P (mks ct1 s1) (mke ct3 se3).

Hint Constructors SmallStepTransitiveClosure2.
Hint Constructors SmallStepTransitiveClosure.

Inductive intra : State -> IntraState -> Prop :=
  intra_def : forall ct s, intra (mks ct s) s.
Hint Constructors intra.
Inductive intrae : State -> IntraException -> Prop :=
  intrae_def : forall ct s, intrae (mke ct s) s.
Hint Constructors intrae.

Lemma SmallStepTransitiveClosure2_not_mke : forall P ct s st,
  ~ SmallStepTransitiveClosure2 P (mke ct s) st.
Proof.
  red; intros.
  inversion_clear H.
Qed.

Lemma SmallStepTransitiveClosure2_IntraNormalStep2_clos : forall P st2 st3,
   SmallStepTransitiveClosure2 P st2 st3 -> forall s2 s1,
   intra st2 s2 ->
   IntraNormalStep2 P (context st2) s1 s2 -> 
   SmallStepTransitiveClosure2 P (mks (context st2) s1) st3.
Proof.
  induction 1; simpl; intros.
  inversion_clear H0 in H1 H; eauto.
  inversion_clear H3 in H; eauto.
  constructor 3 with (2:=H0); eauto.
Qed.  

Lemma SmallStepTransitiveClosure2_IntraExceptionStep2_clos : forall P st2 st3,
   SmallStepTransitiveClosure2 P st2 st3 -> forall s2 se s1,
   intra st2 s2 ->
   IntraExceptionStep2 P (context st2) s1 se ->
   CaughtLocalException P (context st2) se s2 -> 
   SmallStepTransitiveClosure2 P (mks (context st2) s1) st3.
Proof.
  induction 1; intros.
  inversion_clear H0 in H H1 H2; simpl in *; eauto.
  inversion_clear H3 in H5 H4 H; simpl in *; eauto.
  constructor 3 with (2:=H0); eauto.
Qed.  

Lemma SmallStepTransitiveClosure2_IntraStarStep_clos : forall P ct st3 s2 s1,
   IntraStarStep P ct s1 s2 -> 
   SmallStepTransitiveClosure2 P (mks ct s2) st3 ->
   SmallStepTransitiveClosure2 P (mks ct s1) st3.
Proof.
  induction 1; intros; auto.
  apply SmallStepTransitiveClosure2_IntraNormalStep2_clos
   with (st2:=(mks ct s2)) (s2:=s2); auto. 
  eapply SmallStepTransitiveClosure2_IntraExceptionStep2_clos
   with (st2:=(mks ct s2)) (s2:=s2); eauto. 
Qed.

Lemma SmallStepTransitiveClosure2_trans : forall P st1 st2 st3,
  SmallStepTransitiveClosure2 P st1 st2 ->
  SmallStepTransitiveClosure2 P st2 st3 ->
  SmallStepTransitiveClosure2 P st1 st3.
Proof.
  intros.
  induction H.
  apply (SmallStepTransitiveClosure2_IntraStarStep_clos P) with (2:=H0); auto.
  destruct st3; eauto.
  assert (HH:=IHSmallStepTransitiveClosure2 H0);
  clear IHSmallStepTransitiveClosure2 H0 H3.
  inversion HH; subst.
  constructor 3 with (2:=H7); eauto.
  constructor 3 with (init_context ct s2 si); auto.
  constructor 1; apply return_context_init_context_prop2.
  inversion_clear H0.
Qed.

Lemma IntraStarStep_IntraNormalStep2_right_clos : forall P ct s1 s2 s3,
  IntraStarStep P ct s1 s2 ->
  IntraNormalStep2 P ct s2 s3 ->
  IntraStarStep P ct s1 s3.
Proof.
  induction 1; intros; eauto.
Qed.

Lemma IntraStarStep_IntraExceptionStep2_right_clos : forall P ct s1 s2 s3 se,
  IntraStarStep P ct s1 s2 ->
  IntraExceptionStep2 P ct s2 se ->
  CaughtLocalException P ct se s3 ->
  IntraStarStep P ct s1 s3.
Proof.
  induction 1; intros; eauto.
Qed.

Inductive SmallStepTransitiveClosure2_inv_cases (P:program) : State -> State -> Prop :=
 | SmallStepTransitiveClosure2_inv_case1 : forall ct s1 s2,
     IntraStarStep P ct s1 s2 ->
     SmallStepTransitiveClosure2_inv_cases P (mks ct s1) (mks ct s2)
 | SmallStepTransitiveClosure2_inv_case2 : forall st ct1 s1 s2 si s0,
     SmallStepTransitiveClosure2 P st (mks ct1 s1) ->
     CallStep P ct1 s1 si ->
     init_intra ct1 s1 si s0 ->
     IntraStarStep P (init_context ct1 s1 si) s0 s2 ->
     SmallStepTransitiveClosure2_inv_cases P st (mks (init_context ct1 s1 si) s2)
 | SmallStepTransitiveClosure2_inv_case3 : forall ct1 s1 ct2 s2 se2 ct3 se3,
    SmallStepTransitiveClosure2 P (mks ct1 s1) (mks ct2 s2) ->
    IntraSmallStepException P ct2 s2 se2 ->
    clos_refl_trans _ (uncaught P) (ct2,se2) (ct3,se3) ->
    prefix ct3 ct1 -> 
    SmallStepTransitiveClosure2_inv_cases P (mks ct1 s1) (mke ct3 se3).

Inductive SmallStepTransitiveClosure2' (P:program) : State -> State -> Prop :=
 | SmallStepTransitiveClosure2'_inv_case1 : forall ct s1 s2,
     IntraStarStep P ct s1 s2 ->
     SmallStepTransitiveClosure2' P (mks ct s1) (mks ct s2)
 | SmallStepTransitiveClosure2'_inv_case2 : forall st ct1 s1 s2 si s0,
     SmallStepTransitiveClosure2' P st (mks ct1 s1) ->
     CallStep P ct1 s1 si ->
     init_intra ct1 s1 si s0 ->
     IntraStarStep P (init_context ct1 s1 si) s0 s2 ->
     SmallStepTransitiveClosure2' P st (mks (init_context ct1 s1 si) s2)
 | SmallStepTransitiveClosure2'_inv_case3 : forall ct1 s1 ct2 s2 se2 ct3 se3,
    SmallStepTransitiveClosure2' P (mks ct1 s1) (mks ct2 s2) ->
    IntraSmallStepException P ct2 s2 se2 ->
    clos_refl_trans _ (uncaught P) (ct2,se2) (ct3,se3) ->
    prefix ct3 ct1 -> 
    SmallStepTransitiveClosure2' P (mks ct1 s1) (mke ct3 se3).

Lemma SmallStepTransitiveClosure2_inv :  
  forall P st1 st2, 
  SmallStepTransitiveClosure2 P st1 st2 ->
  SmallStepTransitiveClosure2_inv_cases P st1 st2.
Proof.
  induction 1; simpl in *; intros; auto.
  constructor 1; auto.
  inversion_clear IHSmallStepTransitiveClosure2.
  constructor 2 with s0; eauto.
  constructor 2 with s6; eauto.
  constructor 3 with (2:=H0); auto.
Qed.

Hint Resolve call_and_return_exception_equiv1 call_and_return_exception_equiv2
             call_and_return_equiv1 call_and_return_equiv2.

Lemma ExceptionStep_star_prefix : forall P st1 st2,
  clos_refl_trans _ (uncaught P) st1 st2 ->
  prefix (fst st1) (fst st2).
Proof.
  induction 1; intros; subst.
  inversion_clear H.
  constructor 1; auto.
  constructor 2.
  constructor 3 with (fst y); auto.
Qed.

Lemma IntraStarStep_inv3 : forall P ct s1 s2,
      IntraStarStep P ct s1 s2 ->
      forall s3 se,
      IntraExceptionStep2 P ct s2 se ->
      CaughtLocalException P ct se s3 ->
      IntraStarStep P ct s1 s3.
Proof.
  induction 1; intros; eauto.
Qed.

Lemma clos_refl_trans_inv1 : forall (A:Type) (R:A->A->Prop) x y,
  clos_refl_trans A R x y ->
  x=y \/ exists z, clos_refl_trans A R x z /\ R z y.
Proof.
  induction 1; intros.
  right; exists x; split; auto.
  auto.
  destruct IHclos_refl_trans2.
  subst; auto.
  destruct H1.
   right; exists x0.
   destruct H1; split; auto.
   constructor 3 with y; auto.
Qed.

Lemma ExceptionStep_star_eq : forall P st1 st2,
  clos_refl_trans _ (uncaught P) st1 st2 ->
  (fst st1)=(fst st2) -> st1=st2.
Proof.
  induction 1; intros; subst.
  inversion_clear H in H0; simpl in *; subst.
  elim return_context_no_cycle2 with ct2 ct2; auto.
  auto.
  elim clos_refl_trans_inv1 with (1:=H0); intros.
  subst; auto.
  destruct H2 as [t [HH1 HH2]].
  inversion HH2; subst; simpl in *.
  elim return_context_no_cycle2 with ct1 ct2; auto.
  generalize (ExceptionStep_star_prefix _ _ _ HH1); simpl; intros.
  generalize (ExceptionStep_star_prefix _ _ _ H); rewrite H1; intros.
  eauto.
Qed.

Lemma clos_refl_trans_inv2 : forall (A:Type) (R:A->A->Prop) x y,
  clos_refl_trans A R x y ->
  x=y \/ exists z, R x z /\ clos_refl_trans A R  z y.
Proof.
  induction 1; intros.
  right; exists y; split; auto.
  auto.
  destruct IHclos_refl_trans1.
  subst; auto.
  destruct H1.
   right; exists x0.
   destruct H1; split; auto.
   constructor 3 with y; auto.
Qed.

Lemma SmallStepTransitiveClosure2_increase_sf :
  forall P st1 st2,
  SmallStepTransitiveClosure2 P st1 st2 ->
  prefix (context st2) (context st1).
Proof.
  induction 1; simpl; auto.
  simpl in *.
  constructor 3 with (init_context ct s2 si); auto.
  constructor 1.
  apply return_context_init_context_prop2.
Qed.

Lemma SmallStepTransitiveClosure2'_SmallStepTransitiveClosure2 :
  forall P st1 st2,
  SmallStepTransitiveClosure2' P st1 st2 ->
  SmallStepTransitiveClosure2 P st1 st2.
Proof.
  induction 1; eauto.
  apply SmallStepTransitiveClosure2_trans with (1:=IHSmallStepTransitiveClosure2').
  eauto.
Qed.

Hint Constructors SmallStepTransitiveClosure2'.

Lemma SmallStepTransitiveClosure2'_IntraNormalStep2_clos : forall P st2 st1,
   SmallStepTransitiveClosure2' P st1 st2 -> forall s2 s3,
   intra st2 s2 ->
   IntraNormalStep2 P (context st2) s2 s3 -> 
   SmallStepTransitiveClosure2' P st1 (mks (context st2) s3).
Proof.
  induction 1; simpl; intros.
  inversion_clear H0 in H1 H; eauto.
  constructor 1.
  eapply IntraStarStep_IntraNormalStep2_right_clos; eauto.
  inversion H3; subst.
  constructor 2 with (3:=H1); auto.
  eapply IntraStarStep_IntraNormalStep2_right_clos; eauto.
  inversion H3.
Qed.  

Lemma SmallStepTransitiveClosure2'_IntraExceptionStep2_clos : forall P st2 st1,
   SmallStepTransitiveClosure2' P st1 st2 -> forall s2 se s3,
   intra st2 s2 ->
   IntraExceptionStep2 P (context st2) s2 se ->
   CaughtLocalException P (context st2) se s3 -> 
   SmallStepTransitiveClosure2' P st1 (mks (context st2) s3).
Proof.
  induction 1; intros.
  inversion_clear H0 in H H1 H2; simpl in *; eauto.
  constructor 1.
  eapply IntraStarStep_IntraExceptionStep2_right_clos; eauto.
  inversion H3; subst; clear H3; simpl.
  constructor 2 with (3:=H1); auto.
  eapply IntraStarStep_IntraExceptionStep2_right_clos; eauto.
  inversion H3.
Qed.  

Lemma SmallStepTransitiveClosure2'_IntraStarStep_clos : forall P ct st1 s2 s3,
   IntraStarStep P ct s2 s3 -> 
   SmallStepTransitiveClosure2' P st1 (mks ct s2) ->
   SmallStepTransitiveClosure2' P st1 (mks ct s3).
Proof.
  induction 1; intros; auto.
  apply IHIntraStarStep.
  apply SmallStepTransitiveClosure2'_IntraNormalStep2_clos
   with (st2:=(mks ct s1)) (s2:=s1); auto. 
  apply IHIntraStarStep.
  eapply SmallStepTransitiveClosure2'_IntraExceptionStep2_clos
   with (st2:=(mks ct s1)) (s2:=s1); eauto. 
Qed.

Lemma SmallStepTransitiveClosure2'_increase_sf :
  forall P st1 st2,
  SmallStepTransitiveClosure2' P st1 st2 ->
  prefix (context st2) (context st1).
Proof.
  induction 1; simpl; auto.
  simpl in *.
  constructor 3 with ct1; auto.
  constructor 1.
  apply return_context_init_context_prop2.
Qed.

Lemma SmallStepTransitiveClosure2'_not_mke : forall P st1 st2,
  SmallStepTransitiveClosure2' P st1 st2 ->
  forall ct s, ~ st1=(mke ct s).
Proof.
  induction 1; red; intros; try discriminate.
  elim IHSmallStepTransitiveClosure2' with ct s; auto.
Qed.


Lemma SmallStepTransitiveClosure2'_trans : forall P st1 st2 st3,
  SmallStepTransitiveClosure2' P st2 st3 ->
  SmallStepTransitiveClosure2' P st1 st2 ->
  SmallStepTransitiveClosure2' P st1 st3.
Proof.
  intros.
  induction H.
  apply (SmallStepTransitiveClosure2'_IntraStarStep_clos P) with (2:=H0); auto.
  constructor 2 with (3:=H2); eauto.
  destruct st1.
  constructor 3 with (2:=H1); eauto.
  constructor 3 with ct1; auto.
  apply SmallStepTransitiveClosure2'_increase_sf with (1:=H0).
  elim SmallStepTransitiveClosure2'_not_mke with (1:=H0) (ct:=c) (s:=i); auto.
Qed.


Lemma SmallStepTransitiveClosure2_SmallStepTransitiveClosure2' :
  forall P st1 st2,
  SmallStepTransitiveClosure2 P st1 st2 ->
  SmallStepTransitiveClosure2' P st1 st2.
Proof.
  induction 1.
  constructor 1; auto.
  apply SmallStepTransitiveClosure2'_trans with (1:=IHSmallStepTransitiveClosure2).
  constructor 2 with (3:=H1); auto.
  constructor 3 with (2:=H0); auto.
Qed.

Lemma IntraStarStep_trans : forall P ct s1 s2 s3,
  IntraStarStep P ct s1 s2 ->
  IntraStarStep P ct s2 s3 -> IntraStarStep P ct s1 s3.
Proof.
  induction 1; intros; eauto.
Qed.


Lemma SmallStepTransitiveClosure2_caught :
  forall P st1 st2,
  SmallStepTransitiveClosure2' P st1 st2 ->
  forall s2 s2' se2 ct3 se3 s3, intra st2 s2 ->
  IntraStarStep P (context st2) s2 s2' ->
  IntraExceptionStep2 P (context st2) s2' se2 ->
  clos_refl_trans _ (uncaught P) ((context st2),se2) (ct3,se3) ->
  CaughtLocalException P ct3 se3 s3 ->
  prefix ct3 (context st1) -> 
  SmallStepTransitiveClosure2 P st1 (mks ct3 s3).
Proof.
  induction 1; simpl; intros.
  elim clos_refl_trans_inv1 with (1:=H3); intros.
  inversion H6; subst.
  inversion H0; subst.
  constructor 1.
  apply IntraStarStep_inv3 with s2' se3; auto.
  apply IntraStarStep_trans with (1:=H); auto.  
  destruct H6 as [[ct4 se4] [U1 U2]].
  elim return_context_no_cycle3 with ct4 ct3.
  inversion U2; auto.
  constructor 3 with (1:=H5).
  apply ExceptionStep_star_prefix with (1:=U1).
  assert (HH:=SmallStepTransitiveClosure2'_SmallStepTransitiveClosure2 _ _ _ H).
  inversion H3; subst.
  elim clos_refl_trans_inv2 with (1:=H6); intros.
  apply SmallStepTransitiveClosure2_trans with (1:=HH).
  inversion H9; subst; clear H9.
  constructor 2 with (3:=H1); auto.
  constructor 1.
  apply IntraStarStep_inv3 with s2' se3; auto.
  apply IntraStarStep_trans with (1:=H2); auto.  
  destruct H9 as [[ct4 se4] [U1 U2]].
  inversion U1; subst.
  assert (IntraExceptionStep2 P ct1 s1 se4).
  constructor 2 with si sr s0; auto.
  constructor 2 with se2 s2'; auto.
  apply IntraStarStep_trans with (1:=H2); auto.  
  eapply IHSmallStepTransitiveClosure2'; eauto.
  simpl; replace ct1 with ct4; auto.
  apply return_context_functionnal with (init_context ct1 s1 si); auto.
  apply return_context_init_context_prop2.
  inversion H3.
Qed.

Lemma SmallStepTransitiveClosure2_SmallStep_right_clos : forall P st1 st2 st3,
  SmallStep P st2 st3 -> 
  prefix (context st3) (context st1) ->
  SmallStepTransitiveClosure2 P st1 st2 ->
  SmallStepTransitiveClosure2 P st1 st3.
Proof.
  intros P st1 st2 st3 HH; intros.
  destruct st2 as [ct2 s2|]; try (inversion_clear H0;fail).
  generalize (SmallStep_implies_SmallStep_case HH); clear HH; intros.
  inversion_clear H1 in H0 H.
  apply SmallStepTransitiveClosure2_trans with (1:=H0); eauto.
  apply SmallStepTransitiveClosure2_trans with (1:=H0); eauto.
  simpl in *.
  generalize (SmallStepTransitiveClosure2_inv _ _ _ H0); intros.
  inversion_clear H1 in H H2 H0 H H3 H4; simpl in *.
  elim (return_context_no_cycle3 _ _ H3); auto.
  destruct st1 as [ct s|]; try (inversion H5; fail).
  generalize (return_context_init_context_prop1 _ _ _ _ H3); intros; subst.
  apply SmallStepTransitiveClosure2_trans with (1:=H5).
  constructor 1.
  constructor 2 with s'; eauto.
  apply SmallStepTransitiveClosure2_trans with (1:=H0).
  eauto.
  generalize (SmallStep_implies_SmallStep_case HH); clear HH; intros.
  inversion_clear H1 in H0 H.
  simpl in *.
  inversion_clear H0 in H2 H; simpl in *.
  clear H5.
  assert (HH1:=SmallStepTransitiveClosure2_SmallStepTransitiveClosure2' _ _ _ H1).
  apply SmallStepTransitiveClosure2_caught with (1:=HH1)
   (s2:=s2) (s2':=s2) (se2:=se2) (se3:=i); auto.
  inversion_clear H0 in H; simpl in *.
  constructor 3 with (2:=H5); auto.
  constructor 3 with (1:=H6).
  constructor 1; constructor 1 with sr; auto.
Qed.


Theorem equiv_SmallStepTransitiveClosure_SmallStepTransitiveClosure2_aux : forall P st1 st2,
  SmallStepTransitiveClosure P st1 st2 -> 
  (forall ct se, st1 <> mke ct se) ->
  SmallStepTransitiveClosure2 P st1 st2.
Proof.
  induction 1; intros. 
  destruct st; eauto.
  elim (H c i); auto.
  eapply SmallStepTransitiveClosure2_SmallStep_right_clos; eauto.
Qed.

Lemma SmallStepTransitiveClosure2_implies_IntraStarStep : forall P ct s1 s2,
  SmallStepTransitiveClosure2 P (mks ct s1) (mks ct s2) ->
  IntraStarStep P ct s1 s2.
Proof.
  intros.
  inversion_clear H; auto.
  generalize (SmallStepTransitiveClosure2_increase_sf _ _ _ H3); simpl; intros.
  elim (return_context_no_cycle3 (init_context ct s3 si) ct); auto.
  apply return_context_init_context_prop2.
Qed.

Lemma SmallStepTransitiveClosure_increase_sf :
  forall P st1 st2,
  SmallStepTransitiveClosure P st1 st2 ->
  prefix (context st2) (context st1).
Proof.
  induction 1; eauto.
Qed.

Lemma SmallStepTransitiveClosure_trans : forall P st1 st2 st3,
  SmallStepTransitiveClosure P st2 st3 ->
  SmallStepTransitiveClosure P st1 st2 ->
  SmallStepTransitiveClosure P st1 st3.
Proof.
  intros.
  induction H; auto.
  generalize (SmallStepTransitiveClosure_increase_sf _ _ _ H0); intros.
  constructor 2 with st2; auto.
  simpl in *.
  constructor 3 with (context st0); auto.
Qed.

Lemma SmallStep_SmallStepTransitiveClosure : forall P s1 s2,
  context s1 = context s2 -> 
  SmallStep P s1 s2 -> 
  SmallStepTransitiveClosure P s1 s2.
Proof.
  intros.
  constructor 2 with s1; auto.
  rewrite H; auto.
Qed.  

Lemma SmallStepTransitiveClosure_SmallStep_left_clos: forall P st1 st2 st3,
    SmallStepTransitiveClosure P st2 st3 -> 
    SmallStep P st1 st2 -> 
    prefix (context st2) (context st1) ->
    SmallStepTransitiveClosure P st1 st3.
Proof.
  induction 1; intros.
  eauto.
  constructor 2 with st2; eauto.
Qed.

Lemma uncaught_star_SmallStepTransitiveClosure : forall P st1 st2,
  clos_refl_trans _ (uncaught P) st1 st2 ->
  forall st,
  prefix (fst st2) (context st) ->
  SmallStepTransitiveClosure P st (mke (fst st1) (snd st1)) ->
  SmallStepTransitiveClosure P st (mke (fst st2) (snd st2)).
Proof.
  induction 1; simpl; intros; auto.
  destruct x as [ct2 se2]; destruct y as [ct3 se3]; simpl in *.
  constructor 2 with (mke ct2 se2); auto.
  inversion_clear H. 
  apply SmallStep_case_implies_SmallStep.
  constructor 6 with sr; auto.
  apply IHclos_refl_trans2; auto.
  apply IHclos_refl_trans1; auto.
  constructor 3 with (fst z); auto.
  apply ExceptionStep_star_prefix with P; auto.
Qed.

Lemma IntraStarStep_SmallStepTransitiveClosure : forall P ct s1 s2,
   IntraStarStep P ct s1 s2 -> 
   SmallStepTransitiveClosure P (mks ct s1) (mks ct s2).
Proof.
  intros.
  apply (IntraStarStep_ind_strong P
      (fun ct s1 ret => exists s2, SmallStepTransitiveClosure P (mks ct s1) (mks ct s2)
          /\ ((ReturnSmallStep P ct s2 ret \/
             (exists ct1, exists s, SmallStepTransitiveClosure P (mks ct s1) (mks ct1 s)
              /\ exists se1, exists se, IntraSmallStepException P ct1 s se1
              /\ clos_refl_trans _ (uncaught P) (ct1,se1) (ct,se)
              /\ UncaughtLocalException P ct se ret))))
      (fun ct s1 s2 => SmallStepTransitiveClosure P (mks ct s1) (mks ct s2))
      (fun ct s1 se => (IntraSmallStepException P ct s1 se
         \/ (exists ct1, exists s, SmallStepTransitiveClosure P (mks ct s1) (mks ct1 s)
             /\ exists se1, IntraSmallStepException P ct1 s se1
                   /\ clos_refl_trans _ (uncaught P) (ct1,se1) (ct,se))))
      (fun ct s1 s2 => SmallStepTransitiveClosure P (mks ct s1) (mks ct s2)));
  intros; eauto.

  exists s3; split; auto. 

  destruct H3 as [H3|H3].

  right; exists ct0; exists s3; split; auto.
  exists se; exists se; split; auto.
  destruct H3 as [ct1 [s [H3 [se1 [H5 H6]]]]].
  right; exists ct1; exists s; split. 
  apply SmallStepTransitiveClosure_trans with (1:= H3); auto.
  eauto.

  apply SmallStepTransitiveClosure_trans with (2:= H1); auto.

  apply SmallStepTransitiveClosure_trans with (1:=H4).
  destruct H1 as [H1|H1].
  constructor 2 with (mke ct0 se); auto.
  eapply SmallStep_SmallStepTransitiveClosure; eauto.
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
  apply SmallStep_case_implies_SmallStep.
  constructor 5; auto.
  destruct H1 as [ct' [s [H5 [se1 [H6 H7]]]]].
  constructor 2 with (mke ct0 se); auto.
  apply uncaught_star_SmallStepTransitiveClosure with (1:=H7); simpl; auto.
  constructor 2 with (mks ct' s); simpl; auto.
  apply SmallStepTransitiveClosure_increase_sf with (1:=H5).
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
  
  right.  
  destruct H3 as [s3 [H6 [H5|H5]]].
    
  destruct (call_and_return_exception_equiv2 _ _ _ _ _ _ H1).
  elim (ReturnSmallStep_not_return_exception _ _ _ _ _ H5 H4).
  destruct H5 as [ct1 [s0' [H6' [se1 [se2 [H7 [H8 H9]]]]]]].
  exists ct1; exists s0'; split.  
  apply SmallStepTransitiveClosure_trans with (1:=H6').
  constructor 2 with (mks ct0 s); simpl; auto.
  constructor 1; apply return_context_init_context_prop2.
  apply SmallStep_case_implies_SmallStep.
  constructor 2; auto.
  destruct (call_and_return_exception_equiv2 _ _ _ _ _ _ H1); auto.
  exists se1; repeat split; auto.
  destruct (call_and_return_exception_equiv2 _ _ _ _ _ _ H1); intros.
  constructor 3 with (1:=H8).
  constructor 1.
  constructor 1 with sr; auto.
  apply return_context_init_context_prop2.

  eapply SmallStep_SmallStepTransitiveClosure; eauto.
  apply SmallStep_case_implies_SmallStep.
  constructor 1; auto.

  destruct H3 as [s3 [H4' [H5|H5]]].  
  constructor 2 with (mks (init_context ct0 s si) s3); auto.
  apply SmallStepTransitiveClosure_SmallStep_left_clos 
    with (mks (init_context ct0 s si) s0); auto.
  apply SmallStep_case_implies_SmallStep.
  constructor 2; auto.
  elim (call_and_return_equiv2 _ _ _ _ _ _ H1); auto.
  simpl; constructor 1; apply return_context_init_context_prop2.
  apply SmallStep_case_implies_SmallStep.
  constructor 3 with sr; auto.
  apply return_context_init_context_prop2.
  elim (call_and_return_equiv2 _ _ _ _ _ _ H1); auto.

  destruct H5 as [ct1 [s0' [H6 [se1 [se2 [H7 [H8 H9]]]]]]].
  destruct (call_and_return_equiv2 _ _ _ _ _ _ H1).
  elim (return_intra_not_UncaughtLocalException _ _ _ _ _ H9 H4).
Qed.

Lemma IntraExceptionStep2_SmallStepTransitiveClosure : forall P ct s1 se,
   IntraExceptionStep2 P ct s1 se -> 
   (IntraSmallStepException P ct s1 se
         \/ (exists ct1, exists s, SmallStepTransitiveClosure P (mks ct s1) (mks ct1 s)
             /\ exists se1, IntraSmallStepException P ct1 s se1
                   /\ clos_refl_trans _ (uncaught P) (ct1,se1) (ct,se))).
Proof.
  intros.
  apply (IntraExceptionStep2_ind_strong P
      (fun ct s1 ret => exists s2, SmallStepTransitiveClosure P (mks ct s1) (mks ct s2)
          /\ ((ReturnSmallStep P ct s2 ret \/
             (exists ct1, exists s, SmallStepTransitiveClosure P (mks ct s1) (mks ct1 s)
              /\ exists se1, exists se, IntraSmallStepException P ct1 s se1
              /\ clos_refl_trans _ (uncaught P) (ct1,se1) (ct,se)
              /\ UncaughtLocalException P ct se ret))))
      (fun ct s1 s2 => SmallStepTransitiveClosure P (mks ct s1) (mks ct s2))
      (fun ct s1 se => (IntraSmallStepException P ct s1 se
         \/ (exists ct1, exists s, SmallStepTransitiveClosure P (mks ct s1) (mks ct1 s)
             /\ exists se1, IntraSmallStepException P ct1 s se1
                   /\ clos_refl_trans _ (uncaught P) (ct1,se1) (ct,se))))
      (fun ct s1 s2 => SmallStepTransitiveClosure P (mks ct s1) (mks ct s2)));
  intros; eauto.

  exists s2; split; auto. 

  destruct H3 as [H3|H3].

  right; exists ct0; exists s2; split; auto.
  exists se0; exists se0; split; auto.
  destruct H3 as [ct1 [s [H3 [se1 [H5 H6]]]]].
  right; exists ct1; exists s; split. 
  apply SmallStepTransitiveClosure_trans with (1:= H3); auto.
  eauto.

  apply SmallStepTransitiveClosure_trans with (2:= H1); auto.

  apply SmallStepTransitiveClosure_trans with (1:=H4).
  destruct H1 as [H1|H1].
  constructor 2 with (mke ct0 se0); auto.
  eapply SmallStep_SmallStepTransitiveClosure; eauto.
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
  apply SmallStep_case_implies_SmallStep.
  constructor 5; auto.
  destruct H1 as [ct' [s [H5 [se1 [H6 H7]]]]].
  constructor 2 with (mke ct0 se0); auto.
  apply uncaught_star_SmallStepTransitiveClosure with (1:=H7); simpl; auto.
  constructor 2 with (mks ct' s); simpl; auto.
  apply SmallStepTransitiveClosure_increase_sf with (1:=H5).
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
  
  right.  
  destruct H3 as [s3 [H6 [H5|H5]]].
    
  destruct (call_and_return_exception_equiv2 _ _ _ _ _ _ H1).
  elim (ReturnSmallStep_not_return_exception _ _ _ _ _ H5 H4).
  destruct H5 as [ct1 [s0' [H6' [se1 [se2 [H7 [H8 H9]]]]]]].
  exists ct1; exists s0'; split.  
  apply SmallStepTransitiveClosure_trans with (1:=H6').
  constructor 2 with (mks ct0 s); simpl; auto.
  constructor 1; apply return_context_init_context_prop2.
  apply SmallStep_case_implies_SmallStep.
  constructor 2; auto.
  destruct (call_and_return_exception_equiv2 _ _ _ _ _ _ H1); auto.
  exists se1; repeat split; auto.
  destruct (call_and_return_exception_equiv2 _ _ _ _ _ _ H1); intros.
  constructor 3 with (1:=H8).
  constructor 1.
  constructor 1 with sr; auto.
  apply return_context_init_context_prop2.

  eapply SmallStep_SmallStepTransitiveClosure; eauto.
  apply SmallStep_case_implies_SmallStep.
  constructor 1; auto.

  destruct H3 as [s3 [H4' [H5|H5]]].  
  constructor 2 with (mks (init_context ct0 s si) s3); auto.
  apply SmallStepTransitiveClosure_SmallStep_left_clos 
    with (mks (init_context ct0 s si) s0); auto.
  apply SmallStep_case_implies_SmallStep.
  constructor 2; auto.
  elim (call_and_return_equiv2 _ _ _ _ _ _ H1); auto.
  simpl; constructor 1; apply return_context_init_context_prop2.
  apply SmallStep_case_implies_SmallStep.
  constructor 3 with sr; auto.
  apply return_context_init_context_prop2.
  elim (call_and_return_equiv2 _ _ _ _ _ _ H1); auto.

  destruct H5 as [ct1 [s0' [H6 [se1 [se2 [H7 [H8 H9]]]]]]].
  destruct (call_and_return_equiv2 _ _ _ _ _ _ H1).
  elim (return_intra_not_UncaughtLocalException _ _ _ _ _ H9 H4).
Qed.


Lemma equiv_SmallStepTransitiveClosure2_SmallStepTransitiveClosure : forall P st1 st2,
  SmallStepTransitiveClosure2 P st1 st2 ->
  SmallStepTransitiveClosure P st1 st2.
Proof.
  induction 1. 
  apply IntraStarStep_SmallStepTransitiveClosure with (1:=H).
  apply SmallStepTransitiveClosure_trans with (1:= IHSmallStepTransitiveClosure2); auto.
  apply SmallStepTransitiveClosure_trans with (mks ct s2).
  constructor 2 with (mks ct s2); auto.
  simpl; constructor 1; apply return_context_init_context_prop2.
  apply SmallStep_case_implies_SmallStep.
  constructor 2; auto.
  apply IntraStarStep_SmallStepTransitiveClosure; auto.
  apply uncaught_star_SmallStepTransitiveClosure with (1:=H1); simpl; auto.
  constructor 2 with (1:=IHSmallStepTransitiveClosure2); simpl; auto.
  apply SmallStepTransitiveClosure_increase_sf with (1:=IHSmallStepTransitiveClosure2).
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
Qed. 

Inductive IntraBigStep_from_SmallStep (P:program) : ContextCall -> IntraState ->  ReturnState-> Prop :=
 | IntraBigStep_from_SmallStep_value : forall ct s1 s2 sr,
    SmallStepTransitiveClosure P (mks ct s1) (mks ct s2) ->
    ReturnSmallStep P ct s2 sr ->
    IntraBigStep_from_SmallStep P ct s1 sr
 | IntraBigStep_from_SmallStep_exception : forall ct s1 se2 sr,
    SmallStepTransitiveClosure P (mks ct s1) (mke ct se2) ->
    UncaughtLocalException P ct se2 sr ->
    IntraBigStep_from_SmallStep P ct s1 sr.

Lemma SmallStepTransitiveClosure_same_ct_bigstep : forall P ct s1 s2,
    SmallStepTransitiveClosure P (mks ct s1) (mks ct s2) ->
    IntraStarStep P ct s1 s2.
Proof.
  intros.
  apply SmallStepTransitiveClosure2_implies_IntraStarStep.
  apply equiv_SmallStepTransitiveClosure_SmallStepTransitiveClosure2_aux.
  assumption.
  intros; discriminate.
Qed.

Lemma SmallStepTransitiveClosure2_caught' :
  forall P st1 st2,
  SmallStepTransitiveClosure2' P st1 st2 ->
  forall s2 s2' se2 ct3 se3, intra st2 s2 ->
  IntraStarStep P (context st2) s2 s2' ->
  IntraExceptionStep2 P (context st2) s2' se2 ->
  clos_refl_trans _ (uncaught P) ((context st2),se2) (ct3,se3) ->
  prefix ct3 (context st1) -> 
  exists s, 
  SmallStepTransitiveClosure2 P st1 (mks ct3 s)
  /\ IntraExceptionStep2 P ct3 s se3.
Proof.
  induction 1; simpl; intros.
  elim clos_refl_trans_inv1 with (1:=H3); intros.
  inversion H5; subst.
  inversion H0; subst.
  exists s2'; split; auto.
  constructor 1; eapply IntraStarStep_trans; eauto.
  destruct H5 as [[ct4 se4] [U1 U2]].
  elim return_context_no_cycle3 with ct4 ct3.
  inversion U2; auto.
  constructor 3 with (1:=H4).
  apply ExceptionStep_star_prefix with (1:=U1).
  assert (HH:=SmallStepTransitiveClosure2'_SmallStepTransitiveClosure2 _ _ _ H).
  inversion H3; subst.
  elim clos_refl_trans_inv2 with (1:=H6); intros.
  inversion H8; subst; clear H8.
  exists s2'; split; auto.
  apply SmallStepTransitiveClosure2_trans with (1:=HH).  
  constructor 2 with (3:=H1); auto.
  constructor 1.
  apply IntraStarStep_trans with (1:=H2); auto.  
  destruct H8 as [[ct4 se4] [U1 U2]].
  inversion U1; subst.
  assert (IntraExceptionStep2 P ct1 s1 se4).
  constructor 2 with si sr s0; auto.
  constructor 2 with se2 s2'; auto.
  apply IntraStarStep_trans with (1:=H2); auto.  
  elim (IHSmallStepTransitiveClosure2' s1 s1 se4 ct3 se3); auto.
  intros s (T1,T2).
  eauto.
  simpl; replace ct1 with ct4; auto.
  apply return_context_functionnal with (init_context ct1 s1 si); auto.
  apply return_context_init_context_prop2.
  inversion H3.
Qed.

Lemma SmallStepTransitiveClosure_same_ct_bigstep_se : forall P ct s1 se2,
    SmallStepTransitiveClosure P (mks ct s1) (mke ct se2) ->
    exists s2, IntraStarStep P ct s1 s2 /\ IntraExceptionStep2 P ct s2 se2.
Proof.
  intros.
  assert (forall ct0 se, mks ct s1 <> mke ct0 se).
  intros; discriminate.
  generalize (equiv_SmallStepTransitiveClosure_SmallStepTransitiveClosure2_aux _ _ _ H H0); intros.
  clear H H0.
  inversion H1; subst; clear H1.
  generalize (SmallStepTransitiveClosure2_SmallStepTransitiveClosure2' _ _ _ H4); intros.
  elim (SmallStepTransitiveClosure2_caught' _ _ _ H) with s2 s2 se0 ct se2; auto.
  intros s (T1,T2).
  exists s; split; auto.
  apply SmallStepTransitiveClosure_same_ct_bigstep.
  apply equiv_SmallStepTransitiveClosure2_SmallStepTransitiveClosure; auto.
Qed.

Lemma equiv_IntraBigStep_from_SmallStep_IntraBigStep2 : forall p ct s ret,
  IntraBigStep_from_SmallStep p ct s ret -> 
  IntraBigStep2 p ct s ret.
Proof.
  intros p ct s ret H.
  inversion_clear H.
  constructor 1 with s2; auto.
  apply SmallStepTransitiveClosure2_implies_IntraStarStep.
  apply equiv_SmallStepTransitiveClosure_SmallStepTransitiveClosure2_aux; auto.
  intros ;discriminate.
  elim SmallStepTransitiveClosure_same_ct_bigstep_se with (1:=H0).
  intros s' (T1,T2).
  constructor 2 with se2 s'; auto.
Qed.

Lemma equiv_IntraBigStep2_IntraBigStep_from_SmallStep : forall p ct s ret,
  IntraBigStep2 p ct s ret -> IntraBigStep_from_SmallStep p ct s ret.
Proof.
  intros p ct s ret H.
  inversion_clear H.
  constructor 1 with s2; try assumption.
  apply IntraStarStep_SmallStepTransitiveClosure; auto.
  constructor 2 with se; try assumption.
  generalize (IntraStarStep_SmallStepTransitiveClosure _ _ _ _ H0); intros.
  apply SmallStepTransitiveClosure_trans with (2:=H).
  elim IntraExceptionStep2_SmallStepTransitiveClosure with (1:=H1); intros.
  constructor 2 with (mks ct s2); auto.
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
  destruct H3 as [ct1 [s1 [U1 [se1 [U2 U3]]]]].
  generalize (SmallStepTransitiveClosure_increase_sf _ _ _ U1); simpl; intros.
  apply uncaught_star_SmallStepTransitiveClosure with (1:=U3); simpl.
  auto.
  apply SmallStepTransitiveClosure_trans with (2:=U1).
  constructor 2 with (mks ct1 s1); auto.
  apply SmallStep_case_implies_SmallStep.
  constructor; auto.
Qed.   

(** Semantics with intermediate reachables states *)
Inductive IntraNormalStep3 (P:program) : ContextCall -> IntraState -> IntraState -> Prop :=
  | IntraNormalStep3_IntraSmallStep : forall ct s1 s2,
      IntraSmallStep P ct s1 s2 -> 
      IntraNormalStep3 P ct s1 s2
  | IntraNormalStep3_call_return_value' : forall ct s si sr s' s0,
      CallStep P ct s si ->
      call_and_return ct s si s0 sr s' ->
      IntraBigStep3 P (init_context ct s si) s0 sr ->
      IntraNormalStep3 P ct s s'

with IntraExceptionStep3 (P:program) : ContextCall -> IntraState -> IntraException -> Prop :=
  | IntraExceptionStep3_IntraSmallStepException : forall ct s1 se,
      IntraSmallStepException P ct s1 se ->
      IntraExceptionStep3 P ct s1 se
  | IntraExceptionStep3_call_return_value : forall ct s si sr se s0,
      CallStep P ct s si ->
      call_and_return_exception ct s si s0 sr se ->
      IntraBigStep3 P (init_context ct s si) s0 sr ->
      IntraExceptionStep3 P ct s se

with IntraStarStep3 (P:program) : ContextCall -> IntraState -> IntraState -> Prop :=
  | IntraStarStep_refl3 : forall ct s,
      IntraStarStep3 P ct s s
  | IntraStarStep_next3 : forall ct s1 s2 s3,
      IntraStarStep3 P ct s1 s2 ->
      IntraNormalStep3 P ct s2 s3 -> 
      IntraStarStep3 P ct s1 s3
  | IntraStarStep_exception_caught3 : forall ct s1 se s2 s3,
      IntraStarStep3 P ct s1 s2 ->
      IntraExceptionStep3 P ct s2 se ->
      CaughtLocalException P ct se s3 ->
      IntraStarStep3 P ct s1 s3

with IntraBigStep3 (P:program) : ContextCall -> IntraState -> ReturnState -> Prop :=
  | IntraBigStep3_normal : forall ct s1 s2 ret,
      IntraStarStep3 P ct s1 s2 ->
      ReturnSmallStep P ct s2 ret ->
      IntraBigStep3 P ct s1 ret
  | IntraBigStep3_exception_uncaught : forall ct s1 se s2 sr,
      IntraStarStep3 P ct s1 s2 ->
      IntraExceptionStep3 P ct s2 se ->
      UncaughtLocalException P ct se sr ->
      IntraBigStep3 P ct s1 sr.
Hint Constructors IntraBigStep3 IntraExceptionStep3 IntraNormalStep3 IntraStarStep3.

(** We need strong induction principles *)
Scheme IntraBigStep3_ind_strong := Minimality for IntraBigStep3 Sort Prop
  with IntraStarStep3_ind_strong := Minimality for IntraStarStep3 Sort Prop
  with IntraExceptionStep3_ind_strong := Minimality for IntraExceptionStep3 Sort Prop
  with IntraNormalStep3_ind_strong := Minimality for IntraNormalStep3 Sort Prop.

Lemma equiv_IntraBigStep2_IntraBigStep3 : forall P ct s ret,
  IntraBigStep2 P ct s ret -> IntraBigStep3 P ct s ret.
Proof.
  intros P ct s ret.
  apply (IntraBigStep2_ind_strong P
              (fun ct s ret => IntraBigStep3 P ct s ret)
              (fun ct s1 s2 => forall s, IntraStarStep3 P ct s s1 -> IntraStarStep3 P ct s s2)
              (fun ct s1 se => IntraExceptionStep3 P ct s1 se)
              (fun ct s1 s2 => IntraNormalStep3 P ct s1 s2)
  );intros; intuition eauto.
Qed.

Lemma equiv_IntraBigStep3_IntraBigStep2 : forall P ct s ret,
  IntraBigStep3 P ct s ret -> IntraBigStep2 P ct s ret.
Proof.
  intros P ct s ret.
  apply (IntraBigStep3_ind_strong P
              (fun ct s ret => IntraBigStep2 P ct s ret)
              (fun ct s1 s2 => forall s, IntraStarStep P ct s2 s -> IntraStarStep P ct s1 s)
              (fun ct s1 se => IntraExceptionStep2 P ct s1 se)
              (fun ct s1 s2 => IntraNormalStep2 P ct s1 s2)
  );intros; intuition eauto.
Qed.

End NewSemantics.

Module MakeEquivSmallAndBigStep 
  (SS:SMALLSTEP) (BS:BIGSTEP with Module Dom := SS.Dom).

  Import BS SS Dom Prog.   

Module IntermediateSmallStep <: ABSTRACT_SMALLSTEP. 

  Definition program := Program.
  Definition IntraState := IntraNormalState.
  Definition ReturnState := ReturnState.
  Definition InitCallState := InitCallState.
  Definition ContextCall := (Method*CallStack.t)%type.
  Definition IntraException := (PC * Heap.t * Location * LocalVar.t)%type.

  Inductive State : Type := 
    mks : ContextCall -> IntraState -> State
  | mke : ContextCall -> IntraException -> State.

  Definition IntraSmallStep : program -> ContextCall -> IntraState -> IntraState  -> Prop :=
    fun p ct s1 s2 => NormalStep p (fst ct) s1 s2.
  Definition ReturnSmallStep : program -> ContextCall -> IntraState -> ReturnState -> Prop :=
    fun p ct s se => ReturnStep p (fst ct) s se.
  Inductive IntraSmallStepException_ (p:program) : ContextCall -> IntraState -> IntraException -> Prop :=
    IntraSmallStepException_def : forall ct pc h s l h' loc',
     ExceptionStep p (fst ct) (pc,(h,s,l))(h',loc') ->
     IntraSmallStepException_ p ct (pc,(h,s,l)) (pc,h',loc',l).
  Definition IntraSmallStepException := IntraSmallStepException_.
  Definition CallStep : program -> ContextCall -> IntraState -> InitCallState -> Prop :=
    fun p ct s si =>
    CallStep p (fst ct) s si.
  Inductive CaughtLocalException_ : program -> ContextCall -> IntraException -> IntraState -> Prop :=
    | CaughtLocalException_def : forall p m sf pc h loc l pc',
       CaughtException p m (pc,h,loc) pc' ->
       CaughtLocalException_ p (m,sf) (pc,h,loc,l) (pc',(h,(Ref loc::nil),l)).
  Definition CaughtLocalException : program -> ContextCall -> IntraException -> IntraState -> Prop :=
    CaughtLocalException_.

  Inductive UncaughtLocalException_ : program -> ContextCall -> IntraException -> ReturnState -> Prop :=
    UncaughtLocalException_def : forall p h loc m sf pc l,
     UnCaughtException p m (pc,h,loc) ->
     UncaughtLocalException_ p (m,sf) (pc,h,loc,l) (h, Exception loc).
  Definition UncaughtLocalException := UncaughtLocalException_.

  Inductive SmallStep_ (p:program) : State -> State -> Prop :=
   | SmallStep1 : forall h1 m1 pc1 s1 l1 sf1 h2 m2 pc2 s2 l2 sf2,
       step p (St h1 (Fr m1 pc1 s1 l1) sf1) (St h2 (Fr m2 pc2 s2 l2) sf2) ->
       SmallStep_ p (mks (m1,sf1) (pc1,(h1,s1,l1))) (mks (m2,sf2) (pc2,(h2,s2,l2)))
   | SmallStep2 : forall h1 m1 pc1 loc1 l1 sf1 h2 m2 pc2 loc2 l2 sf2,
       step p (StE h1 (FrE m1 pc1 loc1 l1) sf1) (StE h2 (FrE m2 pc2 loc2 l2) sf2) ->
       SmallStep_ p (mke (m1,sf1) (pc1,h1,loc1,l1)) (mke (m2,sf2) (pc2,h2,loc2,l2))
   | SmallStep3 : forall h1 m1 pc1 loc1 l1 sf1 h2 m2 pc2 s2 l2 sf2,
       step p (StE h1 (FrE m1 pc1 loc1 l1) sf1) (St h2 (Fr m2 pc2 s2 l2) sf2) ->
       SmallStep_ p (mke (m1,sf1) (pc1,h1,loc1,l1)) (mks (m2,sf2) (pc2,(h2,s2,l2)))
   | SmallStep4 : forall h1 m1 pc1 s1 l1 sf1 h2 m2 pc2 loc2 l2 sf2,
       step p (St h1 (Fr m1 pc1 s1 l1) sf1) (StE h2 (FrE m2 pc2 loc2 l2) sf2) ->
       SmallStep_ p (mks (m1,sf1) (pc1,(h1,s1,l1))) (mke (m2,sf2) (pc2,h2,loc2,l2)).
  Definition SmallStep := SmallStep_.
      
  Definition init_context : ContextCall -> IntraState -> InitCallState -> ContextCall :=
   fun ct st si => 
     match ct,st,si with
       (m,sf),(pc,(h,s,l)),(m',(s',l')) => (m',(Fr m pc s' l)::sf)
     end.
    
  Inductive init_intra_ : ContextCall -> IntraState -> InitCallState -> IntraState -> Prop :=
    init_intra_def : forall m sf pc h s l m' s' l' bm',
      METHOD.body m' = Some bm' ->
      init_intra_ (m,sf) (pc,(h,s,l)) (m',(s',l')) (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l')).
  Definition init_intra := init_intra_.

  Inductive return_context_ : ContextCall -> ContextCall -> Prop :=
    return_context_def : forall m0 m pc s l sf,
      return_context_ (m0,(Fr m pc s l)::sf) (m,sf).
  Definition return_context := return_context_.

  Inductive return_intra_ : ContextCall -> ReturnState -> IntraState -> Prop :=
    return_intra_def0 : forall m sf m0 pc0 s0 l0 pc' h,
      next m0 pc0 = Some pc' ->
      return_intra_ (m,(Fr m0 pc0 s0 l0)::sf) (h, Normal None) (pc',(h,s0,l0))
  | return_intra_def1 : forall m sf m0 pc0 s0 l0 pc' h v,
      next m0 pc0 = Some pc' ->
      return_intra_ (m,(Fr m0 pc0 s0 l0)::sf) (h, Normal (Some v)) (pc',(h,v::s0,l0)).
  Definition return_intra := return_intra_.

  Inductive return_exception_ : ContextCall -> ReturnState -> IntraException -> Prop :=
    return_exception_def : forall m sf m0 pc0 s0 l0 h loc,
      return_exception_ (m,(Fr m0 pc0 s0 l0)::sf) (h, Exception loc) (pc0,h,loc,l0).
  Definition return_exception := return_exception_.

  Inductive call_and_return_ : ContextCall -> IntraState -> InitCallState -> IntraState -> ReturnState -> IntraState -> Prop:=
  | call_and_return_void : forall m sf pc h s l m' l' bm' h'' s' pc',
      next m pc = Some pc' -> 
      METHOD.body m' = Some bm' ->
      call_and_return_
                 (m,sf)
                 (pc,(h,s,l))
                 (m',(s',l') )
                 (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l'))
                 (h'', Normal None) 
                 (pc',(h'',s',l))
  | call_and_return_value : forall m sf pc h s l m' l' bm' h'' v s' pc',
      next m pc = Some pc' -> 
      METHOD.body m' = Some bm' ->
      call_and_return_
                 (m,sf)
                 (pc,(h,s,l))
                 (m',(s',l'))
                 (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l'))
                 (h'', Normal (Some v)) 
                 (pc',(h'',v::s',l)).
  Definition call_and_return := call_and_return_.


  Inductive call_and_return_exception_  : ContextCall -> IntraState -> InitCallState -> IntraState -> ReturnState -> IntraException -> Prop :=
  | call_and_return_exception_def : forall m sf pc h s l m' l' bm' h'' s' loc,
      METHOD.body m' = Some bm' ->
      call_and_return_exception_
                 (m,sf)
                 (pc,(h,s,l))
                 (m',(s',l')) 
                 (BYTECODEMETHOD.firstAddress bm',(h,OperandStack.empty,l'))
                 (h'', Exception loc) 
                 (pc,h'',loc,l).
  Definition call_and_return_exception := call_and_return_exception_.

  Lemma UncaughtLocalException_functionnal : forall p ct se ret1 ret2,
    UncaughtLocalException p ct se ret1 ->
    UncaughtLocalException p ct se ret2 -> 
    ret1 = ret2.
  Proof.
    intros.
    inversion H; inversion H0; subst.
    inversion H9; inversion H8; subst; auto.
  Qed.

  Lemma return_context_functionnal : forall ct ct1 ct2,
    return_context ct ct1 ->
    return_context ct ct2 ->
    ct1=ct2.
  Proof.
    intros.
    inversion H; inversion H0; subst.
    inversion H3; subst; auto.
  Qed.

  Lemma return_context_init_context_prop2 : forall ct s si,
    return_context (init_context ct s si) ct.
  Proof.
    intros (m,sf) (pc,((h,s),l)) (m',(s',l')); simpl.
    constructor.
  Qed.

  Lemma return_context_length : forall ct1 ct2,
    return_context ct1 ct2 -> (length (snd ct2) < length (snd ct1))%nat.
  Proof.
    intros.
    inversion H; subst; simpl.
    omega.
  Qed.

  Lemma clos_trans_return_context_length : forall ct1 ct2,
    clos_trans _ return_context ct1 ct2 -> (length (snd ct2) < length (snd ct1))%nat.
  Proof.
    induction 1.
    apply return_context_length; auto.
    omega.
  Qed.

  Lemma return_context_no_cycle : forall ct, 
    ~ clos_trans _ return_context ct ct.
  Proof.
    red; intros.  
    assert (HH:=clos_trans_return_context_length _ _ H).
    omega.
  Qed.

  Lemma ReturnSmallStep_not_return_exception : forall P ct s sr se,
   ReturnSmallStep P ct s sr -> return_exception ct sr se -> False.
  Proof.
    intros.
    inversion H; subst; inversion H0.
  Qed.

  Lemma return_intra_not_UncaughtLocalException : forall P ct se sr s,
    UncaughtLocalException P ct se sr ->
    return_intra ct sr s -> False.
  Proof.
    intros.
    inversion H; subst; inversion H0.
  Qed.

  Lemma call_and_return_equiv1 : forall ct s si s0 sr s',
      init_intra ct s si s0 -> 
      return_intra (init_context ct s si) sr s' -> 
      call_and_return ct s si s0 sr s'.
  Proof.
    intros.
    inversion H; subst; simpl in H0;  
    inversion H0; subst; constructor; auto.
  Qed.

  Lemma call_and_return_equiv2 : forall ct s si s0 sr s',
      call_and_return ct s si s0 sr s' ->
      init_intra ct s si s0 /\ return_intra (init_context ct s si) sr s'.
  Proof.
    intros.
    inversion H; subst; split; simpl; constructor; auto. 
  Qed.


  Lemma call_and_return_exception_equiv1 : forall ct s si s0 sr s',
      init_intra ct s si s0 -> 
      return_exception (init_context ct s si) sr s' -> 
      call_and_return_exception ct s si s0 sr s'.
  Proof.
    intros.
    inversion H; subst; simpl in H0.  
    inversion H0; subst; constructor; auto.
  Qed.

  Lemma call_and_return_exception_equiv2 : forall ct s si s0 sr s',
      call_and_return_exception ct s si s0 sr s' ->
      init_intra ct s si s0 /\ return_exception (init_context ct s si) sr s'.
  Proof.
    intros.
    inversion H; subst; split; simpl; constructor; auto. 
  Qed.

  Inductive SmallStep_case (P:program) : State -> State -> Prop :=
  | step_simpl : forall ct s1 s2,
	IntraSmallStep P ct s1 s2 ->
        SmallStep_case P (mks ct s1) (mks ct s2)
  | step_call : forall ct s si s0,
        CallStep P ct s si ->
        init_intra ct s si s0 ->
        SmallStep_case P (mks ct s) (mks (init_context ct s si) s0)
  | step_return : forall ct s sr ct' s' ,
	ReturnSmallStep P ct s sr ->
	return_context ct ct' ->
	return_intra ct sr s' ->
        SmallStep_case P (mks ct s) (mks ct' s')
  | step_thrown_exception : forall ct s se,
       IntraSmallStepException P ct s se ->
       SmallStep_case P (mks ct s) (mke ct se)
  | step_caught_exception : forall ct s se,
       CaughtLocalException P ct se s ->
       SmallStep_case P (mke ct se) (mks ct s)
  | step_uncaught_exception : forall ct1 se1 ct2 se2 sr,
       UncaughtLocalException P ct1 se1 sr ->
       return_context ct1 ct2 ->
       return_exception ct1 sr se2 ->
       SmallStep_case P (mke ct1 se1) (mke ct2 se2).


  Hint Constructors
     BS.NormalStep
     ExceptionStep
     JVMExceptionStep : equiv.

  Lemma SmallStep_implies_SmallStep_case : forall P st1 st2,
    SmallStep P st1 st2 ->
    SmallStep_case P st1 st2.
  Proof.
    intros.
    inversion H; clear H; subst.
    inversion_clear H0.
    constructor 1; unfold IntraSmallStep; eapply aconst_null; eauto.
    constructor 1; unfold IntraSmallStep; eapply arraylength; eauto.
    constructor 1; unfold IntraSmallStep; eapply checkcast1; eauto.
    constructor 1; unfold IntraSmallStep; eapply const; eauto.
    constructor 1; unfold IntraSmallStep; eapply dup; eauto.
    constructor 1; unfold IntraSmallStep; eapply dup_x1; eauto.
    constructor 1; unfold IntraSmallStep; eapply dup_x2; eauto.
    constructor 1; unfold IntraSmallStep; eapply dup2; eauto.
    constructor 1; unfold IntraSmallStep; eapply dup2_x1; eauto.
    constructor 1; unfold IntraSmallStep; eapply dup2_x2; eauto.
    constructor 1; unfold IntraSmallStep; eapply getfield; eauto.
    constructor 1; unfold IntraSmallStep; eapply getstatic; eauto.
    constructor 1; unfold IntraSmallStep; eapply goto; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.i2b; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.i2s; eauto.
    constructor 1; unfold IntraSmallStep; eapply ibinop; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.if_acmp_step_jump; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.if_acmp_step_continue; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.if_icmp_step_jump; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.if_icmpeq_step_continue; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.ifeq_step_jump; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.ifeq_step_continue; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.ifnull_step_jump; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.ifnull_step_continue; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.iinc_step; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.ineg_step; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.instanceof1; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.instanceof2; eauto.
    replace (@pair Prog.Method CallStack.t m2
              (@cons Frame.t (Frame.make m1 pc1 s l1) sf1))
    with (init_context (m1,sf1) (pc1,(h2,args++s,l1)) (m2,(s,l2))); auto.
      inversion H5; subst.
      constructor 2; unfold CallStep.
      simpl; eapply invokestatic; eauto.
      constructor; auto.
    constructor 1; unfold IntraSmallStep; eapply BS.invokestatic_native; eauto.
    inversion H5; subst.
      replace (@pair Prog.Method CallStack.t M
              (@cons Frame.t (Frame.make m1 pc1 s l1) sf1))
      with (init_context (m1,sf1) (pc1,(h2,args++Ref loc::s,l1)) (M,(s,stack2localvar (args ++ Ref loc :: s) (S (length args))))); auto.
      constructor 2; unfold CallStep.
      simpl.
      apply (invokevirtual P h2 m1 pc1 s l1 mid cn M args loc cl bM); auto.
      constructor; auto.
    constructor 1; unfold IntraSmallStep; eapply BS.lookupswitch1; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.lookupswitch2; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.new; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.newarray; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.nop; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.pop; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.pop2; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.putfield; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.putstatic; eauto.
    constructor 3 with (h2, Normal None).
      unfold ReturnSmallStep.
      eapply void_return; eauto.
      constructor.
      constructor; auto.
    constructor 1; unfold IntraSmallStep; eapply BS.swap; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.tableswitch1; eauto.
    constructor 1; unfold IntraSmallStep; eapply BS.tableswitch2; eauto.
    constructor 1; unfold IntraSmallStep; eapply vaload; eauto.
    constructor 1; unfold IntraSmallStep; eapply vastore; eauto.
    constructor 1; unfold IntraSmallStep; eapply vload; eauto.
    constructor 3 with (h2, Normal (Some val)).
      unfold ReturnSmallStep; eapply vreturn; eauto.
      constructor.
      constructor; auto.
    constructor 1; unfold IntraSmallStep; eapply vstore; eauto.
    inversion_clear H0.
    constructor 6 with (h2, Exception loc2).
    constructor; auto.
    constructor.
    constructor.
    inversion_clear H0.
    repeat constructor; auto.
    inversion_clear H0;
    try (constructor;
         try constructor;
         simpl; eauto with equiv).
    eapply athrow; eauto.
  Qed.

  Hint Constructors step : equiv.

  Lemma SmallStep_case_implies_SmallStep : forall P st1 st2,
    SmallStep_case P st1 st2 ->
    SmallStep P st1 st2.
  Proof.
    intros.
    inversion H; subst; clear H.
    destruct ct; inversion H0; subst; 
    constructor; eauto with equiv.
    eapply SS.arraylength_step_ok; eauto.
    eapply SS.const_step_ok; eauto.
    eapply SS.getfield_step_ok; eauto.
    eapply SS.i2b_step_ok; eauto.
    eapply SS.i2s_step_ok; eauto.
    eapply SS.ibinop_step_ok; eauto.
    eapply SS.if_icmp_step_jump; eauto.
    eapply SS.if_icmpeq_step_continue; eauto.
    eapply SS.ifeq_step_jump; eauto.
    eapply SS.ifeq_step_continue; eauto.
    eapply SS.ineg_step; eauto.
    eapply SS.instanceof_step_ok1; eauto.
    eapply SS.lookupswitch_step_ok1; eauto.
    eapply SS.lookupswitch_step_ok2; eauto.
    eapply SS.new_step_ok; eauto.
    eapply SS.newarray_step_ok; eauto.
    eapply SS.putfield_step_ok; eauto.
    eapply SS.tableswitch_step_ok1; eauto.
    eapply SS.tableswitch_step_ok2; eauto.
    eapply SS.vaload_step_ok; eauto.
    eapply SS.vastore_step_ok; eauto.
    inversion H1; subst;
    inversion H0; subst; clear H1 H0; simpl; constructor;
    eauto with equiv.
    inversion H1; subst.
    inversion H2; subst; clear H1 H2; inversion H0; constructor; eauto with equiv.
    destruct ct; inversion H0; subst.
    clear H0; inversion H; clear H; subst; simpl; constructor.
    eapply SS.athrow_step; eauto.
    eapply invokestatic_step_native_exception; eauto.
    inversion H3; subst; eauto with equiv.
    eapply SS.checkcast_step_ClassCastException; eauto.
    eapply SS.ibinop_ArithmeticException; eauto.
    eapply SS.newarray_step_NegativeArraySizeException; eauto.
    eapply SS.vaload_step_NullPointerException; eauto.
    eapply SS.vaload_step_ArrayIndexOutOfBoundsException; eauto.
    eapply SS.vastore_step_NullPointerException; eauto.
    eapply SS.vastore_step_ArrayIndexOutOfBoundsException; eauto.
    eapply SS.vastore_step_ArrayStoreException; eauto.
    inversion H0; subst; clear H0; constructor.
    constructor 1; auto.
    inversion H2; subst; inversion H1; subst;
    inversion H0; subst; constructor; clear H0 H1 H2.
    constructor 2; auto.
  Qed.

End IntermediateSmallStep.

Module N := NewSemantics IntermediateSmallStep.

Hint Constructors
  N.IntraBigStep3 N.IntraExceptionStep3
  N.IntraNormalStep3 N.IntraStarStep3 : equiv.

  Module V2.

   Inductive IntraNormalStep (P:Program) : 
      Method -> IntraNormalState -> IntraNormalState -> Prop :=
     | IntraNormalStep_IntraStep : forall M s1 s2,
        NormalStep P M s1 s2 -> 
        IntraNormalStep P M s1 s2
     | IntraNormalStep1_call_return_value : forall M s si sr s' s0,
        CallStep P M s si ->
        call_and_return M s si s0 sr s' ->
        IntraBigStep P (fst si) s0 sr ->
        IntraNormalStep P M s s'

   with IntraExceptionStep (P:Program) :
         Method -> IntraNormalState -> IntraExceptionState -> Prop :=
     | IntraExceptionStep_IntraExceptionStep : forall M s1 se,
         ExceptionStep P M s1 se ->
         IntraExceptionStep P M s1 se
     | IntraExceptionStep_call_return_value : forall M s si sr se s0,
         CallStep P M s si ->
         call_and_return_exception M s si s0 sr se ->
         IntraBigStep P (fst si) s0 sr ->
         IntraExceptionStep P M s se

   with IntraStarStep (P:Program) :
         Method -> IntraNormalState -> IntraNormalState -> Prop :=
     | IntraStarStep_refl : forall M s,
         IntraStarStep P M s s
     | IntraStarStep_next : forall M s1 s2 s3,
         IntraStarStep P M s1 s2 ->
         IntraNormalStep P M s2 s3 -> 
         IntraStarStep P M s1 s3
     | IntraStarStep_exception_caught : forall M s1 h2 s2 pc pc' h loc l,
         IntraStarStep P M s1 (pc,(h2,s2,l))  ->
         IntraExceptionStep P M (pc,(h2,s2,l)) (h,loc) ->
         CaughtException P M (pc,h,loc) pc' ->
         IntraStarStep P M s1 (pc',(h,Ref loc::OperandStack.empty,l))

   with IntraBigStep (P:Program) :
         Method -> IntraNormalState -> ReturnState -> Prop :=
     | IntraBigStep_normal : forall M s1 s2 ret,
         IntraStarStep P M s1 s2 ->
         ReturnStep P M s2 ret ->
         IntraBigStep P M s1 ret
     | IntraBigStep_exception_uncaught : forall M s1 pc h loc l s2 h2,
         IntraStarStep P M s1 (pc,(h2,s2,l)) ->
         IntraExceptionStep P M (pc,(h2,s2,l)) (h,loc) ->
         UnCaughtException P M (pc,h,loc) ->
         IntraBigStep P M s1 (h, Exception loc).

   (** stronger induction principles *)
   Scheme IntraBigStep_ind_strong := Minimality for IntraBigStep Sort Prop
     with IntraStarStep_ind_strong := Minimality for IntraStarStep Sort Prop
     with IntraExceptionStep_ind_strong := Minimality for IntraExceptionStep Sort Prop
     with IntraNormalStep_ind_strong := Minimality for IntraNormalStep Sort Prop.

  End V2.

Lemma Bigstep_old_and_new : forall p m si sr,
  V2.IntraBigStep p m si sr ->
  forall sf, N.IntraBigStep3 p (m,sf) si sr.
Proof.
  intros p m si sr.
  apply (V2.IntraBigStep_ind_strong p
           (fun m s ret => forall sf, N.IntraBigStep3 p (m,sf) s ret)
           (fun m s1 s2 => forall sf, N.IntraStarStep3 p (m,sf) s1 s2)
           (fun m s1 se => forall sf, N.IntraExceptionStep3 p (m,sf) s1 (fst s1,fst se,snd se, snd (snd s1)))
           (fun m s1 s2 => forall sf, N.IntraNormalStep3 p (m,sf) s1 s2) 
  ); intros; simpl in *.
  eauto with equiv.
  constructor 2 with (pc, h, loc, l) (pc, (h2, s2, l)); auto.
  constructor; auto.
  constructor.
  eauto with equiv.
  constructor 3 with (pc, h, loc, l) (pc, (h2, s2, l)); auto.
  constructor; auto.
  constructor; destruct s1 as [pc1 [[h1 s1] l1]].
  destruct se as [he loc]; simpl.
  constructor; auto.
  constructor 2 with si0 sr0 s0; auto.
  inversion H0; subst.
  eapply IntermediateSmallStep.call_and_return_exception_def; eauto.
  destruct s as [pc [[h s] l]].
  destruct si0 as [m' [s' l']].
  simpl in *.
  auto.
  eauto with equiv.
  constructor 2 with si0 sr0 s0; auto.
  inversion H0; subst.
  eapply IntermediateSmallStep.call_and_return_void; eauto.
  eapply IntermediateSmallStep.call_and_return_value; eauto.
  destruct s as [pc [[h s]l]].
  destruct si0 as [m' [s'' l']].
  simpl in *.
  auto.
Qed.

Lemma IntraExceptionStep3_same_pc_l : forall p ct pc2 h2 s2 l2 pc h loc l,
  N.IntraExceptionStep3 p ct (pc2, (h2, s2, l2)) (pc, h, loc, l) ->
  pc = pc2 /\ l = l2.
Proof.
  intros. 
  inversion_clear H.
  inversion_clear H0; auto.
  inversion_clear H1; auto.
Qed.


Lemma Bigstep_old_and_new_inv : forall p ct si sr,
  N.IntraBigStep3 p ct si sr ->
  V2.IntraBigStep p (fst ct) si sr.
Proof.
  intros p ct si sr.
  apply (N.IntraBigStep3_ind_strong p
           (fun ct s ret => V2.IntraBigStep p (fst ct) s ret)
           (fun ct s1 s2 => V2.IntraStarStep p (fst ct) s1 s2)
           (fun ct s1 se => V2.IntraExceptionStep p (fst ct) s1 (snd (fst (fst se)),snd (fst se)))
           (fun ct s1 s2 => V2.IntraNormalStep p (fst ct) s1 s2) 
  ); intros.
  constructor 1 with s2; auto.
  inversion H3; subst.
  simpl in *.
  destruct s2 as [pc2 [[h2 s2] l2]].
  constructor 2 with pc2 l2 s2 h2; auto.
  elim IntraExceptionStep3_same_pc_l with (1:=H1); intros; subst; auto.
  constructor; auto.
  constructor 2 with s2; auto.
  inversion H3; subst.
  simpl in *.
  destruct s2 as [pc2 [[h2 s2] l2]].
  elim IntraExceptionStep3_same_pc_l with (1:=H1); intros; subst.
  constructor 3 with h2 s2 pc2; auto.
  inversion H; simpl.
  constructor; auto.
  constructor 2 with si0 sr0 s0; auto.
  inversion H0; subst.
  eapply call_and_return_exception_def; eauto.
  destruct ct0 as [m0 sf0].
  destruct s as [pc [[h s] l]].
  destruct si0 as [m' [s' l']].
  simpl in *.
  auto.
  constructor; auto.
  constructor 2 with si0 sr0 s0; auto.
  inversion H0; subst.
  eapply call_and_return_void; eauto.
  eapply call_and_return_value; eauto.
  destruct ct0 as [m0 sf0].
  destruct s as [pc [[h s] l]].
  destruct si0 as [m' [s'' l']].
  simpl in *.
  auto.
Qed.

Definition convertState2Bicolano (s:IntermediateSmallStep.State ): State.t :=
 match s with
    IntermediateSmallStep.mks (m,sf) (pc,(h,os,l)) => St h (Fr m pc os l) sf
  | IntermediateSmallStep.mke (m,sf) (pc,h,loc,l) => StE h (FrE m pc loc l) sf
 end.

Definition convertBicolano2State (s:State.t): IntermediateSmallStep.State :=
 match s with
    St h (Fr m pc os l) sf => IntermediateSmallStep.mks (m,sf) (pc,(h,os,l))
  | StE h (FrE m pc loc l) sf => IntermediateSmallStep.mke (m,sf) (pc,h,loc,l)
 end.

Lemma SmallStep_equiv1 : forall p st1 st2,
  IntermediateSmallStep.SmallStep p st1 st2 ->
  step p (convertState2Bicolano st1) (convertState2Bicolano st2).
Proof.
  intros.
  inversion H; subst; simpl in *; auto.
Qed.

Lemma SmallStep_equiv2 : forall p st1 st2,
  step p st1 st2 ->
  IntermediateSmallStep.SmallStep p
     (convertBicolano2State st1) (convertBicolano2State st2).
Proof.
  intros.
  destruct st1 as [h1 [m1 pc1 s1 l1] sf1|h1 [m1 pc1 loc1 l1] sf1]; simpl;
  destruct st2 as [h2 [m2 pc2 s2 l2] sf2|h2 [m2 pc2 loc2 l2] sf2]; simpl;
  constructor; auto.
Qed.

Lemma prefix_app : forall ct1 ct2,
  N.prefix ct1 ct2 ->
  exists sf, (snd ct1) = List.app sf (snd ct2).
Proof.
  induction 1.
  inversion_clear H; simpl.
  exists (cons (Frame.make m pc s l) nil).
  auto.
  exists (@nil Frame.t); auto.
  destruct IHclos_refl_trans1 as [sf1 H1]; rewrite H1.
  destruct IHclos_refl_trans2 as [sf2 H2]; rewrite H2.
  exists (List.app sf1 sf2).
  rewrite app_ass; auto.
Qed.

Lemma get_sf_context : forall st,
  State.get_sf (convertState2Bicolano st) = snd (N.context st).
Proof.
  destruct st as [[m sf] [pc [[h s] l]]|[m sf] [[[pc h] loc] l]]; 
  simpl; auto.  
Qed.    

Lemma get_sf_context2 : forall st,
  State.get_sf st = snd (N.context (convertBicolano2State st)).
Proof.
  destruct st as [h [m pc s l] sf|h [m pc loc l] sf]; simpl;
  simpl; auto.  
Qed.    

Lemma SmallStepTransitiveClosure_equiv1 : forall p st1 st2,
  N.SmallStepTransitiveClosure p st1 st2 ->
  step_closure_prefix_sf p (convertState2Bicolano st1) (convertState2Bicolano st2).
Proof.
  induction 1.
  constructor 1.
  constructor 2 with (convertState2Bicolano st2); auto.
  repeat rewrite get_sf_context.
  apply prefix_app; auto.
  apply SmallStep_equiv1; auto.
Qed.

Lemma length_app : forall (A:Type) (l1 l2:list A),
  length (l1++l2) = (length l1 + length l2)%nat.
Proof.
  induction l1; simpl; auto.
Qed.

Lemma step_closure_app : forall p st1 st3,
  step_closure_prefix_sf p st1 st3 ->
  exists sf, State.get_sf st3 = sf ++ State.get_sf st1.
Proof.
  induction 1.
  exists (@nil Frame.t); auto.
  destruct IHstep_closure_prefix_sf.
  destruct H0.
  rewrite H0.
  eauto.
Qed.

Lemma app_prefix : forall p st1 st3,
  step_closure_prefix_sf p st1 st3 ->
   N.prefix (N.context (convertBicolano2State st3))
     (N.context (convertBicolano2State st1)).
Proof.
  induction 1; intros.
  constructor 2.
  inversion H1; subst; simpl in *;
  try (apply IHstep_closure_prefix_sf; auto; fail).
  elim N.clos_refl_trans_inv2 with (1:=IHstep_closure_prefix_sf); intros.
  generalize (get_sf_context2 st1).
  destruct (N.context (convertBicolano2State st1)); simpl in *; intros.
  subst.
  injection H3; intros.
  destruct H0.
  assert (length (Frame.make m' pc' s' l' :: x ++ State.get_sf st1)
           = length (State.get_sf st1)).
  rewrite <- H0.
  injection H3; intros eq1 eq2; rewrite eq1; auto.

  simpl in H6; rewrite length_app in H6.
  apply False_ind; omega.

  destruct H3 as [ct' [U1 U2]].
  inversion U1; subst; auto.

  constructor 3 with (m,sf); auto.  
  constructor 1; constructor.
  constructor 3 with (m,sf); auto.  
  constructor 1; constructor.

  elim N.clos_refl_trans_inv2 with (1:=IHstep_closure_prefix_sf); intros.
  generalize (get_sf_context2 st1).
  destruct (N.context (convertBicolano2State st1)); simpl in *; intros.
  subst.
  injection H5; intros.
  destruct H0.
  assert (length (Frame.make m' pc' s' l' :: x ++ State.get_sf st1)
           = length (State.get_sf st1)).
  rewrite <- H0; rewrite H6; auto.
  simpl in H8; rewrite length_app in H8.
  apply False_ind; omega.
  destruct H5 as [ct' [U1 U2]].
  inversion U1; subst; auto.
  elim N.clos_refl_trans_inv2 with (1:=IHstep_closure_prefix_sf); intros.
  generalize (get_sf_context2 st1).
  destruct (N.context (convertBicolano2State st1)); simpl in *; intros.
  subst.
  injection H7; intros.
  destruct H0.
  assert (length (Frame.make m' pc' s' l' :: x ++ State.get_sf st1)
           = length (State.get_sf st1)).
  rewrite <- H0; rewrite H8; auto.
  simpl in H10; rewrite length_app in H10.
  apply False_ind; omega.
  destruct H7 as [ct' [U1 U2]].
  inversion U1; subst; auto.
Qed.

Lemma SmallStepTransitiveClosure_equiv2 : forall p st1 st2,
  step_closure_prefix_sf p st1 st2 ->
  N.SmallStepTransitiveClosure p (convertBicolano2State st1) (convertBicolano2State st2).
Proof.
  induction 1.
  constructor 1.
  constructor 2 with (convertBicolano2State st2); auto.
  apply app_prefix with p; auto.
  constructor 2 with st2; auto.
  apply SmallStep_equiv2; auto.
Qed.

Inductive IntraBigStep_from_SmallStep (P:Program) (sf:CallStack.t) (m:Method) : IntraNormalState -> ReturnState -> Prop := 
 | IntraBigStep_from_SmallStep_value : forall h1 pc1 l1 s1 h2 pc2 s2 l2 sr,
    step_closure_prefix_sf P 
      (St h1 (Fr m pc1 s1 l1) sf)
      (St h2 (Fr m pc2 s2 l2) sf) ->
    ReturnStep P m (pc2,(h2,s2,l2)) sr ->
    IntraBigStep_from_SmallStep P sf m (pc1,(h1,s1,l1)) sr
 | IntraBigStep_from_SmallStep_exception : forall h1 pc1 l1 s1 h2 pc2 loc2 l2,
    step_closure_prefix_sf P 
      (St h1 (Fr m pc1 s1 l1) sf)
      (StE h2 (FrE m pc2 loc2 l2) sf) ->
    UnCaughtException P m (pc2,h2,loc2) ->
    IntraBigStep_from_SmallStep P sf m (pc1,(h1,s1,l1)) (h2,Exception loc2).

Lemma IntraBigStep_from_SmallStep_equiv1 : forall p sf m s ret,
  IntraBigStep_from_SmallStep p sf m s ret ->
  N.IntraBigStep_from_SmallStep p (m,sf) s ret.
Proof.
  induction 1.
  constructor 1 with (pc2, (h2, s2, l2)); auto.
  replace (IntermediateSmallStep.mks (m, sf) (pc1, (h1, s1, l1)))
  with (convertBicolano2State (State.normal h1 (Frame.make m pc1 s1 l1) sf)); auto.
  replace (IntermediateSmallStep.mks (m, sf) (pc2, (h2, s2, l2)))
  with (convertBicolano2State (State.normal h2 (Frame.make m pc2 s2 l2) sf)); auto.
  apply SmallStepTransitiveClosure_equiv2; auto.
  constructor 2 with (pc2, h2, loc2, l2).
  replace (IntermediateSmallStep.mks (m, sf) (pc1, (h1, s1, l1)))
  with (convertBicolano2State (State.normal h1 (Frame.make m pc1 s1 l1) sf)); auto.
  replace (IntermediateSmallStep.mke (m, sf) (pc2, h2, loc2, l2))
  with (convertBicolano2State (State.exception h2 (ExceptionFrame.make m pc2 loc2 l2) sf)); auto.
  apply SmallStepTransitiveClosure_equiv2; auto.
  constructor; auto.
Qed.

Lemma IntraBigStep_from_SmallStep_equiv2 : forall p ct s ret,
  N.IntraBigStep_from_SmallStep p ct s ret ->
  IntraBigStep_from_SmallStep p (snd ct) (fst ct) s ret.
Proof.
  induction 1.
  destruct ct as [m sf].
  destruct s2 as [pc2 [[h2 s2] l2]].
  destruct s1 as [pc1 [[h1 s1] l1]].
  constructor 1 with h2 pc2 s2 l2.
  apply SmallStepTransitiveClosure_equiv1 with (1:=H).
  apply H0.
  destruct ct as [m sf].
  destruct se2 as [[[pc2 h2]loc2]l2].
  destruct s1 as [pc1 [[h1 s1] l1]].
  inversion H0; subst.
  constructor 2 with pc2 l2; auto.
  apply SmallStepTransitiveClosure_equiv1 with (1:=H).
Qed.

Theorem Equiv_SmallStep_BigStepV2 : forall P m s ret sf,
  V2.IntraBigStep P m s ret
    <-> 
  IntraBigStep_from_SmallStep P sf m s ret.
Proof.
  split; intros.
  apply IntraBigStep_from_SmallStep_equiv2 with (ct:=(m,sf)).
  apply N.equiv_IntraBigStep2_IntraBigStep_from_SmallStep.
  apply N.equiv_IntraBigStep3_IntraBigStep2.
  apply Bigstep_old_and_new; auto.
  apply Bigstep_old_and_new_inv with (ct:=(m,sf)); auto.
  apply N.equiv_IntraBigStep2_IntraBigStep3.
  apply N.equiv_IntraBigStep_from_SmallStep_IntraBigStep2.
  apply IntraBigStep_from_SmallStep_equiv1.
  assumption.
Qed.



Lemma V2IntraStarStep_NormalStep_right : forall p m s2 s3,
  V2.IntraStarStep p m s2 s3 -> forall s1,
  exec_intra p m s1 s2 ->
  V2.IntraStarStep p m s1 s3.
Proof.
  induction 1; intros.
  inversion_clear H.
  constructor 2 with s1; auto.
  constructor 1.
  constructor 1; auto.
  constructor 3 with h1 s0 pc1; auto.
  constructor 1.
  constructor 1; auto.
  constructor 2 with s2; auto.
  constructor 3 with h2 s2 pc; auto.
Qed.

Lemma V2IntraBigStep_NormalStep_right : forall p m s1 s2 r,
  V2.IntraBigStep p m s2 r ->
  exec_intra p m s1 s2 ->
  V2.IntraBigStep p m s1 r.
Proof.
  induction 1; intros.
  constructor 1 with s2; auto.
  apply V2IntraStarStep_NormalStep_right with s0; auto.
  constructor 2 with pc l s2 h2; auto.
  apply V2IntraStarStep_NormalStep_right with s0; auto.
Qed.

Lemma V2IntraStarStep_trans : forall p m s2 s3,
  V2.IntraStarStep p m s2 s3 -> forall s1,
  V2.IntraStarStep p m s1 s2 ->
  V2.IntraStarStep p m s1 s3.
Proof.
  induction 1; intros; auto.
  constructor 2 with s2; auto.
  constructor 3 with h2 s2 pc; auto.
Qed.

Lemma V2IntraStarStep_IntraBigStep : forall p m s1 s2 r,
  V2.IntraBigStep p m s2 r ->
  V2.IntraStarStep p m s1 s2 ->
  V2.IntraBigStep p m s1 r.
Proof.
  induction 1; intros.
  constructor 1 with s2; auto.
  apply V2IntraStarStep_trans with s0; auto.
  constructor 2 with pc l s2 h2; auto.
  apply V2IntraStarStep_trans with s0; auto.
Qed.

Lemma IntraStepStar_IntraBigStepV2 : forall P m s ret,
  IntraStepStar P m s (inr _ ret) ->
  V2.IntraBigStep P m s ret.
Proof.
  intros P.
  set (Pr:=fun m s (sr:IntraNormalState + ReturnState) =>
   match sr with
     inl s' => V2.IntraStarStep P m s s'
   | inr r => V2.IntraBigStep P m s r
   end).
  intros.
  apply IntraStepStar_ind with (p:=P) (P:=Pr) (m:=m) (s:=s) (r:=inr IntraNormalState ret);
  unfold Pr; simpl; intros.
  (**)
  constructor 1.
  (**)
  inversion H0; clear H0; subst.
  constructor 1 with s0; auto.
  constructor 1.
  constructor 2 with pc1 l1 s1 h1; auto.
  constructor 1.
  constructor 1; auto.
  (**)
  destruct r.
  apply V2IntraStarStep_NormalStep_right with s'; auto.
  apply V2IntraBigStep_NormalStep_right with s'; auto.
  (**)
  inversion H0; clear H0; subst.
  constructor 2 with pc1 l1 s1 h1; auto.
  constructor 1.
  constructor 2 with (1:=H4) (3:=H2).
  constructor; auto.
  (**)
  assert (V2.IntraStarStep P m0 s0 s'').
  inversion H0; clear H0; subst.
  constructor 2 with (pc1, (h1, s1, l1)).
  constructor 1.
  constructor 2 with (1:=H6) (3:=H2).
  destruct ov; simpl; constructor; auto.
  constructor 3 with h1 s1 pc1; auto.
  constructor 1.
  constructor 2 with (1:=H6) (3:=H2).
  constructor; auto.
  destruct r.
  apply V2IntraStarStep_trans with s''; auto.
  apply V2IntraStarStep_IntraBigStep with s''; auto.
  (**)
  auto.
Qed.

Lemma TransStep_l_trans_aux : forall (A B:Type) 
  (R : A -> A + B -> Prop) y' (x y:A),
  TransStep_l R x y' -> y' = (inl _ y) -> forall z,
  TransStep_l R y z ->
  TransStep_l R x z.
Proof. 
  induction 1; intros.
  inversion_clear H in H0; auto.
  constructor 3 with y; subst; auto.  
  constructor 3 with a2; subst; auto.
Qed.

Lemma TransStep_l_trans : forall (A B:Type) 
  (R : A -> A + B -> Prop) (x y:A) z,
  TransStep_l R x (inl _ y) -> 
  TransStep_l R y z ->
  TransStep_l R x z.
Proof. 
  intros.
  apply TransStep_l_trans_aux with (inl B0 y) y; auto.
Qed.

Lemma ReturnStep_exec_return : forall p m s ret,
  ReturnStep p m s ret ->
  exec_return p m s ret.
Proof.
  intros.
  destruct ret as [h [r|r]].
  constructor 1; auto.
  inversion_clear H.
Qed.

Lemma IntraBigStepV2_IntraStepStar : forall P m s ret,
  V2.IntraBigStep P m s ret ->
  IntraStepStar P m s (inr _ ret).
Proof.
  intros p m s ret.
  apply (V2.IntraBigStep_ind_strong p
           (fun m s ret => IntraStepStar p m s (inr _ ret))
           (fun m s1 s2 => IntraStepStar p m s1 (inl _ s2))
           (fun m s1 se => forall h loc, se = (h,loc) ->
			     (UnCaughtException p m (fst s1, h, loc) ->
			      IntraStepStar p m s1 (inr _ (h, Exception loc)))
                           /\(forall pc', CaughtException p m (fst s1, h, loc) pc' ->
                              IntraStepStar p m s1 (inl _ (pc', (h, Ref loc :: OperandStack.empty, (snd (snd s1)))))))
           (fun m s1 s2 => IntraStepStar p m s1 (inl _ s2))
  ); intros; simpl in *.
  (**)
  unfold IntraStepStar; apply TransStep_l_trans with s2; auto.
  constructor 2; constructor 1.
  apply ReturnStep_exec_return; auto.
  (**)
  unfold IntraStepStar; apply TransStep_l_trans with (pc, (h2, s2, l)); auto.
  destruct (H2 h loc); auto.
  apply H4; auto.
  (**)
  constructor 1.
  (**)
  unfold IntraStepStar; apply TransStep_l_trans with s2; auto.
  (**)
  unfold IntraStepStar; apply TransStep_l_trans with (pc, (h2, s2, l)); auto.
  destruct (H2 h loc); auto.
  apply H5; auto.
  (**)
  subst; split; intros; constructor 2.
  constructor 1.
  destruct s1 as [pc1 [[h1 s1] l1]].
  constructor 2; auto.
  constructor 2.
  destruct s1 as [pc1 [[h1 s1] l1]].
  constructor 2; auto.
  (**)
  subst; split; intros.
  constructor 2.
  constructor 3 with (fst si) s1 sr; auto.
  inversion H0; clear H0; subst; econstructor; eauto.
  constructor 2.
  constructor 3 with (fst si) s1 sr; auto.
  inversion H0; clear H0; subst; econstructor; eauto.
  (**)
  constructor 2; constructor 2; constructor 1; auto.
  (**)
  constructor 2.
  constructor 3 with (fst si) s1 sr; auto.
  inversion H0; clear H0; subst.
  simpl; replace s'0 with (cons_option (@None value) s'0).
  constructor 1; auto.
  destruct s'0; simpl; auto.
  simpl; replace (v::s'0) with (cons_option (Some v) s'0).
  econstructor; eauto.
  destruct s'0; simpl; auto.
Qed.  
    
Theorem Equiv_SmallStep_BigStep : forall P m s ret sf,
  BigStep P m s ret
    <-> 
  IntraBigStep_from_SmallStep P sf m s ret.
Proof.
  intros.
  unfold BigStep.
  elim Equiv_SmallStep_BigStepV2 with P m s ret sf; intros.
  split; intros.
  apply H; apply IntraStepStar_IntraBigStepV2; auto. 
  unfold BigStep.
  apply IntraBigStepV2_IntraStepStar; apply H0; auto.
Qed.

End MakeEquivSmallAndBigStep.

(* 
 Local Variables: 
 coq-prog-name: "coqtop -emacs"
 End:
*)
