Add LoadPath "../Library".
Add LoadPath "../Library/Map".
Require Export ZArith.
Require Export Program.
Require Numeric.
Require Export List.
Require Export BinArray.
Require Export PosAux.
Require Export EqBoolAux.
Require Domain.
Open Scope Z_scope.

Ltac caseeq t := generalize (refl_equal t); pattern t at -1 in |- * ; case t.

(** All semantic domains and basic operation are encapsulated in a module signature *)

Module Make (PROG:PROGRAM 
                with Definition Var := N
                with Definition ShortFieldName := positive) <: Domain.SEMANTIC_DOMAIN.

 (** We depend on the choices done for program data structures *)
 Module Prog :=PROG.
 Import Prog.

 Module SByte <: Numeric.NUMSIZE. Definition power := 7. End SByte.
 Module SShort<: Numeric.NUMSIZE. Definition power := 15. End SShort.
 Module SInt  <: Numeric.NUMSIZE. Definition power := 31. End SInt.

 Module Byte  : Numeric.NUMERIC with Definition power := 7 := Numeric.Make SByte.
 Module Short : Numeric.NUMERIC with Definition power := 15 := Numeric.Make SShort.
 Module Int   : Numeric.NUMERIC with Definition power := 31 := Numeric.Make SInt.

 (** conversion *)
 Parameter b2i : Byte.t -> Int.t.
 Parameter b2i_prop : forall b, Byte.toZ b = Int.toZ (b2i b).

 Parameter s2i : Short.t -> Int.t.
 Parameter s2i_prop : forall s, Short.toZ s = Int.toZ (s2i s).

 Parameter i2b : Int.t -> Byte.t.
 Parameter i2b_prop1 : forall i,
   (Int.toZ i) mod 2^8 < 2^7 ->
   (Int.toZ i) mod 2^8 = Byte.toZ (i2b i).
 Parameter i2b_prop2 : forall i,
   2^7 <= (Int.toZ i) mod 2^8 ->
   (Int.toZ i) mod 2^8 - 2^8 = Byte.toZ (i2b i).

 Parameter i2s : Int.t -> Short.t.
 Parameter i2s_prop1 : forall i,
   (Int.toZ i) mod 2^16 < 2^15 ->
   (Int.toZ i) mod 2^16 = Short.toZ (i2s i).
 Parameter i2s_prop2 : forall i,
   2^15 <= (Int.toZ i) mod 2^16 ->
   (Int.toZ i) mod 2^16 - 2^15 = Short.toZ (i2s i).

 Parameter i2bool : Int.t -> Byte.t.
 Parameter i2bool_prop_true : forall i,
   Byte.toZ (i2bool i) = Int.toZ i mod 2.

 Inductive num : Set :=
   | I : Int.t -> num
   | B : Byte.t -> num
   | Sh : Short.t -> num.
 
 (** Location is the domain of adresses in the heap *)
 Definition Location : Set := N.
 Definition Location_dec : forall loc1 loc2:Location,{loc1=loc2}+{~loc1=loc2} :=
   (Aeq_Dec _ _ Neq_spec).

 Inductive value : Set :=
   | Num : num -> value
   | Ref: Location -> value
   | Null : value.

 Definition init_value (t:type) : value :=
    match t with
     | ReferenceType _ => Null
     | PrimitiveType _ => Num (I (Int.const 0))
    end.

 Definition init_field_value (f:Field) : value :=
   match FIELD.initValue f with
    | FIELD.Int z => Num (I (Int.const z))
    | FIELD.NULL => Null
    | FIELD.UNDEF => init_value (FIELDSIGNATURE.type (FIELD.signature f))
  end.
 
 (** Domain of local variables *)
 Module Type LOCALVAR.
   Parameter t : Type.
   Parameter get : t-> Var -> option value.
   Parameter update : t -> Var -> value -> t.
   Parameter get_update_new : forall l x v, get (update l x v) x = Some v.
   Parameter get_update_old : forall l x y v,
     x<>y -> get (update l x v) y = get l y.
 End LOCALVAR.

 Module MapVar <: MAP with Definition key := Var := BinNatMap.

 Module LocalVar <: LOCALVAR.
   Definition t := MapVar.t value.
   Definition get : t -> Var -> option value := @MapVar.get value.
   Definition update : t -> Var -> value -> t := @MapVar.update value.
   Lemma get_update_new : forall l x v, get (update l x v) x = Some v.
   Proof. exact (MapVar.get_update1 value). Qed.
   Lemma get_update_old : forall l x y v,
     x<>y -> get (update l x v) y = get l y.
   Proof. 
    intros;refine (MapVar.get_update2 value _ _ _ _ _). 
    intro;apply H;subst;trivial.
  Qed.
  Definition empty := MapVar.empty value.
 End LocalVar.

 (* Domain of operand stacks *) 
 Module Type OPERANDSTACK.
   Definition t : Set := list value.
   Definition empty : t := nil.
   Definition push : value -> t -> t := fun v t => cons v t.
   Definition size : t -> nat := fun t  => length t .
   Definition get_nth : t -> nat -> option value := fun s n => nth_error s n.
 End OPERANDSTACK.

 Module OperandStack <: OPERANDSTACK.
   Definition t : Set := list value.
   Definition empty : t := nil.
   Definition push : value -> t -> t := fun v t => cons v t.
   Definition size : t -> nat := fun t  => length t .
   Definition get_nth : t -> nat -> option value := fun s n => nth_error s n.
  End OperandStack.

 (** Transfert fonction between operand stack and local variables necessary for invoke instructions *)
  
 Fixpoint stack2localvar_rec
    (os:OperandStack.t) (n:nat) (l:LocalVar.t) {struct n}: LocalVar.t :=
   match n with 
   | O => l
   | S n => 
      match os with
      | nil => l
      | v::os => 
         stack2localvar_rec os n (LocalVar.update l (N_toVar n) v)
      end
   end.
 Definition stack2localvar os n := stack2localvar_rec os n LocalVar.empty.

 Lemma stack2locvar_rec_prop1 :
   forall n l os x, (n <= Var_toN x)%nat -> 
     LocalVar.get (stack2localvar_rec  os n l) x = LocalVar.get l x.
 Proof.
  induction n;simpl;intros;trivial.
  destruct os;simpl;intros;trivial.
  assert (n <= Var_toN x)%nat. omega.
  rewrite IHn;trivial.
  unfold LocalVar.get, LocalVar.update;rewrite MapVar.get_update2;trivial.
  intro Heq;rewrite Heq in H.
  rewrite Var_toN_bij2 in H;omega.
 Qed.

 Lemma stack2locvar_prop1 :
   forall s n x, (n <= Var_toN x)%nat -> LocalVar.get (stack2localvar s n) x = None.
 Proof.
   unfold stack2localvar;intros.
   rewrite stack2locvar_rec_prop1;trivial.
   exact (MapVar.get_empty _ _).
 Qed.

 Lemma stack2locvar_rec_prop2 :
   forall n os l x, (Var_toN x < n)%nat ->
     (forall y, (Var_toN y < n)%nat -> LocalVar.get l y = None) ->
     LocalVar.get (stack2localvar_rec os n l) x = 
       OperandStack.get_nth os (n-(Var_toN x)-1)%nat.
 Proof.
   induction n;simpl;intros.
   elimtype False;omega.
   destruct os.
   rewrite H0;trivial.
   destruct (Var_toN x);simpl;trivial.
   destruct (n-0)%nat;trivial.
   destruct (n-n0-1)%nat;trivial.
   change (match Var_toN x with
   | O => (S n)
   | S l0 => (n - l0)
   end - 1)%nat with (S n- Var_toN x -1)%nat.
   assert (Var_toN x = n \/ Var_toN x < n)%nat. omega.
   destruct H1.
   pattern n at 2;rewrite <- H1;rewrite Var_toN_bij1.
   rewrite stack2locvar_rec_prop1. 2:omega.
   rewrite LocalVar.get_update_new.
   replace (S n - Var_toN x - 1)%nat with O;try omega.
   simpl;trivial.
   rewrite IHn;trivial.
   replace (S n - Var_toN x - 1)%nat with (S (n - Var_toN x - 1))%nat;try omega.
   simpl;trivial.
   intros;rewrite LocalVar.get_update_old.
   apply H0;omega.
   intro Heq;rewrite <-Heq in H2;rewrite Var_toN_bij2 in H2;omega.
  Qed.
   
 Lemma stack2locvar_prop2 :
   forall s n x, (Var_toN x < n)%nat ->
     LocalVar.get (stack2localvar s n) x = OperandStack.get_nth s (n-(Var_toN x)-1)%nat.
 Proof.
  unfold stack2localvar;intros;apply stack2locvar_rec_prop2;trivial.
  intros;exact (MapVar.get_empty _ _).
 Qed.

 Module Type HEAP.
   Parameter t : Type.

   Inductive AdressingMode : Set :=
     | StaticField : FieldSignature -> AdressingMode
     | DynamicField : Location -> FieldSignature -> AdressingMode
     | ArrayElement : Location -> Int.t -> AdressingMode.

   Inductive LocationType : Set :=
     | LocationObject : ClassName -> LocationType  
     | LocationArray : Int.t -> type -> LocationType.
   (** (LocationArray length element_type) *)

   Parameter get : t -> AdressingMode -> option value.
   Parameter update : t -> AdressingMode -> value -> t.
   Parameter typeof : t -> Location -> option LocationType.   
     (** typeof h loc = None -> no object, no array allocated at location loc *)
   Parameter new : t -> Program -> LocationType -> option (Location * t).
     (** program is required to compute the size of the allocated element, i.e. to know
        the Class associated with a ClassName  *)

   (** Compatibility between a heap and an adress *)
   Inductive Compat (h:t) : AdressingMode -> Prop :=
     | CompatStatic : forall f,
         Compat h (StaticField f)
     | CompatObject : forall cn loc f,
         typeof h loc = Some (LocationObject cn) ->
         Compat h (DynamicField loc f)
     | CompatArray : forall length tp loc i,
         0 <= Int.toZ i < Int.toZ length ->
         typeof h loc = Some (LocationArray length tp) ->
         Compat h (ArrayElement loc i).

   Parameter get_update_same : forall h am v, Compat h am ->  get (update h am v) am = Some v.
   Parameter get_update_old : forall h am1 am2 v, am1<>am2 -> get (update h am1 v) am2 = get h am2.
   Parameter get_uncompat : forall h am, ~ Compat h am -> get h am = None.

   Parameter typeof_update_same : forall h loc am v,
     typeof (update h am v) loc = typeof h loc.

   Parameter new_fresh_location : forall (h:t) (p:Program) (lt:LocationType) (loc:Location) (h':t),
     new h p lt = Some (loc,h') ->
     typeof h loc = None.

   Parameter new_typeof : forall (h:t) (p:Program) (lt:LocationType) (loc:Location) (h':t),
     new h p lt = Some (loc,h') ->
     typeof h' loc = Some lt.

   Parameter new_typeof_old : forall (h:t) (p:Program) (lt:LocationType) (loc loc':Location) (h':t),
     new h p lt = Some (loc,h') ->
     loc <> loc' ->
     typeof h' loc' = typeof h loc'.

   Parameter new_defined_object_field : forall (h:t) (p:Program) (cn:ClassName) (fs:FieldSignature) (f:Field) (loc:Location) (h':t),
     new h p (LocationObject cn) = Some (loc,h') ->
     is_defined_field p cn fs f ->
     get h' (DynamicField loc fs) = Some (init_field_value f).

   Parameter new_undefined_object_field : forall (h:t) (p:Program) (cn:ClassName) (fs:FieldSignature) (loc:Location) (h':t),
     new h p (LocationObject cn) = Some (loc,h') ->
     ~ defined_field p cn fs ->
     get h' (DynamicField loc fs) = None.
 
  Parameter new_object_no_change : 
     forall (h:t) (p:Program) (cn:ClassName) (loc:Location) (h':t) (am:AdressingMode),
     new h p (LocationObject cn) = Some (loc,h') ->
     (forall (fs:FieldSignature), am <> (DynamicField loc fs)) ->
     get h' am = get h am.

  Parameter new_valid_array_index : forall (h:t) (p:Program) (length:Int.t) (tp:type) (i:Int.t) (loc:Location) (h':t),
     new h p (LocationArray length tp) = Some (loc,h') ->
     0 <= Int.toZ i < Int.toZ length ->
     get h' (ArrayElement loc i) = Some (init_value tp).

  Parameter new_unvalid_array_index : forall (h:t) (p:Program) (length:Int.t) (tp:type) (i:Int.t) (loc:Location) (h':t),
     new h p (LocationArray length tp) = Some (loc,h') ->
     ~ 0 <= Int.toZ i < Int.toZ length ->
     get h' (ArrayElement loc i) = None.

  Parameter new_array_no_change : 
     forall (h:t) (p:Program) (length:Int.t) (tp:type) (loc:Location) (h':t) (am:AdressingMode),
     new h p (LocationArray length tp) = Some (loc,h') ->
     (forall (i:Int.t), am <> (ArrayElement loc i)) ->
     get h' am = get h am.

(* These properties should be useless
   Parameter get_static_some : forall (h:t) (p:Program) (fs:FieldSignature),
     isStatic p fs ->
     exists v, get h (StaticField fs) = Some v.

   Parameter get_static_some : forall (h:t) (p:Program) (fs:FieldSignature),
     ~ isStatic p fs ->
     exists v, get h (StaticField fs) = None.
*)

 End HEAP.

  Module FieldSignatureProj.
    Definition element:= FieldSignature.
    Definition key:=positive.
    Definition  proj (f:element) := FIELDSIGNATURE.name (snd f).
  End FieldSignatureProj.

  Module MapSignatureField := MapProj FieldSignatureProj.

Module Heap.
  Module Object.
    Record t : Type := Obj { class : ClassName; fieldsval : BinMap.t value }.
    Definition get (o:t) := BinMap.get value o.(fieldsval).
    Definition op_get (o:option t) :=
      match o with
        | Some o => get o
        | None => fun f => None
      end.
    Definition op_update (f:ShortFieldSignature) (v:value) (o:option t) : option t :=
      match o with
        | None => None
        | Some o => Some (Obj o.(class) (BinMap.update _ o.(fieldsval) (FIELDSIGNATURE.name f) v))
      end.

    Definition init_fields :  list Field -> BinMap.t value :=
      @fold_right (BinMap.t value) Field
        (fun f m => BinMap.update _ m (FIELDSIGNATURE.name (FIELD.signature f)) (init_field_value f))
        (BinMap.empty _).

    Definition default (c:Class) : t :=
      Obj
      (CLASS.name c) 
      (init_fields (CLASS.definedFields c)).

  End Object.
  Module Array.
    Record t : Type := Arr {
      length : Int.t;
      element_type : type; 
      elements : BinNatMap.t value 
    }.
    Definition get (a:t) (i:Int.t) :=
      match Int.toZ i with
        | Z0 => BinNatMap.get value a.(elements) N0
        | Zpos p => 
          match Zcompare (Int.toZ i) (Int.toZ a.(length)) with
            | Lt =>  BinNatMap.get value a.(elements) (Npos p)
            | _ => None
          end
        | Zneg _ => None
      end.
    Definition op_get (a:option t) :=
      match a with
        | Some a => get a
        | None => fun i => None
      end.
    Definition op_update (i:Int.t) (v:value) (a:option t) : option t :=
      match a with
        | None => None
        | Some a => 
          match Int.toZ i with
            | Z0 => Some (Arr a.(length) a.(element_type) (BinNatMap.update _ a.(elements) N0 v))
            | Zpos p => 
              match Zcompare (Int.toZ i) (Int.toZ a.(length)) with
                | Lt => Some 
                         (Arr a.(length) a.(element_type) (BinNatMap.update _ a.(elements) (Npos p) v))
                | _ => Some a
              end
            | Zneg _ => Some a
          end
      end.

    Fixpoint init_array_rec (n:nat) (v:value) {struct n} : BinNatMap.t value :=
      match n with 
        | O => BinNatMap.empty _
        | S p => BinNatMap.update _ (init_array_rec p v) (N_of_nat p) v
      end.

    Definition default (lgth:Int.t) (n:nat) (tp:type) : t :=
      Arr lgth tp (init_array_rec n (init_value tp)).
  End Array.

  Module LocMap := BinNatMap.

  Record t_ : Type := hp {
    statics : BinMap.t value;
    objects : LocMap.t Object.t;
    arrays : LocMap.t Array.t;
    next_loc : Location
  }.
  Definition t : Type := t_.
  
  Inductive AdressingMode : Set :=
  | StaticField : FieldSignature -> AdressingMode
  | DynamicField : Location -> FieldSignature -> AdressingMode
  | ArrayElement : Location -> Int.t -> AdressingMode.

  Inductive LocationType : Set :=
  | LocationObject : ClassName -> LocationType  
  | LocationArray : Int.t -> type -> LocationType.

   Definition get (h:t) (am:AdressingMode) : option value :=
     match am with
       | StaticField (_,f) =>  BinMap.get _ h.(statics) (FIELDSIGNATURE.name f)
       | DynamicField l (_,f) => Object.op_get (LocMap.get _ h.(objects) l) (FIELDSIGNATURE.name f)
       | ArrayElement l i => Array.op_get (LocMap.get _ h.(arrays) l) i 
     end.
   Definition update (h:t) (am:AdressingMode) (v:value) : t :=
     match am with
       | StaticField (_,f) => 
           hp
            (BinMap.update _ h.(statics) (FIELDSIGNATURE.name f) v)
            h.(objects)
            h.(arrays)
            h.(next_loc)
       | DynamicField l (_,f) => 
           hp
            h.(statics)
            (LocMap.modify _ h.(objects) l (Object.op_update f v))
            h.(arrays)              
            h.(next_loc)
       | ArrayElement l i => 
           hp
            h.(statics)
            h.(objects)
            (LocMap.modify _ h.(arrays) l (Array.op_update i v))
            h.(next_loc)
     end.     
   Definition typeof (h:t) (l:Location) : option LocationType :=
     match LocMap.get _ h.(objects) l with
       | Some o => Some (LocationObject o.(Object.class))
       | None =>
         match LocMap.get _ h.(arrays) l with
           | Some a => Some (LocationArray a.(Array.length) a.(Array.element_type))
           | None => None
         end
     end.

   Definition Int_to_nat (i:Int.t) : option nat :=
     match Int.toZ i with
       | Zpos p => Some (nat_of_P p)
       | Z0 => Some O
       | Zneg _ => None
     end.

   Definition new (h:t) (p:Program) (ltp:LocationType) : option (Location * t) :=
     match ltp with
       | LocationObject cn => 
         match PROG.class p cn with
           | None => None
           | Some c => Some
               (h.(next_loc),
                hp
                 h.(statics)
                 (LocMap.update _ h.(objects) h.(next_loc) (Object.default c))
                 h.(arrays)              
                 (Nsucc h.(next_loc)))
         end
       | LocationArray lgth tp => 
         match Int_to_nat lgth with
           | None => None
           | Some n =>
             Some
               (h.(next_loc),
                hp
                 h.(statics)
                 h.(objects)              
                 (LocMap.update _ h.(arrays) h.(next_loc) (Array.default lgth n tp))
                 (Nsucc h.(next_loc)))
         end
     end.

   (** Compatibility between a heap and an adress *)
   Inductive Compat (h:t) : AdressingMode -> Prop :=
     | CompatStatic : forall f,
         Compat h (StaticField f)
     | CompatObject : forall cn loc f,
         typeof h loc = Some (LocationObject cn) ->
         Compat h (DynamicField loc f)
     | CompatArray : forall length tp loc i,
         0 <= Int.toZ i < Int.toZ length ->
         typeof h loc = Some (LocationArray length tp) ->
         Compat h (ArrayElement loc i).

   Lemma get_update_same : forall h am v, Compat h am ->  get (update h am v) am = Some v.
   Proof.
     destruct 1; simpl.
     destruct f as [_ f]; simpl.
     rewrite BinMap.get_update1; auto.
     destruct f as [_ f]; simpl.
     rewrite LocMap.get_modify1; auto.
     caseeq (BinNatMap_Base.get Object.t (objects h) loc); intros; simpl.
     unfold Object.get; simpl.
     rewrite BinMap.get_update1; auto.
     unfold typeof in H.
     replace LocMap.get with BinNatMap_Base.get in H; auto.
     rewrite H0 in H.
     destruct (BinNatMap_Base.get Array.t (arrays h) loc); discriminate.
     rewrite LocMap.get_modify1; auto.
     unfold Array.op_update.
     unfold typeof in H0.
     destruct (LocMap.get Object.t (objects h) loc); simplify_eq H0.
     clear H0; intros.
     replace LocMap.get with BinNatMap_Base.get in H0; auto.
     caseeq (BinNatMap_Base.get _ (arrays h) loc); intros.
     rewrite H1 in H0.
     injection H0; intros; subst; clear H0.
     caseeq (Int.toZ i); intros.
     simpl.
     unfold Array.get; simpl.
     rewrite H0.
     destruct t0; simpl.
     destruct elements; simpl; reflexivity.
     rewrite <- H0.
     destruct H.
     generalize (Zlt_compare _ _ H2).
     destruct (Int.toZ i ?= Int.toZ (Array.length t0)); intros H3; try elim H3.
     simpl; unfold Array.get.
     rewrite H0.
     rewrite <- H0.
     generalize (Zlt_compare _ _ H2).
     simpl Array.length.
     destruct (Int.toZ i ?= Int.toZ (Array.length t0)); intros H'; try elim H'.
     simpl.
     apply (BinNatMap.get_update1 _ (Array.elements t0) (Npos p) v).
     rewrite H0 in H.
     destruct H.
     destruct H.
     reflexivity.
     rewrite H1 in H0; discriminate.
   Qed.

   Axiom get_update_old : forall h am1 am2 v, am1<>am2 -> get (update h am1 v) am2 = get h am2.
(* Proof.
     destruct am1; destruct am2; simpl; intros.
     destruct f0; destruct f; simpl.
     2: destruct f0; destruct f; simpl; auto.
     2: destruct f; simpl; auto.
     2: destruct f0; destruct f; simpl; auto.
     2: destruct f0; destruct f; simpl; auto.
     3: destruct f; simpl; auto.
     3: destruct f; simpl; auto.
     3: destruct f; simpl; auto.
*)
     
   Parameter get_uncompat : forall h am, ~ Compat h am -> get h am = None.

   Parameter typeof_update_same : forall h loc am v,
     typeof (update h am v) loc = typeof h loc.

   Parameter new_fresh_location : forall (h:t) (p:Program) (lt:LocationType) (loc:Location) (h':t),
     new h p lt = Some (loc,h') ->
     typeof h loc = None.

   Parameter new_typeof : forall (h:t) (p:Program) (lt:LocationType) (loc:Location) (h':t),
     new h p lt = Some (loc,h') ->
     typeof h' loc = Some lt.

   Parameter new_typeof_old : forall (h:t) (p:Program) (lt:LocationType) (loc loc':Location) (h':t),
     new h p lt = Some (loc,h') ->
     loc <> loc' ->
     typeof h' loc' = typeof h loc'.

   Parameter new_defined_object_field : forall (h:t) (p:Program) (cn:ClassName) (fs:FieldSignature) (f:Field) (loc:Location) (h':t),
     new h p (LocationObject cn) = Some (loc,h') ->
     is_defined_field p cn fs f ->
     get h' (DynamicField loc fs) = Some (init_field_value f).

   Parameter new_undefined_object_field : forall (h:t) (p:Program) (cn:ClassName) (fs:FieldSignature) (loc:Location) (h':t),
     new h p (LocationObject cn) = Some (loc,h') ->
     ~ defined_field p cn fs ->
     get h' (DynamicField loc fs) = None.
 
  Parameter new_object_no_change : 
     forall (h:t) (p:Program) (cn:ClassName) (loc:Location) (h':t) (am:AdressingMode),
     new h p (LocationObject cn) = Some (loc,h') ->
     (forall (fs:FieldSignature), am <> (DynamicField loc fs)) ->
     get h' am = get h am.

  Parameter new_valid_array_index : forall (h:t) (p:Program) (length:Int.t) (tp:type) (i:Int.t) (loc:Location) (h':t),
     new h p (LocationArray length tp) = Some (loc,h') ->
     0 <= Int.toZ i < Int.toZ length ->
     get h' (ArrayElement loc i) = Some (init_value tp).

  Parameter new_unvalid_array_index : forall (h:t) (p:Program) (length:Int.t) (tp:type) (i:Int.t) (loc:Location) (h':t),
     new h p (LocationArray length tp) = Some (loc,h') ->
     ~ 0 <= Int.toZ i < Int.toZ length ->
     get h' (ArrayElement loc i) = None.

  Parameter new_array_no_change : 
     forall (h:t) (p:Program) (length:Int.t) (tp:type) (loc:Location) (h':t) (am:AdressingMode),
     new h p (LocationArray length tp) = Some (loc,h') ->
     (forall (i:Int.t), am <> (ArrayElement loc i)) ->
     get h' am = get h am.

(* These properties should be useless
   Parameter get_static_some : forall (h:t) (p:Program) (fs:FieldSignature),
     isStatic p fs ->
     exists v, get h (StaticField fs) = Some v.

   Parameter get_static_some : forall (h:t) (p:Program) (fs:FieldSignature),
     ~ isStatic p fs ->
     exists v, get h (StaticField fs) = None.
*)

 End Heap.

  Inductive ReturnVal : Set :=
   | Normal : option value -> ReturnVal
   | Exception : Location -> ReturnVal.

 (** input-output native method execution *)
 Inductive NativeMethod : Type := 
 | NoImplem
 | Implem :  (Heap.t * list value -> Heap.t * ReturnVal -> Prop) -> NativeMethod.
 Parameter native_implementation : Method -> NativeMethod.

(*
 Parameter foo : Method.
 Parameter foo_is_native : METHOD.isNative foo = true.
 Inductive foo_implementation : Heap.t * list value -> Heap.t * ReturnVal -> Prop :=
   foo_implementation_def : forall h,
     foo_implementation (h,nil) (h,Normal (Some (Num (I (Int.fromZ 1))))).
 Parameter foo_implementation_prop : METHOD.isNative foo = true.
   native_implementation foo = Implem foo_implementation. 
*)

 (** Domain of frames *)
 Module Type FRAME.
   Inductive t : Type := 
      make : Method -> PC -> OperandStack.t -> LocalVar.t -> t.
 End FRAME.
 
 Module Frame.
   Inductive t : Type := 
      make : Method -> PC -> OperandStack.t -> LocalVar.t -> t.
 End Frame.

 (** Domain of call stacks *)
 Module Type CALLSTACK.
   Definition t : Type := list Frame.t.
 End CALLSTACK.

 Module CallStack.
   Definition t : Type := list Frame.t.
 End CallStack.

 Module Type EXCEPTION_FRAME.
   Inductive t : Type := 
      make : Method -> PC -> Location -> LocalVar.t -> t.
 End EXCEPTION_FRAME.
 Module ExceptionFrame.
    Inductive t : Type := 
      make : Method -> PC -> Location -> LocalVar.t -> t.
  End ExceptionFrame.
 
 (** Domain of states *)
 Module Type STATE.
   Inductive t : Type := 
      normal : Heap.t -> Frame.t -> CallStack.t -> t
    | exception : Heap.t -> ExceptionFrame.t -> CallStack.t -> t.
   Definition get_sf (s:t) : CallStack.t :=
     match s with
       normal _ _ sf => sf
     | exception _ _ sf => sf
     end.
   Definition get_m (s:t) : Method :=
     match s with
       normal _ (Frame.make m _ _ _)_ => m
     | exception _ (ExceptionFrame.make m _ _ _) _ => m
     end.
 End STATE.

 Module State.
  Inductive t : Type := 
      normal : Heap.t -> Frame.t -> CallStack.t -> t
    | exception : Heap.t -> ExceptionFrame.t -> CallStack.t -> t.
   Definition get_sf (s:t) : CallStack.t :=
     match s with
       normal _ _ sf => sf
     | exception _ _ sf => sf
     end.
   Definition get_m (s:t) : Method :=
     match s with
       normal _ (Frame.make m _ _ _)_ => m
     | exception _ (ExceptionFrame.make m _ _ _) _ => m
     end.
 End State.
 (** Some notations *)
 Notation St := State.normal.
 Notation StE := State.exception.
 Notation Fr := Frame.make.
 Notation FrE := ExceptionFrame.make.

  (** compatibility between ArrayKind and type *) 
  Inductive compat_ArrayKind_type : ArrayKind -> type -> Prop :=
    | compat_ArrayKind_type_ref : forall rt,
        compat_ArrayKind_type Aarray (ReferenceType rt)
    | compat_ArrayKind_type_int : 
        compat_ArrayKind_type Iarray (PrimitiveType INT)
    | compat_ArrayKind_type_byte : 
        compat_ArrayKind_type Barray (PrimitiveType BYTE)
    | compat_ArrayKind_type_bool : 
        compat_ArrayKind_type Barray (PrimitiveType BOOLEAN)
    | compat_ArrayKind_type_short : 
        compat_ArrayKind_type Sarray (PrimitiveType SHORT).

  Inductive isReference : value -> Prop :=
  | isReference_null : isReference Null
  | isReference_ref : forall loc, isReference (Ref loc).

  (** compatibility between ValKind and value *) 
  Inductive compat_ValKind_value : ValKind -> value -> Prop :=
    | compat_ValKind_value_ref : forall v,
        isReference v -> compat_ValKind_value Aval v
    | compat_ValKind_value_int : forall n,
        compat_ValKind_value Ival (Num (I n)).

  (** compatibility between ArrayKind and value *) 
  Inductive compat_ArrayKind_value : ArrayKind -> value -> Prop :=
    | compat_ArrayKind_value_ref : forall v,
        isReference v -> compat_ArrayKind_value Aarray v
    | compat_ArrayKind_value_int : forall n,
        compat_ArrayKind_value Iarray (Num (I n))
    | compat_ArrayKind_value_byte : forall n,
        compat_ArrayKind_value Barray (Num (B n))
    | compat_ArrayKind_value_short : forall n,
        compat_ArrayKind_value Sarray (Num (Sh n)).

  (* convert a value to be pushed on the stack *)
  Definition conv_for_stack (v:value) : value :=
    match v with
    | Num (B b) => Num (I (b2i b))
    | Num (Sh s) => Num (I (s2i s))
    | _ => v
    end.

  (* convert a value to be store in an array *)
  Definition conv_for_array (v:value) (t:type) : value :=
    match v with
    | Ref loc => v
    | Num (I i) =>
       match t with
         ReferenceType _ => v (* impossible case *)
       | PrimitiveType INT => v
       | PrimitiveType BOOLEAN => Num (B (i2bool i))
       | PrimitiveType BYTE => Num (B (i2b i))
       | PrimitiveType SHORT => Num (Sh (i2s i))         
       end
    | _ => v (* impossible case *)
    end.

  (** [assign_compatible_num source target] holds if a numeric value [source] can be 
    assigned to a variable of type [target]. This point is not clear in the JVM spec. *)
  Inductive assign_compatible_num : num -> primitiveType -> Prop :=
   | assign_compatible_int_int : forall i, assign_compatible_num (I i) INT
   | assign_compatible_short_int : forall sh, assign_compatible_num (Sh sh) INT
   | assign_compatible_byte_int : forall b, assign_compatible_num (B b) INT
   | assign_compatible_short_short : forall sh, assign_compatible_num (Sh sh) SHORT
   | assign_compatible_byte_byte : forall b, assign_compatible_num (B b) BYTE
   | assign_compatible_byte_boolean : forall b, assign_compatible_num (B b) BOOLEAN.

  (** [assign_compatible h source target] holds if a value [source] can be 
    assigned to a variable of type [target] *)
  Inductive assign_compatible (p:Program) (h:Heap.t) : value -> type -> Prop :=
   | assign_compatible_null : forall t, assign_compatible p h Null (ReferenceType t)
   | assign_compatible_ref_object_val : forall (loc:Location) (t:refType) (cn:ClassName), 
       Heap.typeof h loc = Some (Heap.LocationObject cn) ->
       compat_refType p (ClassType cn) t ->
       assign_compatible p h (Ref loc) (ReferenceType t)
   | assign_compatible_ref_array_val : forall (loc:Location) (t:refType) (length:Int.t) (tp:type), 
       Heap.typeof h loc = Some (Heap.LocationArray length tp) ->
       compat_refType p (ArrayType tp) t ->
       assign_compatible p h (Ref loc) (ReferenceType t)
   | assign_compatible_num_val : forall (n:num) (t:primitiveType),
       assign_compatible_num n t -> assign_compatible p h (Num n) (PrimitiveType t).

  Inductive SemCompRef : CompRef -> value -> value -> Prop :=
  | SemCompRef_eq : forall v1 v2,
       isReference v1 -> isReference v2 -> v1 = v2 ->
     (****************************************************)
          SemCompRef EqRef v1 v2
  | SemCompRef_ne : forall v1 v2,
       isReference v1 -> isReference v2 -> v1 <> v2 ->
     (****************************************************)
          SemCompRef NeRef v1 v2.

  Definition SemCompInt (cmp:CompInt) (z1 z2: Z) : Prop :=
    match cmp with
      EqInt =>  z1=z2
    | NeInt => z1<>z2
    | LtInt => z1<z2
    | LeInt => z1<=z2
    | GtInt => z1>z2
    | GeInt => z1>=z2
    end.

  Definition SemBinopInt (op:BinopInt) (i1 i2:Int.t) : Int.t :=
    match op with 
    | AddInt => Int.add i1 i2
    | AndInt => Int.and i1 i2
    | DivInt => Int.div i1 i2
    | MulInt => Int.mul i1 i2
    | OrInt => Int.or i1 i2
    | RemInt => Int.rem i1 i2
    | ShlInt => Int.shl i1 i2
    | ShrInt => Int.shr i1 i2
    | SubInt => Int.sub i1 i2
    | UshrInt => Int.ushr i1 i2
    | XorInt => Int.xor i1 i2
    end.

  Inductive handler_catch (p:Program) : ExceptionHandler -> PC -> ClassName -> Prop :=
  (* The current handler catches all the exceptions in its range *)
  | handler_catch_all : forall pc ex e,
    EXCEPTIONHANDLER.catchType ex = None ->
    EXCEPTIONHANDLER.isPCinRange ex pc = true ->
    handler_catch p ex pc e

  (* The current handler specifies the type of exception it catches *)
  | handler_catch_one : forall pc ex cl1 cl2,
    EXCEPTIONHANDLER.catchType ex = Some cl1 ->
    EXCEPTIONHANDLER.isPCinRange ex pc = true ->
    subclass_name p cl2 cl1 ->
    handler_catch p ex pc cl2.


  (** Lookup in the given list of exception handlers if one of them catches
      the current exception *)

  Inductive lookup_handlers (p:Program) : list ExceptionHandler -> PC -> ClassName -> PC -> Prop :=
  (* The current handler catches the exception *)
  | lookup_handlers_hd_catchAll : forall pc ex exl e,
    handler_catch p ex pc e ->
    lookup_handlers p (ex::exl) pc e (EXCEPTIONHANDLER.handler ex)

  (* Continue with the next handler *)
  | lookup_handlers_tl : forall ex exl pc e pc',
    ~ handler_catch p ex pc e ->
    lookup_handlers p exl pc e pc' ->
    lookup_handlers p (ex::exl) pc e pc'.

  (** Lookup in the callstack if one frame catches the thrown exception *)
  (* If an handler can catch the exception then the control flow is transferred 
     to the beginning of the handler and the exception caught is the only element 
     of the operand stack *)
  (* If lookup in the topmost frame fails, the frame is popped and the lookup 
     continues in the next frame *)
  (* FIXME: Check that the object pointed by loc is an instance of Throwable? - gd *)

  Inductive CaughtException (p:Program) : Method -> PC*Heap.t*Location -> PC -> Prop :=
    CaughtException_def : forall m pc h loc bm pc' e,
      METHOD.body m = Some bm ->
      Heap.typeof h loc = Some (Heap.LocationObject e) ->
      lookup_handlers p (BYTECODEMETHOD.exceptionHandlers bm) pc e pc' ->
      CaughtException p m (pc,h,loc) pc'.

  Inductive UnCaughtException (p:Program) : Method -> PC*Heap.t*Location -> Prop :=
    UnCaughtException_def : forall m pc h loc bm e,
      METHOD.body m = Some bm ->
      Heap.typeof h loc = Some (Heap.LocationObject e) ->
      (forall pc', ~ lookup_handlers p (BYTECODEMETHOD.exceptionHandlers bm) pc e pc') ->
      UnCaughtException p m (pc,h,loc).

End Make.

