Add LoadPath "../Library".
Require Import Sumbool.
Require Export EqBoolAux.
Require Export Misc.

Section AssListT.
  Set Implicit Arguments.
  Variables (A:Set) (B:Type).
  Fixpoint assListLookup (a:A) (f:A->A->bool) (l:listT (A**B)) {struct l}
    : optionT B :=
    match l with
      | NilT => NoneT
      | ConsT ab l' => if f a (fstT ab) then SomeT (sndT ab)
        else assListLookup a f l'
    end.

  Inductive AllListT (P:B -> Prop) : listT (A**B) -> Prop :=
    | alT_nil : AllListT P NilT
    | alT_cons: forall (ab:A**B) (l:listT (A**B)),
      P (sndT ab) -> AllListT P l -> AllListT P (ConsT ab l).
  Unset Implicit Arguments.
End AssListT.


Module Type ASSLIST.
    (* Index set must be decidable to work as intended. *)
  Record IX : Type := {C:Set; decf:C->C->bool}.
  Parameter t : IX -> Type -> Type.
  Section Index.
    Variables (I:IX) (B:Type).
    Parameter lookup : t I B -> C I -> optionT B.
    Parameter delete : t I B -> C I -> t I B.
    Parameter upd: t I B -> C I -> B -> t I B.
    Parameter mty: t I B.
    Parameter isMty: t I B -> bool.

      (* listify proves that the mapping is finite, and allows
	 recursion over all elements of a map 
      *)
    Parameter listify : t I B -> listT (C I ** B).

    Parameter listify_correct : 
      forall (m:t I B) (x:C I),
        assListLookup x (decf I) (listify m) = lookup m x.
    Parameter listify_complete :
      forall (m:t I B) (x:C I) e,
        lookup m x = SomeT e -> inT (x,,e) (listify m).

    Parameter allAssList :
      forall (m:t I B) (P:B->Prop), AllListT P (listify m) ->
	forall (x:C I) (b:B), lookup m x = SomeT b -> P b.
  End Index.
End ASSLIST.

  (* make the computations visible using "<:" *)
Module AssList <: ASSLIST.
  Record IX : Type := {C:Set; decf:C->C->bool}.
  Section Index.
    Variables (ix:IX) (B:Type).
    Definition t : Type := listT (C ix ** B).
    Fixpoint lookup (al:t) (a:C ix) {struct al} : optionT B :=
      match al with
	| NilT => NoneT 
	| ConsT a' l' => 
	  match decf ix a (fstT a') with 
	    | true => SomeT (sndT a')
	    | false => lookup l' a
	  end
      end.
    Parameter delete : t -> C ix -> t.
    Definition upd (al:t) (a:C ix) (b:B) : t := ConsT (a,,b) al.
    Definition mty : t := @NilT (C ix ** B).
    Definition isMty (al:t) : bool :=
      match al with
	| NilT => true
	| ConsT _ _ => false
      end.
    Definition listify (al:t) : listT (C ix ** B) := al.

    Lemma IX_decidable:
      forall (ix:IX) (c d:C ix), {decf ix c d = true}+{decf ix c d = false}.
    Proof.
      intros. apply (sumbool_of_bool (decf ix0 c d)).
    Qed.

    Lemma listify_correct : 
      forall (m:t) (x:C ix),
        assListLookup x (decf ix) (listify m) = lookup m x.
    Proof.
      induction m.
    (* NilT *)
      simpl. intros; trivial.
    (* ConsT *)
      intro x.
      change (listify (ConsT a m)) with (ConsT a (listify m)).
      assert (h: forall y, y = decf _ x (fstT a) ->
        assListLookup x (decf ix) (ConsT a (listify m)) = 
        lookup (ConsT a m) x).
      induction y; intro yy.
      unfold assListLookup, lookup. rewrite <- yy. trivial.
      unfold assListLookup, lookup. rewrite <- yy.
      change (assListLookup x (decf ix) (listify m) = lookup m x).
      apply IHm.

      apply (h (decf ix x (fstT a))); trivial.
    Qed.

    Parameter listify_complete :
      forall (m:t) (x:C ix) e,
        lookup m x = SomeT e -> inT (x,,e) (listify m).

    Lemma listifyNil_None:
      forall (m:t), listify m = NilT -> forall x, lookup m x = NoneT.
    Proof.
      intros. rewrite <- (listify_correct m x). rewrite H. trivial.
    Qed.

    Lemma allAssList :
      forall (m:t) (P:B->Prop),  AllListT P (listify m) ->
	forall (x:C ix) (b:B), lookup m x = SomeT b -> P b.
    Proof.
      unfold t, listify. induction m; intros.
  (* NilT *)
      discriminate H0.
  (* ConsT *)
      inversion H; clear H.
      unfold lookup in H0.
      case (IX_decidable ix x (fstT a)); intro hdecf; rewrite hdecf in H0.
  (* head *)
      injection H0; intro k; rewrite <- k; trivial.
  (* tail *)
      fold lookup in H0. apply (IHm P H4 _ _ H0).
    Qed.

  End Index.
End AssList.
