  Export V5 Dom Prog.

  Definition InitState  := LocalVar.t * Heap.t.
  Definition LocalState := Heap.t * OperandStack.t * LocalVar.t.
 
  (* Local specification table *)
  Definition Assertion := InitState -> LocalState -> Prop.
  Definition LocalPost := InitState -> LocalState -> ReturnState -> Prop.
  Definition StrongInv := LocalState -> LocalState -> Prop.

  Definition LocalSpec      := Assertion ** LocalPost ** StrongInv.
  Definition LocalSpecTable := PCM.t LocalSpec.

  (* Local annotation table *)
  Definition LocalAnnoTable := PCM.t Assertion.

  (* Method specification table *)


  Definition MethPre  := InitState -> Prop.
  Definition MethPost := InitState -> ReturnState -> Prop.
  Definition MethInv  := InitState -> LocalState -> Prop.

  Definition MethSpec := MethPre ** MethPost ** MethInv.

  Definition MethSpecTable := 
     MM.t (MethSpec ** LocalSpecTable ** LocalAnnoTable).
  
  Definition mkMethSpec (pre:MethPre) (post:MethPost) s0 r := 
    pre s0 -> post s0 r.

  Inductive kindOfInstr : Set :=
   | BasicInstr
   | CallInstr
   | ReturnInstr
   | IfInstr (o:OFFSET.t)
   | GotoInstr (o:OFFSET.t)
   | LSwitchInstr (def:OFFSET.t) (l:list (Z*OFFSET.t)) 
   | TSwitchInstr (def:OFFSET.t) (low high:Z) (loffset:list OFFSET.t).

   Definition kind_of_instr (ins:Instruction) : kindOfInstr := 
   match ins with
   | Aconst_null         => BasicInstr
   | Arraylength         => BasicInstr
   | Athrow              => BasicInstr
   | Checkcast _         => BasicInstr
   | Const _ _           => BasicInstr
   | Dup                 => BasicInstr
   | Dup_x1              => BasicInstr
   | Dup_x2              => BasicInstr
   | Dup2                => BasicInstr
   | Dup2_x1             => BasicInstr
   | Dup2_x2             => BasicInstr
   | Getfield _          => BasicInstr
   | Getstatic _         => BasicInstr
   | Goto o              => GotoInstr o
   | I2b                 => BasicInstr
   | I2s                 => BasicInstr
   | Ibinop _            => BasicInstr
   | If_acmp _ o         => IfInstr o
   | If_icmp _ o         => IfInstr o 
   | If0 _ o             => IfInstr o
   | Ifnull _ o          => IfInstr o
   | Iinc _ _            => BasicInstr
   | Ineg                => BasicInstr
   | Instanceof _        => BasicInstr
   | Invokeinterface _   => CallInstr
   | Invokespecial _     => CallInstr
   | Invokestatic _      => CallInstr
   | Invokevirtual _     => CallInstr
   | Lookupswitch def listkey => LSwitchInstr def listkey
(*   | Multianewarray (t:refType) (d:Z) | New (cl:ClassName) *)
   | New _               => BasicInstr
   | Newarray _          => BasicInstr
   | Nop                 => BasicInstr
   | Pop                 => BasicInstr
   | Pop2                => BasicInstr
   | Putfield _          => BasicInstr
   | Putstatic _         => BasicInstr
   | Return              => ReturnInstr
   | Swap                => BasicInstr
   | Tableswitch def low high loff => TSwitchInstr def low high loff
   | Vaload _            => BasicInstr
   | Vastore _           => BasicInstr
   | Vload _ _           => BasicInstr
   | Vreturn _           => ReturnInstr
   | Vstore _ _          => BasicInstr
   end.

  Definition isBasicInstr (M:Method) (l:PC) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | BasicInstr => True
      | _ => False
      end
    | None => False
    end.

  Definition isCallInstr (M:Method) (l:PC) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | CallInstr => True
      | _ => False
      end
    | None => False
    end.
  

  Definition isReturnInstr (M:Method) (l:PC) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | ReturnInstr => True
      | _ => False
      end
    | None => False
    end.

  Definition isIfInstr (M:Method) (l:PC) (o:OFFSET.t) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | IfInstr o' => o = o'
      | _ => False
      end
    | None => False
    end.

  Definition isGotoInstr (M:Method) (l:PC) (o:OFFSET.t) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | GotoInstr o' => o = o' 
      | _ => False
      end
    | None => False
    end.

  Definition isLSwitchInstr (def:OFFSET.t) (listkey:list (Z*OFFSET.t)) (M:Method) (l:PC) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | LSwitchInstr d lk => ins = Lookupswitch d lk /\ (d=def) /\ (lk = listkey)
      | _ => False
      end
    | None => False
    end.

  Definition isTSwitchInstr (def:OFFSET.t) (low high:Z) (list_offset:list OFFSET.t) (M:Method) (l:PC) : Prop :=
    match instructionAt M l with
    | Some ins => 
      match kind_of_instr ins with
      | TSwitchInstr d lo hi loff => ins = Tableswitch d lo hi loff /\ (d=def) /\ (lo = low) /\ (hi = high) /\ (loff = list_offset)
      | _ => False
      end
    | None => False
    end.

  Inductive SP_pre (P:Program) (M:Method) (l l':PC) (A:Assertion) 
                   (s0:InitState) (t:LocalState) : Prop :=
   | SP_pre_def: forall s:LocalState,
        A s0 s ->
        NormalStep P M (l,s) (l',t) ->
       (**********************************)
        SP_pre P M l l' A s0 t.

  
  Definition SP_post (P:Program) (M:Method) (l l':PC) (B:LocalPost)
                    (s0:InitState) (r:LocalState) (t:ReturnState) : Prop :=
     forall s:LocalState, 
        NormalStep P M (l,s) (l',r) -> B s0 s t.

  Definition SP_inv (P:Program) (M:Method) (l l' :PC) (I:StrongInv)
                   (r:LocalState) (t:LocalState) : Prop:=
     forall s:LocalState, 
        NormalStep P M (l,s) (l',r) -> I s t.

 
  (* Case of execption *)
  Inductive SP_preEC (P:Program) (M:Method) (l:PC) (e:ClassName)
                     (A:Assertion) (s0:InitState) (t:LocalState) : Prop :=
   | SP_preEC_def:  forall  (ops:OperandStack.t)(s:LocalVar.t)(h:Heap.t),
                  A s0 (h,ops,s) ->   
                  forall (h2:Heap.t) loc2,
                  ExceptionStep P M (l,(h,ops,s)) (h2,loc2) -> 
                  Heap.typeof h2 loc2 = Some (Heap.LocationObject e) ->
                  t = (h2,Ref loc2::OperandStack.empty, s) ->
     (*************************************************)
                  SP_preEC P M l e A s0 t.

  
  Inductive SP_postEC (P:Program) (M:Method) (l:PC) (e:ClassName)
                      (B:LocalPost) (s0:InitState) (r:LocalState)
                      (t:ReturnState) : Prop :=
   | SP_postEC_def: forall (h2:Heap.t) loc2 (s:LocalVar.t),
              (forall  (ops:OperandStack.t) (h:Heap.t),
                  ExceptionStep P M (l,(h,ops,s)) (h2,loc2) -> 
                  Heap.typeof h2 loc2 = Some (Heap.LocationObject e) ->
                  B s0 (h,ops,s) t) -> 
                  r = (h2,Ref loc2::OperandStack.empty, s) ->
     (*************************************************)
                  SP_postEC P M l e B s0 r t.


  Inductive SP_invEC (P:Program) (M:Method) (l:PC) (e:ClassName)
                     (I:StrongInv) (r:LocalState) (t:LocalState) : Prop:=
   | SP_invEC_def:  forall (h2:Heap.t) loc2  (s:LocalVar.t),
		  r = (h2,Ref loc2::OperandStack.empty, s) ->
                  Heap.typeof h2 loc2 = Some (Heap.LocationObject e) ->
                  (forall (ops:OperandStack.t) (h:Heap.t),
                  ExceptionStep P M (l,(h,ops,s)) (h2,loc2) -> 
                  I (h,ops,s) t) ->
     (*************************************************)
                  SP_invEC P M l e I r t.

       
  (*Operators for Static invoke-non-exceptional case, non-void return value).*)

  Inductive SINV_pre (pre:MethPre) (post:MethPost) (A:Assertion) (N:nat) 
                      (s0:InitState) (s:LocalState)  : Prop :=
   SINV_pre_def : forall (ops:OperandStack.t) (S:LocalVar.t) (h:Heap.t) 
                           (H:Heap.t) (args:OperandStack.t) (w:option value),
     length args = N ->
     mkMethSpec pre post (stack2localvar (args++ops) N, H)(h, Normal w) ->
     A s0 (H,args++ops,S) ->
     s = (h, cons_option w ops, S) ->
    (*************************************************)
     SINV_pre pre post A N s0 s.


  Inductive SINV_inv (pre:MethPre) (post:MethPost)  (J:StrongInv) (N:nat) 
                     (s:LocalState) (t:LocalState) : Prop :=
   SINV_inv_def:  
    (forall ops S h H args w, 
       length args = N ->
       mkMethSpec pre post (stack2localvar (args++ops) N,H) (h, Normal w) ->
       s=(h,cons_option w ops,S) -> J (H,args++ops,S) t) ->
     (*************************************************)
      SINV_inv pre post J N s t.


  Inductive SINV_post (pre:MethPre) (post:MethPost) (B:LocalPost) (N:nat) 
                      (s0:InitState) (s:LocalState) (t:ReturnState) : Prop :=
   SINV_post_def: 
    (forall ops S h H args w r,
       length args = N ->
       mkMethSpec pre post (stack2localvar (args++ops) N, H) (h, Normal w) ->
       s=(h, cons_option w ops,S) ->
       r=(H,args++ops,S) ->
       B s0 r t) ->
     (*************************************************)
     SINV_post pre post B N s0 s t.



  Inductive SINV_preEC (pre:MethPre) (post:MethPost) (A:Assertion) (N:nat) 
            (l:PC) (e:ClassName) (s0:InitState) (s:LocalState)  : Prop :=
   SINV_preEC_def : forall (ops:OperandStack.t) (S:LocalVar.t) (h:Heap.t) 
                           (H:Heap.t) (loc:Location)(args:OperandStack.t),
     length args = N ->
     mkMethSpec pre post (stack2localvar (args++ops) N, H) (h, Exception loc)->
     Heap.typeof h loc = Some (Heap.LocationObject e) ->
     A s0 (H,args++ops,S) ->
     s = (h,Ref loc::OperandStack.empty, S) ->
     (*************************************************)
     SINV_preEC pre post A N l e s0 s.



  Inductive SINV_invEC (pre:MethPre) (post:MethPost) (J:StrongInv) (N:nat)  
             (l:PC) (e:ClassName) (s:LocalState) (t:LocalState) : Prop :=
    SINV_invEC_def:  
     (forall ops S h H args loc, 
        length args = N ->
        mkMethSpec pre post (stack2localvar(args++ops) N,H) (h,Exception loc)->
        Heap.typeof h loc = Some (Heap.LocationObject e) ->
        s=(h,Ref loc::OperandStack.empty, S) -> 
        J (H,args++ops,S) t) ->
     (*************************************************)
     SINV_invEC pre post J N l e s t.

 

  Inductive SINV_postEC (pre:MethPre) (post:MethPost) (B:LocalPost) (N:nat) 
       (l:PC) (e:ClassName)(s0:InitState)(s:LocalState)(t:ReturnState): Prop :=
   SINV_postEC_def: 
     (forall ops S h H args r loc,
        length args = N ->
        mkMethSpec pre post 
          (stack2localvar (args++ops) N, H) (h, Exception loc) ->
        Heap.typeof h loc = Some (Heap.LocationObject e) ->
        s=(h,Ref loc::OperandStack.empty, S) -> 
        r=(H,args++ops,S) -> 
        B s0 r t) ->
     (*************************************************)
      SINV_postEC pre post B N l e s0 s t.


  (*Operators for Virtual invoke-non-exceptional case, non-void return value).*)
  Inductive VINV_pre (pre:MethPre) (post:MethPost) (A:Assertion) (N:nat)
                     (s0:InitState) (s:LocalState)  : Prop :=
    VINV_pre_def : forall (ops:OperandStack.t) (S:LocalVar.t) (h:Heap.t) 
                           (H:Heap.t) (args:OperandStack.t) (w:option value) a,
           length args = N ->
           mkMethSpec pre post 
             (stack2localvar (args++(Ref a)::ops) (1+N), H) (h, Normal w) ->
           A s0 (H,args++(Ref a)::ops,S) ->
           s = (h, cons_option w ops, S) ->
     (*************************************************)
           VINV_pre pre post A N s0 s.


  Inductive VINV_inv (pre:MethPre) (post:MethPost) (J:StrongInv) (N:nat)
                     (s:LocalState) (t:LocalState) : Prop :=
   VINV_inv_def:  (forall ops S h H args w a, 
                   length args = N ->
                   mkMethSpec pre post 
                     (stack2localvar (args++(Ref a)::ops) (1+N),H) (h, Normal w) ->
                   s=(h,cons_option w ops,S) -> J (H,args++(Ref a)::ops,S) t) ->
     (*************************************************)
         VINV_inv pre post J N s t.

  Inductive VINV_post (pre:MethPre) (post:MethPost) (B:LocalPost) (N:nat)
                      (s0:InitState) (s:LocalState) (t:ReturnState) : Prop :=
   VINV_post_def: (forall ops S h H args w r a,
                    length args = N ->
                    mkMethSpec pre post 
                      (stack2localvar (args++(Ref a)::ops) (1+N), H) (h, Normal w) ->
                    s=(h, cons_option w ops,S) ->
                    r=(H,args++(Ref a)::ops,S) ->
                    B s0 r t) ->
     (*************************************************)
          VINV_post pre post B N s0 s t.


 Inductive VINV_preEC (pre:MethPre) (post:MethPost) (A:Assertion) (N:nat)
                      (l:PC) (e:ClassName)
                      (s0:InitState) (s:LocalState)  : Prop :=
   | VINV_preEC_def : forall (ops:OperandStack.t) (S:LocalVar.t) (h:Heap.t) 
                           (H:Heap.t) (loc:Location)(args:OperandStack.t) a,
           length args = N ->
           mkMethSpec pre post 
             (stack2localvar (args++(Ref a)::ops) (1+N), H) (h, Exception loc) ->
           Heap.typeof h loc = Some (Heap.LocationObject e) ->
           A s0 (H,args++(Ref a)::ops,S) ->
           s = (h,Ref loc::OperandStack.empty, S) ->
     (*************************************************)
           VINV_preEC pre post A N l e s0 s.


  Inductive VINV_invEC (pre:MethPre) (post:MethPost) (J:StrongInv) (N:nat)
                       (l:PC) (e:ClassName)
                       (s:LocalState) (t:LocalState) : Prop :=
   | VINV_invEC_def:  
       (forall ops S h H args loc a, 
          length args = N ->
          mkMethSpec pre post 
            (stack2localvar (args++(Ref a)::ops) (1+N),H) (h, Exception loc) ->
          Heap.typeof h loc = Some (Heap.LocationObject e) ->
          s=(h,Ref loc::OperandStack.empty, S) -> 
          J (H,args++(Ref a)::ops,S) t) ->
     (*************************************************)
       VINV_invEC pre post J N l e s t.

 
  Inductive VINV_postEC (pre:MethPre) (post:MethPost) (B:LocalPost) (N:nat)
                        (l:PC) (e:ClassName)
                        (s0:InitState) (s:LocalState) (t:ReturnState) : Prop :=
  | VINV_postEC_def: 
       (forall ops S h H args r loc a,
          length args = N ->
          mkMethSpec pre post 
            (stack2localvar (args++(Ref a)::ops) (1+N), H) (h, Exception loc) ->
          Heap.typeof h loc = Some (Heap.LocationObject e) ->
          s=(h,Ref loc::OperandStack.empty, S) -> 
          r=(H,args++(Ref a)::ops,S) -> 
          B s0 r t) ->
     (*************************************************)
       VINV_postEC pre post B N l e s0 s t.

 Inductive VINV_preECNull (P:Program) (A:Assertion) (N:nat)
                          (s0:InitState) (s:LocalState)  : Prop :=
   | VINV_preECNull_def : forall (ops:OperandStack.t) (S:LocalVar.t) (h:Heap.t) 
                           (H:Heap.t) (loc:Location)(args:OperandStack.t),
           length args = N ->
           Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
           A s0 (H,args++Null::ops,S) ->
           s = (h,Ref loc::OperandStack.empty, S) ->
     (*************************************************)
           VINV_preECNull P A N s0 s.

  Inductive VINV_invECNull (P:Program) (J:StrongInv) (N:nat)
                           (s:LocalState) (t:LocalState) : Prop :=
   | VINV_invECNull_def:  
       (forall ops S h H args loc, 
          length args = N ->
           Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
           s=(h,Ref loc::OperandStack.empty, S) -> 
           J (H,args++Null::ops,S) t) ->
     (*************************************************)
       VINV_invECNull P J N s t.

 
  Inductive VINV_postECNull (P:Program) (B:LocalPost) (N:nat) 
                            (s0:InitState) (s:LocalState) (t:ReturnState) : Prop :=
  | VINV_postECNull_def: 
       (forall ops S h H args r loc,
           length args = N ->
           Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
           s=(h,Ref loc::OperandStack.empty, S) -> 
           r=(H,args++Null::ops,S) -> 
           B s0 r t) ->
     (*************************************************)
       VINV_postECNull P B N s0 s t.


  Inductive SP_LSwitch_Def_pre (P:Program) (M:Method) (l l':PC) (A:Assertion) (i:Int.t) 
                   (s0:InitState) (t:LocalState) : Prop :=
   | SP_LSwitch_def_pre_def: forall s:LocalState,
        A s0 s ->
        (exists h, exists ss, exists lv, s=(h,Num(I i) :: ss, lv)) ->
        NormalStep P M (l,s) (l',t) ->
       (**********************************)
        SP_LSwitch_Def_pre P M l l' A i s0 t.

  
  Definition SP_LSwitch_Def_post (P:Program) (M:Method) (l l':PC) (B:LocalPost)(i:Int.t) 
                    (s0:InitState) (r:LocalState) (t:ReturnState)  : Prop :=
     forall s:LocalState, 
        (exists h, exists ss, exists lv, s=(h,Num(I i) :: ss, lv)) ->
        NormalStep P M (l,s) (l',r) -> B s0 s t.

  Definition SP_LSwitch_Def_inv (P:Program) (M:Method) (l l' :PC) (I0:StrongInv)(i:Int.t) 
                   (r:LocalState) (t:LocalState) : Prop:=
     forall s:LocalState, 
        (exists h, exists ss, exists lv, s=(h,Num(I i) :: ss, lv)) ->
        NormalStep P M (l,s) (l',r) -> I0 s t.

 

 Section SP_Judgement.
  Variable P   : Program.
  Variable MST : MethSpecTable.


  Section METHOD.

  Variable M: Method. 
  Variable LocSpecTab : LocalSpecTable.
  Variable LocAnnTab : LocalAnnoTable.

  Definition Anno_LOOKUP pc := PCM.get LocAnnTab pc.
  Definition ASSUM_LOOKUP pc := PCM.get LocSpecTab pc.
  Definition MSPEC_LOOKUP m := MM.get MST m.

  Inductive SP_Judgement :
     bool -> PC -> Assertion -> LocalPost -> StrongInv -> Prop :=

  | SP_INSTR : forall  (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv),

      isBasicInstr M l ->
      (* Local Invariant is satisfied *)
      (forall Q , Anno_LOOKUP l = Some Q -> 
                 (forall s0 s , A s0 s -> Q s0 s)) ->
      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Next instruction is valid, 3 cases : *)
         (* 1 - Normal execution, SP_pre, SP_post, SP_inv use
            NormalStep who does not contain the case of
            exception *)
         (forall ll, next M l = Some ll -> 
              SP_Judgement true ll (SP_pre P M l ll A)
                                   (SP_post P M l ll B)
                                   (SP_inv P M l ll I)) ->
   
         (* 2 - Exception caught by the handler *)
         (forall ll e bm, 
              METHOD.body M = Some bm ->
              lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                             l e ll ->
              SP_Judgement true ll (SP_preEC P M l e A)
                                    (SP_postEC P M l e B)
                                    (SP_invEC P M l e I)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 s, A s0 s ->
            forall h2 loc2,
               ExceptionStep P M (l,s) (h2,loc2) ->
               UnCaughtException  P M (l,h2,loc2) ->
               B s0 s (h2,Exception loc2)) ->
     (********************************************************************)
      SP_Judgement false l A B I

   | SP_LOOKUPSWITCH : forall  (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv)
                                                       (def:OFFSET.t) (listkey:list (Z*OFFSET.t)) ,

        isLSwitchInstr def listkey M l ->

        (* Local Invariant is satisfied *)
        (forall Q , Anno_LOOKUP l = Some Q -> 
                   (forall s0 s , A s0 s -> Q s0 s)) ->
        (* Strong Invariant is satisfied *)
        (forall s0 s, A s0 s -> I s s) ->

        (forall o', (exists i', exists i, List.In (pair i' o')listkey /\ i' = Int.toZ i) ->
              SP_Judgement true (OFFSET.jump l o')
                                   (SP_pre P M l (OFFSET.jump l o') A)
                                   (SP_post P M l (OFFSET.jump l o') B)
                                   (SP_inv P M l (OFFSET.jump l o') I)) ->

         (forall i, (forall i' o', List.In (pair i' o')listkey -> i' <> Int.toZ i) ->
              SP_Judgement true (OFFSET.jump l def)
                                   (SP_LSwitch_Def_pre P M l (OFFSET.jump l def) A i)
                                   (SP_LSwitch_Def_post P M l (OFFSET.jump l def) B i)
                                   (SP_LSwitch_Def_inv P M l (OFFSET.jump l def) I i))->
     (********************************************************************)
      SP_Judgement false l A B I
(*
  | tableswitch1 : forall h m pc s l i def low high list_offset,

    instructionAt m pc = Some (Tableswitch def low high list_offset) ->
    Z_of_nat (length list_offset) = (high - low + 1)%Z ->
    Int.toZ i < low \/ high < Int.toZ i ->
   
   NormalStep p m (pc,(h,(Num (I i)::s),l)) (OFFSET.jump pc def,(h,s,l))

  | tableswitch2 : forall h m pc s l n o i def low high list_offset,

    instructionAt m pc = Some (Tableswitch def low high list_offset) ->
    Z_of_nat (length list_offset) = (high - low + 1)%Z ->
    low <= Int.toZ i <= high ->
    Z_of_nat n = (Int.toZ i) - low ->
    nth_error list_offset n = Some o ->
   
   NormalStep p m (pc,(h,(Num (I i)::s),l)) (OFFSET.jump pc o,(h,s,l))
*)
   | SP_TABLESWITCH : forall  (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv) 
                                                              (def:OFFSET.t) (low high:Z) (list_offset:list OFFSET.t) ,

        isTSwitchInstr def low high list_offset M l->

        (* Local Invariant is satisfied *)
        (forall Q , Anno_LOOKUP l = Some Q -> 
                   (forall s0 s , A s0 s -> Q s0 s)) ->
        (* Strong Invariant is satisfied *)
        (forall s0 s, A s0 s -> I s s) ->

         
           ((exists i, Z_of_nat (length list_offset) = (high - low + 1)%Z /\
                             (Int.toZ i < low \/ high < Int.toZ i)) ->
              SP_Judgement true (OFFSET.jump l def)
                                   (SP_pre P M l (OFFSET.jump l def) A)
                                   (SP_post P M l (OFFSET.jump l def) B)
                                   (SP_inv P M l (OFFSET.jump l def) I)) ->

              (forall o, (exists i, exists n, Z_of_nat (length list_offset) = (high - low + 1)%Z /\
                                              low <= Int.toZ i <= high /\
                                              Z_of_nat n = (Int.toZ i) - low /\
                                             nth_error list_offset n = Some o) ->
                              SP_Judgement true (OFFSET.jump l o)
                                               (SP_pre P M l (OFFSET.jump l o) A)
                                               (SP_post P M l (OFFSET.jump l o) B)
                                              (SP_inv P M l (OFFSET.jump l o) I))->
     (********************************************************************)
      SP_Judgement false l A B I

  | SP_RETURN : forall (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv),
      isReturnInstr M l ->
      (* Local Invariant is satisfied *)
      (forall Q, Anno_LOOKUP l = Some Q -> 
                 (forall s0 s , A s0 s -> Q s0 s)) ->
      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Local Post is satisfied *)
      (forall s0 s, A s0 s ->
         forall h2 ov, ReturnStep P M (l,s) (h2,Normal ov) ->
         B s0 s (h2,Normal ov)) ->
      (********************************************************************)
      SP_Judgement false l A B I

   | SP_GOTO : forall (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv) 
                    (o:OFFSET.t),   
      instructionAt M l = Some (Goto o) ->
    
      (* Local Assertion is satisfied *) 
      (forall Q, Anno_LOOKUP l = Some Q ->
                  (forall s0 s , A s0 s -> Q s0 s)) ->

      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Local Post is satisfied *)
      SP_Judgement true (OFFSET.jump l o)
                        (SP_pre P M l (OFFSET.jump l o) A)
                        (SP_post P M l (OFFSET.jump l o) B)
                        (SP_inv P M l (OFFSET.jump l o) I) ->
    (********************************************************************)
      SP_Judgement false l A B I

  | SP_IF : forall (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv)
                 (ll:PC)  (o:OFFSET.t),
      isIfInstr M l o ->
      next M l = Some ll ->
      (forall Q , Anno_LOOKUP l = Some Q ->
                  (forall s0 s , A s0 s -> Q s0 s)) ->
      (forall s0 s, A s0 s -> I s s) ->

      SP_Judgement true (OFFSET.jump l o)
        (SP_pre P M l (OFFSET.jump l o) A)
        (SP_post P M l (OFFSET.jump l o) B)
        (SP_inv P M l (OFFSET.jump l o) I) -> 
      SP_Judgement true ll
        (SP_pre P M l ll A)
        (SP_post P M l ll B)
        (SP_inv P M l ll I) ->
    (********************************************************************)
      SP_Judgement false l A B I

  | SP_INVS: forall (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv) 
                 (msig:MethodSignature) (m:Method)
                 (mpre:MethPre) (mpost:MethPost) (MI:MethInv) 
                 (LS:LocalSpecTable) (LA:LocalAnnoTable)
                 (N:nat),

     instructionAt M l = Some(Invokestatic msig) ->
     findMethod P msig = Some m ->
     METHOD.isNative m = false ->
     MSPEC_LOOKUP msig = Some ((((mpre,,mpost),,MI),,LS),,LA) ->
     (forall Q , Anno_LOOKUP l = Some Q ->
                 (forall s0 s , A s0 s -> Q s0 s)) ->
     (forall s0 s, A s0 s -> I s s) ->
     N = length (METHODSIGNATURE.parameters (snd msig)) ->
     (forall s0 args ops S H t ,
          length args = N ->
           A s0 (H,args++ops,S) ->
           (mpre (stack2localvar (args++ops) N,H) ->
                MI (stack2localvar (args++ops) N,H) t) -> 
           I (H,args++ops,S) t) -> 
     (* Evaluation of the callee does not raise exception*)
     (forall ll, next M l = Some ll ->
           SP_Judgement true ll (SINV_pre mpre mpost A N)
                                (SINV_post mpre mpost B N)
                                (SINV_inv mpre mpost I N)) ->
     (* Evaluation of the callee raise an exception which is caught*)
     (forall ll e bm, 
              METHOD.body M = Some bm ->
              lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l e ll ->
              SP_Judgement true ll (SINV_preEC mpre mpost A N l e)
                                    (SINV_postEC mpre mpost B N l e)
                                    (SINV_invEC mpre mpost I N l e)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 args ops S H, A s0 (H,args++ops,S) ->
            forall h loc,
              length args = N ->
              mkMethSpec mpre mpost
                (stack2localvar (args++ops) N,H) (h, Exception loc) ->
              UnCaughtException  P M (l,h,loc) ->
              B s0  (H,args++ops,S) (h,Exception loc)) ->
     (*******************************************************************)
      SP_Judgement false l A B I

(*********************************************************************)
(***********************INVOKEVIRTUAL**********************************)

  | SP_INVV: forall (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv) 
                 (msig:MethodSignature) (m:Method)
                 (mpre:MethPre) (mpost:MethPost) (MI:MethInv) 
                 (LS:LocalSpecTable) (LA:LocalAnnoTable)
                 (N:nat),

     instructionAt M l = Some(Invokevirtual msig) ->
     findMethod P msig = Some m ->
     METHOD.isNative m = false ->
     MSPEC_LOOKUP msig = Some ((((mpre,,mpost),,MI),,LS),,LA) ->
     (forall Q , Anno_LOOKUP l = Some Q ->
                 (forall s0 s , A s0 s -> Q s0 s)) ->
     (forall s0 s, A s0 s -> I s s) ->
     N = length (METHODSIGNATURE.parameters (snd msig)) ->
     (forall s0 args ops S H t a, length args = N ->
                                 A s0 (H,args++(Ref a)::ops,S) ->
                                 (mpre (stack2localvar (args++(Ref a)::ops) (1+N),H) ->
                                    MI (stack2localvar (args++(Ref a)::ops) (1+N),H) t) -> 
                                 I (H,args++(Ref a)::ops,S) t) -> 
     (* Evaluation of the callee does not raise exception*)
     (forall ll, next M l = Some ll ->
           SP_Judgement true ll (VINV_pre mpre mpost A N)
                                (VINV_post mpre mpost B N)
                                (VINV_inv mpre mpost I N)) ->
     (* Evaluation of the callee raise an exception which is caught*)
     (forall ll e bm, 
              METHOD.body M = Some bm ->
              lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l e ll ->
              SP_Judgement true ll (VINV_preEC mpre mpost A N l e)
                                    (VINV_postEC mpre mpost B N l e)
                                    (VINV_invEC mpre mpost I N l e)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 args ops S H a, A s0 (H,args++(Ref a)::ops,S) ->
            forall h loc,
              length args = N ->
              mkMethSpec mpre mpost 
                (stack2localvar (args++(Ref a)::ops) (1+N),H) (h, Exception loc) ->
              UnCaughtException  P M (l,h,loc) ->
              B s0  (H,args++(Ref a)::ops,S) (h,Exception loc)) ->

     (*Nullpointerexception*)
         (forall s0 args ops S H , A s0 (H,args++Null::ops,S) ->
            forall h loc,
              length args = N -> 
              Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
              ( (forall ll bm, METHOD.body M = Some bm ->
                               lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l (javaLang,NullPointerException) ll ->
                               SP_Judgement true ll (VINV_preECNull P A N)
                                                    (VINV_postECNull P B N)
                                                    (VINV_invECNull P I N)) /\
                (UnCaughtException P M (l,h,loc) ->
                 B s0 (H,args++Null::ops,S) (h,Exception loc)))) ->
           
     (*******************************************************************)
      SP_Judgement false l A B I

(*********************************END OF INVOKEVIRTUAL******************)
(*********************************************************************)

  | SP_CONSEQ : forall (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv) 
                    (b:bool) (AA:Assertion) (BB:LocalPost) (II:StrongInv) ,
      SP_Judgement b l AA BB II ->
      (forall s0 s, A s0 s -> AA s0 s) ->
      (forall s0 s t, A s0 s -> BB s0 s t -> B s0 s t) ->
      (forall t s , II t s -> I t s) ->
  (********************************************************************)
      SP_Judgement b l A B I 

  | SP_INJECT : forall (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv),
      SP_Judgement false l A B I ->
  (********************************************************************)
      SP_Judgement true l A B I

  | SP_AX : forall (l:PC) (A:Assertion) (B:LocalPost) (I:StrongInv), 
      ASSUM_LOOKUP l = Some ((A ,, B) ,, I) ->
      (forall Q, Anno_LOOKUP l = Some Q ->
         forall s0 s, A s0 s -> Q s0 s) ->
      (forall s0 s, A s0 s -> I s s) ->
  (********************************************************************)
      SP_Judgement true l A B I.



 Inductive SP_Judgement_call (l:PC) (msig:MethodSignature)(AA:Assertion)
                             (BB:LocalPost)(II:StrongInv): Prop :=
 | SP_INVS_call: forall (A:Assertion) (B:LocalPost) (I:StrongInv) 
                  (m:Method) 
                  (mpre:MethPre) (mpost:MethPost) (MI:MethInv) 
                  (LS:LocalSpecTable) (LA:LocalAnnoTable)(N:nat),

     findMethod P msig = Some m ->
     METHOD.isNative m = false ->
     MSPEC_LOOKUP msig =  Some ((((mpre,,mpost),,MI),,LS),,LA) ->
     (forall Q , Anno_LOOKUP l = Some Q ->
                 (forall s0 s , A s0 s -> Q s0 s)) ->
     (forall s0 s, A s0 s -> I s s) ->
     N = length (METHODSIGNATURE.parameters (snd msig)) ->
     (forall s0 args ops S H t , 
         length args = N ->
         A s0 (H,args++ops,S) ->
         (mpre (stack2localvar (args++ops) N,H) ->
            MI (stack2localvar (args++ops) N,H) t) -> 
         I (H,args++ops,S) t) -> 
     (* Evaluation of the callee does not raise exception*)
     (forall ll, next M l = Some ll ->
              SP_Judgement true ll (SINV_pre mpre mpost A N)
                                   (SINV_post mpre mpost B N)
                                   (SINV_inv mpre mpost I N)) ->
     (* Evaluation of the callee raise an exception which is caught*)
     (forall ll e bm, 
              METHOD.body M = Some bm ->
              lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                       l e ll ->
              SP_Judgement true ll (SINV_preEC mpre mpost A N l e)
                                    (SINV_postEC mpre mpost B N l e)
                                    (SINV_invEC mpre mpost I N l e)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 args ops S H, A s0 (H,args++ops,S) ->
            forall h loc,
                length args = N ->
                mkMethSpec mpre mpost
                   (stack2localvar (args++ops) N,H) (h, Exception loc) ->
                UnCaughtException  P M (l,h,loc) ->
                B s0  (H,args++ops,S) (h,Exception loc)) ->
     (*******************************************************************)
      (forall s0 s, AA s0 s -> A s0 s) ->
      (forall s0 s t, AA s0 s -> B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
    SP_Judgement_call l msig AA BB II.

 Inductive SP_Judgement_virtualcall (l:PC) (msig:MethodSignature)(AA:Assertion)
                             (BB:LocalPost)(II:StrongInv): Prop :=
  | SP_INVV_call: forall  (A:Assertion) (B:LocalPost) (I:StrongInv) 
                  (m:Method) 
                  (mpre:MethPre) (mpost:MethPost) (MI:MethInv) 
                  (LS:LocalSpecTable) (LA:LocalAnnoTable)(N:nat),

     findMethod P msig = Some m ->
     METHOD.isNative m = false ->
     MSPEC_LOOKUP msig = Some ((((mpre,,mpost),,MI),,LS),,LA) ->
     (forall Q , Anno_LOOKUP l = Some Q ->
                 (forall s0 s , A s0 s -> Q s0 s)) ->
     (forall s0 s, A s0 s -> I s s) ->
     N = length (METHODSIGNATURE.parameters (snd msig)) ->
     (forall s0 args ops S H t a, length args = N ->
                                 A s0 (H,args++(Ref a)::ops,S) ->
                                 (mpre (stack2localvar (args++(Ref a)::ops) (1+N),H) ->
                                    MI (stack2localvar (args++(Ref a)::ops) (1+N),H) t) -> 
                                 I (H,args++(Ref a)::ops,S) t) -> 
     (* Evaluation of the callee does not raise exception*)
     (forall ll, next M l = Some ll ->
           SP_Judgement true ll (VINV_pre mpre mpost A N)
                                (VINV_post mpre mpost B N)
                                (VINV_inv mpre mpost I N)) ->
     (* Evaluation of the callee raise an exception which is caught*)
     (forall ll e bm, 
              METHOD.body M = Some bm ->
              lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l e ll ->
              SP_Judgement true ll (VINV_preEC mpre mpost A N l e)
                                    (VINV_postEC mpre mpost B N l e)
                                    (VINV_invEC mpre mpost I N l e)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 args ops S H a, A s0 (H,args++(Ref a)::ops,S) ->
            forall h loc,
              length args = N ->
              mkMethSpec mpre mpost 
                (stack2localvar (args++(Ref a)::ops) (1+N),H) (h, Exception loc) ->
              UnCaughtException  P M (l,h,loc) ->
              B s0  (H,args++(Ref a)::ops,S) (h,Exception loc)) ->

     (*Nullpointerexception*)
         (forall s0 args ops S H , A s0 (H,args++Null::ops,S) ->
            forall h loc,
              length args = N -> 
              Heap.new H P (Heap.LocationObject (javaLang,NullPointerException)) = Some (loc,h) ->
              ( (forall ll bm, METHOD.body M = Some bm ->
                               lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                              l (javaLang,NullPointerException) ll ->
                               SP_Judgement true ll (VINV_preECNull P A N)
                                                    (VINV_postECNull P B N)
                                                    (VINV_invECNull P I N)) /\
                (UnCaughtException P M (l,h,loc) ->
                 B s0 (H,args++Null::ops,S) (h,Exception loc)))) ->

     (*******************************************************************)
      (forall s0 s, AA s0 s -> A s0 s) ->
      (forall s0 s t, AA s0 s -> B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
    SP_Judgement_virtualcall l msig AA BB II.

 Inductive SP_Judgement_basic (l:PC) (AA:Assertion) (BB:LocalPost) (II:StrongInv) : Prop :=
  | SP_INSTR_basic : forall  (A:Assertion) (B:LocalPost) (I:StrongInv),

      (* Local Invariant is satisfied *)
      (forall Q , Anno_LOOKUP l = Some Q -> 
                 (forall s0 s , A s0 s -> Q s0 s)) ->
      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Next instruction is valid, 3 cases : *)
         (* 1 - Normal execution, SP_pre, SP_post, SP_inv use
            NormalStep who does not contain the case of
            exception *)
         (forall ll, next M l = Some ll -> 
              SP_Judgement true ll (SP_pre P M l ll A)
                                   (SP_post P M l ll B)
                                   (SP_inv P M l ll I)) ->
   
         (* 2 - Exception caught by the handler *)
         (forall ll e bm, 
              METHOD.body M = Some bm ->
              lookup_handlers P (BYTECODEMETHOD.exceptionHandlers bm)
                                                 l e ll ->
              SP_Judgement true ll (SP_preEC P M l e A)
                                    (SP_postEC P M l e B)
                                    (SP_invEC P M l e I)) ->
         (* - Uncaught exception (ie return) *) 
         (forall s0 s, A s0 s ->
            forall h2 loc2,
               ExceptionStep P M (l,s) (h2,loc2) ->
               UnCaughtException  P M (l,h2,loc2) ->
               B s0 s (h2,Exception loc2)) ->
     (********************************************************************)
      (forall s0 s, AA s0 s -> A s0 s) ->
      (forall s0 s t, AA s0 s -> B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
      SP_Judgement_basic l AA BB II.

 
 Inductive SP_Judgement_lookupswitch (def:OFFSET.t) (listkey:list (Z*OFFSET.t)) 
        (l:PC) (AA:Assertion) (BB:LocalPost) (II:StrongInv) : Prop :=
  | SP_LOOKUPSWITCH_switch: forall  (A:Assertion) (B:LocalPost) (I:StrongInv),

        (* Local Invariant is satisfied *)
        (forall Q , Anno_LOOKUP l = Some Q -> 
                   (forall s0 s , A s0 s -> Q s0 s)) ->
        (* Strong Invariant is satisfied *)
        (forall s0 s, A s0 s -> I s s) ->


        (forall o', (exists i', exists i, List.In (pair i' o')listkey /\ i' = Int.toZ i) ->
              SP_Judgement true (OFFSET.jump l o')
                                   (SP_pre P M l (OFFSET.jump l o') A)
                                   (SP_post P M l (OFFSET.jump l o') B)
                                   (SP_inv P M l (OFFSET.jump l o') I)) ->

         (forall i, (forall i' o', List.In (pair i' o')listkey -> i' <> Int.toZ i) ->
              SP_Judgement true (OFFSET.jump l def)
                                   (SP_LSwitch_Def_pre P M l (OFFSET.jump l def) A i)
                                   (SP_LSwitch_Def_post P M l (OFFSET.jump l def) B i)
                                   (SP_LSwitch_Def_inv P M l (OFFSET.jump l def) I i))->
     (********************************************************************)
      (forall s0 s, AA s0 s -> A s0 s) ->
      (forall s0 s t, AA s0 s -> B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
      SP_Judgement_lookupswitch def listkey l AA BB II.

Inductive SP_Judgement_tableswitch (def:OFFSET.t) (low high :Z) (list_offset:list OFFSET.t)
        (l:PC) (AA:Assertion) (BB:LocalPost) (II:StrongInv) : Prop :=

   | SP_TABLESWITCH_switch : forall  (A:Assertion) (B:LocalPost) (I:StrongInv) ,

        (* Local Invariant is satisfied *)
        (forall Q , Anno_LOOKUP l = Some Q -> 
                   (forall s0 s , A s0 s -> Q s0 s)) ->
        (* Strong Invariant is satisfied *)
        (forall s0 s, A s0 s -> I s s) ->


         
           ((exists i, Z_of_nat (length list_offset) = (high - low + 1)%Z /\
                             (Int.toZ i < low \/ high < Int.toZ i)) ->
              SP_Judgement true (OFFSET.jump l def)
                                   (SP_pre P M l (OFFSET.jump l def) A)
                                   (SP_post P M l (OFFSET.jump l def) B)
                                   (SP_inv P M l (OFFSET.jump l def) I)) ->

              (forall o, (exists i, exists n, Z_of_nat (length list_offset) = (high - low + 1)%Z /\
                                              low <= Int.toZ i <= high /\
                                              Z_of_nat n = (Int.toZ i) - low /\
                                             nth_error list_offset n = Some o) ->
                              SP_Judgement true (OFFSET.jump l o)
                                               (SP_pre P M l (OFFSET.jump l o) A)
                                               (SP_post P M l (OFFSET.jump l o) B)
                                              (SP_inv P M l (OFFSET.jump l o) I))->
     (********************************************************************)
      (forall s0 s, AA s0 s -> A s0 s) ->
      (forall s0 s t, AA s0 s -> B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
      SP_Judgement_tableswitch def low high list_offset l AA BB II.

 Inductive SP_Judgement_return (l:PC) (AA:Assertion) (BB:LocalPost)
                               (II:StrongInv) : Prop := 
  | SP_RETURN_return : forall (A:Assertion) (B:LocalPost) (I:StrongInv),
      (* Local Invariant is satisfied *)
      (forall Q, Anno_LOOKUP l = Some Q -> 
                 (forall s0 s , A s0 s -> Q s0 s)) ->
      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Local Post is satisfied *)
      (forall s0 s, A s0 s ->
         forall h2 ov, ReturnStep P M (l,s) (h2,Normal ov) ->
         B s0 s (h2,Normal ov)) ->
      (********************************************************************)
      (forall s0 s, AA s0 s -> A s0 s) ->
      (forall s0 s t, AA s0 s -> B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
      SP_Judgement_return l AA BB II.

 Inductive SP_Judgement_goto (l:PC)(o:OFFSET.t)(AA:Assertion)
                             (BB:LocalPost) (II:StrongInv) : Prop := 

   | SP_GOTO_goto : forall (A:Assertion) (B:LocalPost) (I:StrongInv) ,
      (* Local Assertion is satisfied *) 
      (forall Q, Anno_LOOKUP l = Some Q ->
                  (forall s0 s , A s0 s -> Q s0 s)) ->

      (* Strong Invariant is satisfied *)
      (forall s0 s, A s0 s -> I s s) ->

      (* Local Post is satisfied *)
      SP_Judgement true (OFFSET.jump l o)
                        (SP_pre P M l (OFFSET.jump l o) A)
                        (SP_post P M l (OFFSET.jump l o) B)
                        (SP_inv P M l (OFFSET.jump l o) I) ->
    (********************************************************************)
      (forall s0 s, AA s0 s -> A s0 s) ->
      (forall s0 s t, AA s0 s -> B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
      SP_Judgement_goto l o AA BB II.

 Inductive SP_Judgement_if (l:PC)(o:OFFSET.t)(AA:Assertion) (BB:LocalPost)
                           (II:StrongInv) : Prop := 
  | SP_IF_if : forall (A:Assertion) (B:LocalPost) (I:StrongInv) (ll:PC),
      next M l = Some ll ->
      (forall Q , Anno_LOOKUP l = Some Q ->
                  (forall s0 s , A s0 s -> Q s0 s)) ->
      (forall s0 s, A s0 s -> I s s) ->

      SP_Judgement true (OFFSET.jump l o)
        (SP_pre P M l (OFFSET.jump l o) A)
        (SP_post P M l (OFFSET.jump l o) B)
        (SP_inv P M l (OFFSET.jump l o) I) -> 
      SP_Judgement true ll
        (SP_pre P M l ll A)
        (SP_post P M l ll B)
        (SP_inv P M l ll I) ->
    (********************************************************************)
      (forall s0 s, AA s0 s -> A s0 s) ->
      (forall s0 s t, AA s0 s -> B s0 s t -> BB s0 s t) ->
      (forall t s , I t s -> II t s) ->
     SP_Judgement_if l o AA BB II.


 End METHOD.

(*The good program property, with behavioural subtyping*)
 Definition mkMethInv (pre:MethPre) (inv:MethInv) s0 s := 
    pre s0 -> inv s0 s.

  Inductive IMPLIES (M:MethSpec) (M1:MethSpec) : Prop :=
  IMPLIES_def: forall (A:MethPre) (B:MethPost) (I:MethInv)
                              (A1:MethPre) (B1:MethPost) (I1:MethInv), 
          (forall s t, mkMethSpec A B s t -> mkMethSpec A1 B1 s t) ->
          (forall s t, mkMethInv A I s t -> mkMethInv A1 I1 s t) ->
         M = ((A,,B),,I) ->
         M1 = ((A1,,B1),,I1) -> 
         IMPLIES M M1.

(*The property charcterising behavioural subtyping BST requires
    that the dynamically identified method has a specification that
    is finer (i.e. implies) the specification of the statically identieifed method.
    Probably, this property should be split into 2 parts:
    1. The class of the object on which the method is dynamically invoked
        is a subclass of the class in which the statically identified method is
        defined (hence, the dynamically identified method overrides the static
        one). This property holds (I believe) for bytecode-verified programs - 
        whether we prove this fact explicitly is another matter. An alternative
        would be to include this condition as a side condition of the proof rule, 
        for example in an (implicit consequence of any) annotation Q.
    2. The condition that overriding methods have a specification that is
        stronger than the overridden method's specification. This is a condition
        on the method specification table which we require to be satisfied.
    Whether this split is actually performed is to a certain extent orthogonal
    to the soundness proof.*)
 Definition BST : Prop :=
       forall m l mid m0 ms LST LAT cl m',
          instructionAt m l = Some(Invokevirtual mid) ->
          findMethod P mid =Some m0 ->
          MSPEC_LOOKUP mid = Some((ms,,LST),,LAT) ->
          lookup P (fst mid) (snd mid) (cl,m') ->
         (exists ms1, exists LST1, exists LAT1, 
                MSPEC_LOOKUP (cl, METHOD.signature m') = Some ((ms1,,LST1),,LAT1) /\
                                             IMPLIES ms1 ms).

 Definition StrongInvConversion (I:MethInv) : StrongInv :=
    fun s t => (forall S H , s=(H,OperandStack.empty,S) -> I (S,H) t).

 Definition GP : Prop :=
   BST /\
   ( forall msig pre post MI LST LAT,
      MSPEC_LOOKUP msig = Some ((((pre,,post),,MI),,LST),,LAT) ->
      forall m, findMethod P msig = Some m -> 
      (   (forall MB,
            METHOD.body m = Some MB ->
            SP_Judgement m LST LAT false 
            (BYTECODEMETHOD.firstAddress MB)
            (fun s0 s => let (S,H) := s0 in pre s0 /\ s = (H,OperandStack.empty,S))
            (fun s0 s t => let (S,H):= s0 in 
               s = (H,OperandStack.empty,S) -> post (S,H) t) 
            (StrongInvConversion MI))
        /\ (forall l A B I, 
              ASSUM_LOOKUP LST l = Some ((A,,B),,I) ->
              SP_Judgement m LST LAT false l A B I))).
