Add LoadPath "../Bicolano".
Add LoadPath "../Library".
Require Export ZArith.
Require Export Misc.
Require Export Tactic.
Require Export BigStep.
Add LoadPath "../Library/Map".
Require Export MapSignatures.
Require Export BaseLogic2.

Module SWp (BL:RULES).
 Export BL V5 Dom Prog.

 
 Section ____TYPE_SYNTACTIFICATION____.

 Inductive WpType : Set := 
   | BtLocation | BtInt | BtClassName | BtType
   | BtHeap | BtReturnVal | BtLocationType | BtValue
   | BtOperandStack | BtLocalVar.

  Section ____VALUE_DEPENDENT____.

   Set Implicit Arguments.
   Set Contextual Implicit.
   Variable V : Type.

   Definition SS := list V.

   Inductive SLv : Type :=
     | SLvUpdate : SLv -> Var -> V -> SLv
     | SLvVar :> LocalVar.t -> SLv
     | SLvStack2LocalVar : SS -> nat -> SLv.

   Inductive SLoc : Type :=
     | SLValue :> V -> SLoc
     | SLVar :> Location -> SLoc.

   Inductive SRV : Type :=
     | SRVVar :> ReturnVal -> SRV
     | SRVNormalSome : V -> SRV
     | SRVNormalNone : SRV
     | SRVException : SLoc -> SRV.

   Inductive SLT : Type :=
     (*| SLTHTypeOf : SH -> SLoc -> SLT*)
     | SLTVar :> Heap.LocationType -> SLT.
 
   Inductive SInt : Type :=
     | SIVar :> Int.t -> SInt
     (*| SItypeArray : SH -> SLoc -> SInt*)
     | SIValue : V -> SInt
     | SI2b : SInt -> SInt
     | SI2s : SInt -> SInt
     | SIBinop : BinopInt -> SInt -> SInt -> SInt
     | SINeg : SInt -> SInt.

   Inductive SAM : Type :=
     | SAMStaticField :> FieldSignature -> SAM
     | SAMDynamicField : SLoc -> FieldSignature -> SAM
     | SAMArrayElement : SLoc -> SInt -> SAM.
 
   Inductive SH : Type :=
     | SHVar : Heap.t -> SH
     | SHUpdate : SH -> SAM -> V -> SH.

   Inductive SClassName : Type :=
     | SCNVar :> ClassName -> SClassName.

   Inductive Snum : Type :=
     | SnVar :> num -> Snum
     | SnI :> SInt -> Snum.
 
   Inductive SType : Type :=
     | STVar :> type -> SType.

  End ____VALUE_DEPENDENT____.

  Inductive SValue : Type :=
    | SVnum : Snum SValue -> SValue
    (*| SVNull : SValue*)
    | SVRef : SLoc SValue -> SValue
    | SVVar :> value -> SValue
    | SConv4Stack : SValue -> SValue
    | SVGetNth : nat -> SS SValue -> SValue
    | SConv4Array : SValue -> SType -> SValue
    | SLvGet : SLv SValue -> Var -> SValue
    | SHGet : SH SValue -> SAM SValue -> SValue.

   Definition SSVar := map SVVar.



 End ____TYPE_SYNTACTIFICATION____.

 Definition SLocation := SLoc SValue.
 Definition SOperandStack := SS SValue. 
 Definition SLocalVar := SLv SValue.
 Definition SHeap := SH SValue.
 Definition SReturnVal := SRV SValue.
 Definition SAdressingMode := SAM SValue.
 Definition SLocationType := SLT. (*SValue.*)

 Definition shvar (h : Heap.t) : SHeap := @SHVar SValue h.
 Coercion shvar : Heap.t >-> SHeap.

 Definition wp_type (t : WpType) : Type :=
   match t with
   | BtLocation => SLocation
   | BtInt => SInt SValue
   | BtClassName => SClassName
   | BtType => SType
   | BtHeap => SHeap
   | BtReturnVal => SReturnVal
   | BtLocationType => SLocationType
   | BtValue => SValue
   | BtOperandStack => SOperandStack
   | BtLocalVar => SLocalVar
   end.


 Definition interp_wptype (t : WpType) : Type :=
   match t with
   | BtLocation => Location
   | BtInt => Int.t
   | BtClassName => ClassName
   | BtType => type
   | BtHeap => Heap.t
   | BtReturnVal => ReturnVal
   | BtLocationType => Heap.LocationType
   | BtValue => value
   | BtOperandStack => OperandStack.t
   | BtLocalVar => LocalVar.t
   end.


 Parameter unknown_constant: forall t : WpType, interp_wptype t.
 Parameter unknown_wp_constant: forall t : WpType, wp_type t.
 Parameter unknown_location : SValue -> SLocation.
 Parameter unknown_int : SValue -> SInt SValue.


 (* Parameter value2int : value -> Int.t. *)
 Definition value2int (v:value) : Int.t :=
   match v with
    | Num (I i) => i
    | _ => unknown_constant BtInt
   end.
 Coercion value2int : value >-> Int.t.

(* Parameter value2loc : value -> Location.*)
 Definition value2loc (v:value) : Location :=
   match v with
    | Ref loc => loc
    | _ => unknown_constant BtLocation
   end.
 Coercion value2loc : value >-> Location.

(* Parameter svalue2sloc : SValue -> SLocation.*)
 Definition svalue2sloc (v : SValue) : SLocation :=
   match v with
   | SVRef l => l
   | SVVar (Ref l) => SLVar l
   | SLvGet l x => SLValue v
   | SHGet h am => SLValue v
   | l => unknown_location l
   end.
 Coercion svalue2sloc : SValue >-> SLocation.

(* Parameter svalue2sint : SValue -> SInt SValue.*)
 Definition svalue2sint (v : SValue) : SInt SValue :=
   match v with
   | SVnum (SnI n) => n
   | SVVar (Num (I n)) => SIVar n
   | SVnum (SnVar (I n)) => SIVar n
   | SLvGet l x => SIValue v
   | SHGet h am => SIValue v
   | v => unknown_int v
   end.
 Coercion svalue2sint : SValue >-> SInt.

 
 Fixpoint var_nat_eq (n m : nat) { struct n } : bool :=
   match n, m with
   | O, O => true
   | S n, S m => var_nat_eq n m
   | _, _ => false
   end.

  Definition var_eq x y := var_nat_eq (Var_toN x) (Var_toN y).

Definition mkupdate := @SLvUpdate SValue.
Definition mkget := SLvGet.

 Definition SInitState := SLocalVar ** SHeap.
 Definition SLocalState := (SHeap ** SOperandStack) ** SLocalVar.
 Definition SReturnState := SHeap ** SReturnVal.

 Inductive SWp : Type :=
   (* Heap *)
   | HTypeArray : PC -> SHeap -> SValue -> SInt SValue -> SType -> SWp
   | HTypeObject : PC -> SHeap -> SValue -> SClassName -> SWp
   | HNewArray : PC -> SHeap -> SInt SValue -> SType -> SLocation -> SHeap -> SWp
   | HNewObject : PC -> SHeap -> SClassName -> SLocation -> SHeap -> SWp
   (* Comparison *)
   | SCompRef : PC -> CompRef -> SValue -> SValue -> SWp
   | SCompInt : PC -> CompInt -> SInt SValue -> SInt SValue -> SWp
   (* Logic *)
   | SFalse : PC -> SWp
   | STrue : PC -> SWp
   | SAnd : SWp -> SWp -> SWp
   | SImp : SWp -> SWp -> SWp
   | SNeg : SWp -> SWp
   | SForall : forall (t:WpType), (wp_type t -> SWp) -> SWp
   (* Other *)
   | SPre : PC -> option SWp -> MethodSignature -> SInitState -> SWp
   | SPost : PC -> option SWp -> MethodSignature -> SInitState -> SReturnState -> SWp
   | SAssert : PC -> option SWp -> MethodSignature -> PC -> SInitState -> SLocalState -> SWp
   (* Predicates *)
   | SisNull : PC -> SValue -> SWp
   | SisStatic : PC -> FieldSignature -> SWp
   | SCheckByteSize : PC -> Z -> SWp
   | SisZero : PC -> SValue -> SWp
   | SBound : PC -> SValue -> SValue -> SWp
   | SIntBound : PC -> primitiveType -> Z -> SWp
   | Ssubclass_name : PC -> SClassName -> ClassName -> SWp
   | Sassign_compatible : PC -> SHeap -> SValue -> SType -> SWp
   | SNonNegative : PC -> SValue -> SWp
   | Sdefined_field : PC -> SClassName -> FieldSignature -> SWp
   | Scompat_ArrayKind_type : PC -> ArrayKind -> SType -> SWp
   | Scompat_ArrayKind_value : PC -> ArrayKind -> SValue -> SWp
   | Scompat_ValKind_value : PC -> ValKind -> SValue -> SWp.



 Notation "! x" := (SNeg x) (at level 75).
 Notation "x --> y" := (SImp x y) (at level 80, right associativity).
 Notation "'Forall' x : t , P" := (SForall t (fun x : wp_type t => P)) (at level 200, x ident). 
 Notation "x & y" := (SAnd x y) (at level 80, right associativity).


 Definition SAssertion := SInitState -> SLocalState -> SWp.


 Section ____WPS_TRANSLATION____.


  Section ____VALUE_DEPENDENT____.

   Variable interp_v: forall v : SValue, value.
   Fixpoint interp_os (s : SOperandStack) : OperandStack.t :=
     match s with
     | nil => OperandStack.empty
     | v :: vs => interp_v v :: interp_os vs
     end.

   Fixpoint interp_lv (l : SLocalVar) : LocalVar.t :=
     match l with
     | SLvUpdate lv x v => LocalVar.update (interp_lv lv) x (interp_v v)
     | SLvVar lv => lv
     | SLvStack2LocalVar os n => stack2localvar (interp_os os) n
     end.

   Parameter another_location : Location.

   Definition interp_location (l : SLocation) : Location :=
     match l with
     | SLVar loc => loc
     | SLValue v => interp_v v
     end.

   Parameter another_int : Int.t.

   Fixpoint interp_int (i : SInt SValue) : Int.t :=
     match i with
     | SIVar i => i
     | SIValue v => interp_v v 
     | SI2b i => b2i (i2b (interp_int i))
     | SI2s i => s2i (i2s (interp_int i))
     | SIBinop op i1 i2 => SemBinopInt op (interp_int i1) (interp_int i2)
     | SINeg i => Int.neg (interp_int i)
     end.

   Definition interp_am (a : SAdressingMode) : Heap.AdressingMode :=
     match a with
     | SAMStaticField f => Heap.StaticField f
     | SAMDynamicField l f => Heap.DynamicField (interp_location l) f
     | SAMArrayElement l i => Heap.ArrayElement (interp_location l) (interp_int i)
     end.

   Fixpoint interp_h (h : SHeap) : Heap.t :=
     match h with
     | SHVar h => h
     | SHUpdate h a v => Heap.update (interp_h h) (interp_am a) (interp_v v)
     end.

   Definition interp_rv (r : SReturnVal) : ReturnVal :=
     match r with
     | SRVVar rv => rv
     | SRVNormalSome v => Normal (Some (interp_v v))
     | SRVNormalNone => Normal None
     | SRVException l => Exception (interp_location l)
     end.

Definition do_htypeof (h : Heap.t) (loc : Location) :=
  match Heap.typeof h loc with
  | Some v => v
  | None => unknown_constant BtLocationType
  end.

Definition interp_locationtype (l : SLocationType) : Heap.LocationType :=
match l with
| SLTVar l => l
end.

Definition interp_classname (c : SClassName) : ClassName :=
match c with
| SCNVar c => c
end.

Definition interp_num (n : Snum SValue ) : num :=
match n with
| SnVar n => n
| SnI i => I (interp_int i)
end.

Definition interp_type (t : SType) : type :=
match t with
| STVar t => t
end.

End ____VALUE_DEPENDENT____.

Section ____VALUE____.

Variable interp_os : (SValue -> value) -> (SOperandStack -> OperandStack.t).

Definition do_lvget (l : LocalVar.t) (x : Var) :=
  match LocalVar.get l x with
    | Some v => v
    | None => unknown_constant BtValue
    end.

Definition do_hget (h : Heap.t) (am : Heap.AdressingMode) :=
  match Heap.get h am with
  | Some v => v
  | None => unknown_constant BtValue
  end.


Fixpoint interp_v (n : nat) (v : SValue) { struct n } : value :=
match n with 
| O => unknown_constant BtValue
| S m => match v with
         | SVnum n => Num (interp_num (interp_v m) n)
         | SVRef l => Ref (interp_location (interp_v m) l)
         | SVVar v => v
         | SConv4Stack v => conv_for_stack (interp_v m v)
         | SVGetNth n os => match n, interp_os (interp_v m) os with
	                    | O, nil => unknown_constant BtValue
                            | O, v :: _ => v
			    | _, os' => match OperandStack.get_nth os' n with
			                | Some v => v
					| None => unknown_constant BtValue
					end
                            end
         | SConv4Array v t => conv_for_array (interp_v m v) (interp_type t)
         | SLvGet l x => do_lvget (interp_lv (interp_v m) l) x
         | SHGet h am => do_hget (interp_h (interp_v m) h) (interp_am (interp_v m) am)
         end
end.


End ____VALUE____.

Definition interp_value := interp_v interp_os 3.
Definition interp_operandstack := interp_os interp_value.
Definition interp_heap := interp_h interp_value.
Definition interp_returnval := interp_rv interp_value.
Definition interp_localvar := interp_lv interp_value.
Definition interp_adressingmode := interp_am interp_value.

Definition InitState := LocalVar.t ** Heap.t.
Definition LocalState := Heap.t ** OperandStack.t ** LocalVar.t.
Definition ReturnState := Heap.t ** ReturnVal.
Definition Assertion := InitState -> LocalState -> Prop.

Definition interp_initstate (s : SInitState) : InitState :=
match s with
| pair l h => (interp_localvar l, interp_heap h)
end.

Definition interp_localstate (s : SLocalState) : LocalState :=
match s with
| pair (pair h os) l => (interp_heap h, interp_operandstack os, interp_localvar l)
end.

Definition interp_returnstate (s : SReturnState) : ReturnState :=
match s with
| pair h rv => (interp_heap h, interp_returnval rv)
end.

 Inductive Assumption : Set :=
   | not_null : list nat -> Assumption.

(** A deplacer *)
 Lemma isRef_notassig_Ref : 
     forall p h v t,
       isReference v -> 
        ~ assign_compatible p h v (ReferenceType t) ->
        exists loc, v = Ref loc.
 Proof.
  intros p0 h v t H;inversion_mine H;intros HH;eauto.   
  elim HH. constructor. 
 Qed.    
 
 Lemma lookup_unique :  
      forall p exl pc e pc' pc'',
            Dom.lookup_handlers p exl pc e pc' -> 
            Dom.lookup_handlers p exl pc e pc'' ->
            pc' = pc''.
 Proof.
  induction 1;intros HH;inversion_mine HH;simpl_eq;auto.
  elim (H2 H). elim (H H6).
 Qed. 

(** Fin a deplacer *)


 Definition GlobalMethSpec := (InitState -> list Prop) ** (InitState -> ReturnState -> list Prop).
 Definition LocMethSpec := PCM.t ((InitState -> LocalState -> list Prop) ** nat) ** PCM.t Assumption.
 
 Definition MethSpecTab := MM.t (nat**GlobalMethSpec**LocMethSpec).

 Definition assertLookup (lms:LocMethSpec) pc := 
   match PCM.get (fst lms) pc with
   | Some (pairT A _) => Some A
   | _ => None 
   end.

 Definition globalLookup (mst:MethSpecTab) msig := 
   match MM.get mst msig with
   | Some (pairT (pairT _ gms) _) => Some gms
   | None => None 
   end.

 Definition lookupPre mst msig := 
   match globalLookup mst msig with 
   | Some (pre, _ ) => pre
   | _ => fun _ => True :: nil
   end.

 Definition lookupPost mst msig := 
   match globalLookup mst msig with 
   | Some (_, post) => post
   | _ => fun _ _ => True :: nil
   end.
  
 Definition localLookup (mst:MethSpecTab) msig pc := 
   match MM.get mst msig with
   | Some (pairT (pairT _ gms) lms) => 
       match PCM.get (fst lms) pc with
       | Some (a,_) => a
       | _ => fun _ _ => True :: nil
       end
   | None => fun _ _ => True :: nil
   end.
 
 Record annotatedProg : Type := AnnoProg {
   ap_prog     :> Prog.Program;
   ap_subclass : ClassName -> ClassName -> bool;
      (* A decidable version of subclass_name relation of p *)
   ap_mst      :> MethSpecTab
 }.

 Record annotatedMethod : Type := AnnoMeth {
   am_needed : nat;
   am_meth_sign:> MethodSignature;
   am_meth     :> Method;
   am_bytecode :> BytecodeMethod;
   am_excHand  :> list ExceptionHandler;
   am_pre      : InitState -> list Prop;
   am_post     : InitState -> ReturnState -> list Prop;
   am_local    :> LocMethSpec
 }.



Section INTERP.
 Variable  (mst:MethSpecTab).
 Variable p : Program.
 Variable wps_label : PC -> Assertion.

 Definition type_constr (t : WpType) : interp_wptype t -> wp_type t := 
   match t as w return (interp_wptype w -> wp_type w) with
   | BtLocation => @SLVar SValue
   | BtInt => @SIVar SValue
   | BtClassName => SCNVar
   | BtType => STVar
   | BtHeap => @SHVar SValue
   | BtReturnVal => @SRVVar SValue
   | BtLocationType => SLTVar (*SValue*)
   | BtValue => SVVar
   | BtOperandStack => SSVar
   | BtLocalVar => @SLvVar SValue
   end.
 

 Fixpoint interp_assert (l:list Prop) : Prop :=
   match l with
   | nil => True
   | x::nil => x
   | x::xs => x /\ interp_assert xs
   end.

 Fixpoint interp_assume (l:list Prop) c : Prop :=
   match l with
   | nil => c
   | x::xs => x -> interp_assume xs c
   end.
 
Fixpoint interp_swp (w : SWp) : Prop :=
match w with
| HTypeArray _ h v i t => Heap.typeof (interp_heap h) (interp_value v) = 
                        Some (Heap.LocationArray (interp_int interp_value i) (interp_type t))
| HTypeObject _ h v c => Heap.typeof (interp_heap h) (interp_value v) =
                       Some (Heap.LocationObject (interp_classname c))
| HNewArray _ h i t l h' =>
     Heap.new (interp_heap h) p (Heap.LocationArray (interp_int interp_value i) (interp_type t)) =
     Some (interp_location interp_value l, interp_heap h')
| HNewObject _ h c l h' =>
     Heap.new (interp_heap h) p (Heap.LocationObject (interp_classname c)) =
     Some (interp_location interp_value l, interp_heap h')
| SCompRef _ op v1 v2 => SemCompRef op (interp_value v1) (interp_value v2)
| SCompInt _ op i1 i2 => SemCompInt op (Int.toZ (interp_int interp_value i1)) (Int.toZ (interp_int interp_value i2))
| SFalse _ => False
| STrue _ => True
| SAnd w1 w2 => interp_swp w1 /\ interp_swp w2
| SImp w1 w2 => interp_swp w1 -> interp_swp w2
| SNeg w => ~ interp_swp w
| SForall x f => match x with
	         | BtLocation => forall loc : interp_wptype x, interp_swp (f (type_constr x loc))
	         | BtInt => forall i : interp_wptype x, interp_swp (f (type_constr x i))
	         | BtClassName => forall cn : interp_wptype x, interp_swp (f (type_constr x cn))
	         | BtType => forall t : interp_wptype x, interp_swp (f (type_constr x t))
	         | BtHeap => forall h : interp_wptype x, interp_swp (f (type_constr x h))
	         | BtReturnVal => forall rv : interp_wptype x, interp_swp (f (type_constr x rv))
	         | BtLocationType => forall l : interp_wptype x, interp_swp (f (type_constr x l))
	         | BtValue => forall v : interp_wptype x, interp_swp (f (type_constr x v))
	         | BtOperandStack => forall os : interp_wptype x, interp_swp (f (type_constr x os))
	         | BtLocalVar => forall lv : interp_wptype x, interp_swp (f (type_constr x lv))
                 end
| SPre _ post m ss0 =>
    let s0 := interp_initstate ss0 in
    let l := lookupPre mst m s0 in
    match post with
    | Some p => interp_assume l (interp_swp p)
    | None => interp_assert l
    end
| SPost _ post m ss0 sr =>
    let s0 := interp_initstate ss0 in
    let r := interp_returnstate sr in
    let l := lookupPost mst m s0 r in
    match post with
    | Some p => interp_assume l (interp_swp p)
    | None => interp_assert l
    end
| SAssert _ post m pc ss0 ss => 
    let s0 := interp_initstate ss0 in
    let s := interp_localstate ss in
    let l := localLookup mst m pc s0 s in
    match post with
    | Some p => interp_assume l (interp_swp p)
    | None => interp_assert l
    end
                      
| SisNull _ v => interp_value v = Null
| SisStatic _ f => isStatic p f
| SCheckByteSize _ z => (Byte.range z)
| SisZero _ v => Int.toZ (interp_value v) = 0
| SBound _ v1 v2 => 0 <= Int.toZ (interp_value v1) < Int.toZ (interp_value v2)
| SIntBound _ t z => match t with
                   | BYTE  => Byte.range z 
                   | SHORT => Short.range z 
                   | INT   => Int.range z 
                   | _ => False
                   end
| Ssubclass_name _ c c' => subclass_name p (interp_classname c) c'
| Sassign_compatible _ h v t => assign_compatible p (interp_heap h) (interp_value v) (interp_type t)
| SNonNegative _ v => 0 <= Int.toZ (interp_value v)
| Sdefined_field _ c f => defined_field p (interp_classname c) f
| Scompat_ArrayKind_type _ k t => compat_ArrayKind_type k (interp_type t)
| Scompat_ArrayKind_value _ k v => compat_ArrayKind_value k (interp_value v)
| Scompat_ValKind_value _ k v => compat_ValKind_value k (interp_value v)
end.

End INTERP.

End ____WPS_TRANSLATION____.

 Section ____WPS____.

  Variable p:annotatedProg.
  
  Hypothesis ap_subclass_correct :
    forall cn cn', 
      if p.(ap_subclass) cn cn' then subclass_name p cn cn' 
      else ~subclass_name p cn cn'.

  Lemma ap_subclass_complete_t : 
    forall cn cn', subclass_name p cn cn' -> p.(ap_subclass) cn cn' = true.
  Proof.
   intros cn cn' H;assert (HH:= ap_subclass_correct cn cn');
    destruct (p.(ap_subclass) cn cn');trivial.  elim (HH H).
  Qed.

  Fixpoint lookup_handlers 
    (exl:list ExceptionHandler) (pc:PC) (cl:ClassName) {struct exl} : 
                                                               option PC :=
    match exl with
    | nil => None
    | ex :: exl' => 
      if EXCEPTIONHANDLER.isPCinRange ex pc then
	match EXCEPTIONHANDLER.catchType ex with
        | None => Some (EXCEPTIONHANDLER.handler ex)
        | Some cl' =>
	  if p.(ap_subclass) cl cl' then Some (EXCEPTIONHANDLER.handler ex)
	  else lookup_handlers exl' pc cl
        end
      else lookup_handlers exl' pc cl
    end.

  Section ____WPS_METH____.
  
   Variable m:annotatedMethod.
   Record valid_am : Prop := mkVAM {
     vam_Bytecode : METHOD.body m = Some m.(am_bytecode);   
     vam_Handler : m.(am_excHand) = BYTECODEMETHOD.exceptionHandlers m
   }.
 
   Hypothesis isValidAm : valid_am.
 
   Lemma lookup_handler_correct :
     forall pc e,
       match lookup_handlers m pc e with
       | Some pc' => Dom.lookup_handlers p m pc e pc'
       | None => forall pc', ~Dom.lookup_handlers p m pc e pc'
       end.
   Proof.
    intros pc e;generalize m.(am_excHand).
    induction l;simpl;intros.
    intro H;inversion H.
    case_eq (EXCEPTIONHANDLER.isPCinRange a pc);intros Heq.
    case_eq (EXCEPTIONHANDLER.catchType a);intros.
    generalize (ap_subclass_correct e c);destruct (ap_subclass p e c);intros.
    econstructor;eauto. econstructor 2;eauto.
    destruct (lookup_handlers l pc e).
    econstructor 2;eauto.
    intro HH;inversion_mine HH;simpl_eq. elim (H0 H3).
    intros pc' HH;inversion_mine HH.
    inversion_mine H6;simpl_eq. elim (H0 H3). elim (IHl _ H7).
    econstructor;eauto. econstructor;eauto.
    destruct (lookup_handlers l pc e).
    econstructor 2;eauto.
    intro HH;inversion_mine HH;simpl_eq.
    intros pc' HH;inversion_mine HH.
    inversion_mine H4;simpl_eq. elim (IHl _ H5).
   Qed.
  
   Lemma lookup_complete_t : 
     forall pc e pc', Dom.lookup_handlers p m pc e pc' -> 
       lookup_handlers m pc e = Some pc'.
   Proof.
    intros;generalize (lookup_handler_correct pc e);
     destruct (lookup_handlers m pc e).
    intros Heq;rewrite (lookup_unique  _ _ _ _ _ _ H Heq);trivial.
    intros HH;elim (HH _ H).
   Qed.

    Section ____WPS_INSTR____.

     Variable wps_label : PC -> SAssertion.
     Definition interp_p := interp_swp p p.
     Coercion interp_p : SWp >-> Sortclass.


     Definition wps_next pc : SAssertion :=
       match BYTECODEMETHOD.nextAddress m pc with
       | None => fun s0 s => STrue pc
       | Some pc' => wps_label pc'
       end.

    Lemma wp_next_wp_label : 
      forall pc pc' s0 s, 
        interp_p (wps_next pc s0 s) ->
        next m pc = Some pc' ->
        interp_p (wps_label pc' s0 s).
    Proof.
      unfold wps_next, next;intros pc pc' s0 s.
      rewrite isValidAm.(vam_Bytecode);intros H Heq;rewrite Heq in H;trivial.
    Qed.

     Definition handlers_in_range m pc := 
       filter (fun ex => EXCEPTIONHANDLER.isPCinRange ex pc) m.

     Fixpoint wps_Exception_handlers_post exl (pc:PC) (loc:SLocation) (h:SHeap) l (s0:SInitState) { struct exl } : SWp := 
       match exl with
       | nil => SPost pc None m s0 (h, SRVException loc)
       | ex :: exl' => 
         if EXCEPTIONHANDLER.isPCinRange ex pc then
           let pc' := EXCEPTIONHANDLER.handler ex in
           match EXCEPTIONHANDLER.catchType ex with
           | None => wps_label pc' s0 (h, SVRef loc :: nil, l)
           | Some cl' =>
            ((Sassign_compatible pc h (SVRef loc) (ReferenceType (ClassType cl'))) -->
                   wps_label pc' s0 (h, SVRef loc :: nil , l))
             & (!(Sassign_compatible pc h (SVRef loc) (ReferenceType (ClassType cl'))) -->
                   wps_Exception_handlers_post exl' pc loc  h l s0)
           end
         else wps_Exception_handlers_post exl' pc loc  h l s0
       end.


      Definition wps_Exception_unknown := wps_Exception_handlers_post m.
      
(*
     Lemma wps_Exception_unknown_caught :
      forall pc cn pc' loc2 h2 l s0,
        Dom.lookup_handlers p m pc cn pc' ->
        wps_Exception_unknown pc loc2 cn h2 l s0 ->
        wps_label pc' s0 (h2, SVRef loc2 :: nil, l).
*)

(*       
   Lemma wps_Exception_unknown_uncaught :
     forall pc cn loc2 h2 l s0,
       Heap.typeof h2 loc2 = Some (Heap.LocationObject cn) ->
       UnCaughtException p m (pc, h2, loc2) ->
       wps_Exception_unknown pc (SLVar loc2) cn h2 l s0 ->
       SPost pc None m s0 (SHVar h2, SRVException (SLVar loc2)).
   Proof.
*)

   Parameter pc_wps_exception : PC.
   Definition wps_Exception pc loc cn h l s0 := 
      match lookup_handlers m pc cn with
      | None => (* UncaughtException *)
          SPost pc None m s0 (h, SRVException loc)
      | Some pc' =>
 	  wps_label pc' s0 (h, (SVRef loc) :: nil, l)
      end.

(*  
    Lemma wp_Exception_caught :
      forall pc cn pc' loc2 h2 l s0,
        Dom.lookup_handlers p m pc cn pc' ->
        interp_p (wps_Exception pc loc2 cn h2 l s0) ->
        interp_p (wps_label pc' s0 (h2, SVRef loc2 :: nil, l)).
    Proof.
*)
 
(*
   Lemma wp_Exception_uncaught :
      forall pc cn loc2 h2 l s0,
        Heap.typeof h2 loc2 = Some (Heap.LocationObject cn) ->
        UnCaughtException p m (pc, h2, loc2) ->
        interp_p (wps_Exception pc (SLVar loc2) cn (SHVar h2) l s0) ->
        interp_p (SPost pc None m s0 (SHVar h2, SRVException (SLVar loc2))).
    Proof.
*)

   Definition wps_jvmException pc (e:ShortClassName) (h:SHeap) (l:SLocalVar) (s0:SInitState) :=  
    let jexc : ClassName := (javaLang,e) in
    Forall h' : BtHeap,
    Forall l' : BtLocation,
     HNewObject pc h jexc l' h' --> (wps_Exception pc l' jexc h' l s0).

(* 
   Lemma wp_jvmException_caught :
      forall pc e pc' loc2 h2 h l s0,
        Dom.lookup_handlers p m pc (javaLang, e) pc' ->
        Heap.new h p (Heap.LocationObject (javaLang, e)) = Some (loc2, h2) ->
        interp_p (wps_jvmException pc e (SHVar h) l s0) ->
        interp_p (wps_label pc' s0 (SHVar h2, SVRef (SLVar loc2) :: nil, l)).
    Proof. 
*)

(*
    Lemma wp_jvmException_uncaught :
      forall pc e loc2 h2 h l s0,
        Heap.new h p (Heap.LocationObject (javaLang, e)) = Some (loc2, h2) ->
        UnCaughtException p m (pc, h2, loc2) ->
        interp_p (wps_jvmException pc e (SHVar h) l s0) ->
        interp_p (SPost pc None m s0 (SHVar h2, SRVException (SLVar loc2))).
    Proof. 
*)

    Definition wps_NullPointerException pc :=
      wps_jvmException pc NullPointerException.

(*  
    Lemma wp_NullPointerException_caught :
      forall pc pc' loc2 h2 h l s0,
        Dom.lookup_handlers p m pc (javaLang, NullPointerException) pc' ->
        Heap.new h p (Heap.LocationObject (javaLang, NullPointerException)) =
           Some (loc2, h2) ->
        interp_p (wps_NullPointerException pc (SHVar h) l s0) ->
        interp_p (wps_label pc' s0 (SHVar h2, SVRef (SLVar loc2) :: nil, l)).
    Proof.
*)

(*
    Lemma wp_NullPointerException_uncaught :
      forall pc  loc2 h2 h l s0,
        Heap.new h p (Heap.LocationObject (javaLang, NullPointerException)) =
           Some (loc2, h2) ->
       UnCaughtException p m (pc, h2, loc2) ->
       wps_NullPointerException pc (SHVar h) l s0 ->
       SPost pc None m s0 (SHVar h2, SRVException (SLVar loc2)).
    Proof.
*)

   Definition wps_cond (c t f : SWp) : SWp := (c --> t) & (! c --> f).

    Lemma wp_cond_true : 
      forall cond wpT wpF, 
        interp_p (wps_cond cond wpT wpF) -> interp_p cond -> interp_p wpT.
    Proof. intros cond wpT wpF (H1,H2);auto. Qed.

    Lemma wp_cond_false : 
      forall cond wpT wpF, 
        interp_p (wps_cond cond wpT wpF) -> interp_p (!cond) -> interp_p wpF.
    Proof. intros cond wpT wpF (H1,H2);auto. Qed.
 

    Definition wps_jvmExc cond wpT pc e h l s0 :=
      wps_cond cond wpT (wps_jvmException pc e h l s0).

(*    
    Lemma wp_jvmExc_true : 
      forall cond wpT pc e h l s0, 
       interp_p (wps_jvmExc cond wpT pc e h l s0) -> interp_p cond -> interp_p wpT.
    Proof.
*)


(*
    Lemma wp_jvmExc_false :
      forall cond wpT pc e h l s0,
        interp_p (wps_jvmExc cond wpT pc e h l s0) -> interp_p (!cond) ->
        interp_p (wps_jvmException pc e h l s0).
    Proof.
*)

    Definition wps_bound (i length:SValue) wpIn pc :=
      wps_jvmExc (SBound pc i length)
         wpIn pc ArrayIndexOutOfBoundsException.


(*
    Lemma wp_bound_true :
      forall (i length : SValue) wpIn pc h l s0,
        interp_p (wps_bound i length wpIn pc h l s0) -> 
        (0 <= Int.toZ (interp_value i) < Int.toZ (interp_value length)) ->
        interp_p wpIn.
    Proof. 
*)

(*
    Lemma wp_bound_false :
      forall (i length : SValue) wpIn pc h l s0,
        interp_p (wps_bound  i length wpIn pc h l s0) -> 
        ~(0 <= Int.toZ (interp_value i) < Int.toZ (interp_value length)) -> 
        interp_p (wps_jvmException pc ArrayIndexOutOfBoundsException h l s0).
    Proof. 
*)

    Fixpoint is_in_list (l : list nat) (n : nat) { struct l } :=
      match l with
      | nil => false
      | x :: xs => match nat_compare x n with
                   | Eq => true
                   | _ => is_in_list xs n
                   end
      end.

(*
    Definition wps_checkNull pc val n wp h l s0 :=
      let np :=    (SisNull pc val --> wps_NullPointerException pc h l s0) 
                & (! SisNull pc val --> wp) in
      match PCM.get (snd m.(am_local)) pc with
      | Some (not_null al) => if is_in_list al n then wp else np
      | _ => np
      end.
*)
    Definition wps_checkNull pc val (n:nat) wp h l s0 :=
      (SisNull pc val --> wps_NullPointerException pc h l s0) 
      & (! SisNull pc val --> wp).


    Fixpoint pop (A:Type) (n:nat) (l:list A) {struct n}: list A := 
      match n, l with
      | O, _ => l
      | S n, _ :: l => pop A n l
      | S n, nil => nil
      end.

    Lemma pop_append : 
      forall A n l1 l2, length l1 = n  -> pop A n (l1 ++ l2) = l2.
    Proof.
     induction n;simpl;intros.
     destruct l1;try discriminate;trivial.
     destruct l1;try discriminate;simpl.
     injection H;intros;auto.
    Qed.

    Definition init_args n os := (SLvStack2LocalVar os n, pop SValue n os).

      
    Lemma init_args_correct :
      forall n args os l' os', 
        init_args n (args ++ os) = (l', os') ->
        length args = n  ->
        SLvStack2LocalVar (args ++ os) (length args) = l' /\ os' = os.
    Proof.
     unfold init_args;simpl;intros.
     injection H;intros;subst.
     rewrite pop_append;auto.
   Qed.

    Definition wps_i pc (s0 : SInitState) (s: SLocalState) : SWp:=
      match BYTECODEMETHOD.instructionAt m pc with
      | Some ins =>
(** Basic instruction *)
        match ins, s with
        | Aconst_null, (h, os, l) => wps_next pc s0 (h, SVVar Null :: os, l)

        | Arraylength, (h, val :: os', l) =>
          let wp := Forall length : BtInt, 
                    Forall tp : BtType,
                     HTypeArray pc h val length tp -->
                     wps_next pc s0 (h, SVnum length :: os', l) in
          wps_checkNull pc val O wp h l s0

        | Checkcast t, (h, val :: os', l) =>
          (wps_jvmExc (Sassign_compatible pc h val (ReferenceType t))
             (wps_next pc s0 (h, val :: os', l))
             pc ClassCastException h l s0)

        | Const t z, (h, os, l) => SIntBound pc t z -->
            wps_next pc s0 (h, (SVnum ((SIVar (Int.const z)))) :: os, l)

        | Dup, (h, v :: os, l) => wps_next pc s0 (h, v :: v :: os, l)
        | Dup_x1, (h, v1 :: v2 :: os',l) => 
          (wps_next pc s0 (h, v1 :: v2 :: v1 :: os', l))
        | Dup_x2, (h, v1 :: v2 :: v3 :: os', l) =>
          (wps_next pc s0 (h, v1 :: v2 :: v3 :: v1 :: os', l))
        | Dup2, (h, v1 :: v2 :: os', l) =>
          (wps_next pc s0 (h, v1 :: v2 :: v1 :: v2 :: os', l))
        | Dup2_x1, (h, v1 :: v2 :: v3 :: os', l) =>
          (wps_next pc  s0 (h, v1 :: v2 :: v3 :: v1 :: v2 :: os', l))
        | Dup2_x2, (h, v1 :: v2 :: v3 :: v4 :: os', l) =>
          (wps_next pc  s0 (h, v1 :: v2 :: v3 :: v4 :: v1 :: v2 :: os', l))
   
      	| Getfield f, (h, val :: os', l) =>
          let ft := FIELDSIGNATURE.type (snd f) in
          (*let wp := match ft with
                    | PrimitiveType _ => Forall i : BtInt,
                    (HGet pc h (SAMDynamicField (svalue2sloc val) f) (SVnum (SnI i)) --> 
                       wps_next pc s0 (h, (SVnum (SnI i)) :: os', l))
                    | ReferenceType _ => Forall loc : BtLocation,
                    (HGet pc h (SAMDynamicField (svalue2sloc val) f) (SVRef loc) --> 
                       wps_next pc s0 (h, (SVRef loc) :: os', l))
                    end in
                    *)
          let wp2 := let v := SHGet h (SAMDynamicField (svalue2sloc val) f) in 
                     wps_next pc s0 (h, v :: os', l)
          in
          wps_checkNull pc val O wp2 h l s0

        | Getstatic f, (h, os, l) =>
          let v := SHGet h (SAMStaticField f) in 
          SisStatic pc f --> wps_next pc s0 (h, v :: os, l)

        | I2b, (h, i :: os', l) => 
          (wps_next pc s0 (h, SVnum (SI2b i) :: os', l))
        | I2s, (h, i :: os', l) =>
          (wps_next pc s0 (h, SVnum (SI2s i) :: os', l))

        | Ibinop DivInt, (h, i2 :: i1 :: os', l) =>
          (wps_jvmExc (SNeg (SisZero pc i2))
            (wps_next pc s0 (h, SVnum (SIBinop DivInt i1 i2) :: os', l))
            pc ArithmeticException h l s0)
        | Ibinop RemInt, (h, i2 :: i1 :: os', l) =>
          (wps_jvmExc (SNeg (SisZero pc i2))
            (wps_next pc s0 (h, SVnum (SIBinop RemInt i1 i2) :: os', l))
            pc ArithmeticException h l s0)
        | Ibinop op, (h, i2 :: i1 :: os', l) =>
          (wps_next pc s0 (h, SVnum (SIBinop op i1 i2) :: os', l))

        | Iinc x z, (h, os, l) =>
(*          Forall i : BtInt,*)
          (SCheckByteSize pc z -->
(*            SLvGet l x i --> *)
              wps_next pc s0 (h, os, mkupdate l x (SVnum (SIBinop AddInt (SLvGet l x) (SVnum (SIVar (Int.const z)))))))

        | Ineg, (h, i :: os', l) =>
          (wps_next pc s0 (h, SVnum (SINeg i) :: os', l))

        | Instanceof t, (h, loc :: os', l) =>
          ((SisNull pc loc --> wps_next pc s0 (h, (SVnum (SIVar (Int.const 0))) :: os', l))
          & (! SisNull pc loc -->
                wps_cond (Sassign_compatible pc h loc (ReferenceType t))
                  (wps_next pc s0 (h, (SVnum (SIVar (Int.const 1))) :: os', l))
                  (wps_next pc s0 (h, (SVnum (SIVar (Int.const 0))) :: os', l))))

        | New c, (h, os, l) =>
          Forall loc : BtLocation,
          (Forall h' : BtHeap,
           (HNewObject pc h c loc h' -->
            wps_next pc s0 (h', SVRef loc :: os, l)))

        | Newarray t, (h, i :: os', l) => 
          (wps_jvmExc (SNonNegative pc i)
            (Forall loc : BtLocation,
            (Forall h' : BtHeap,
             (HNewArray pc h i t loc h' -->
               wps_next pc s0 (h', SVRef loc :: os', l))))
            pc NegativeArraySizeException h l s0)

        | Nop, s => wps_next pc s0 s

        | Pop, (h, v :: os, l) => wps_next pc s0 (h, os, l)
        | Pop2, (h, v1 :: v2 :: os, l) => wps_next pc s0 (h, os, l)

        | Putfield f, (h, v :: loc :: os', l) =>
	  let wp := ((Forall cn : BtClassName,
                 HTypeObject pc h loc cn -->
                 Sdefined_field pc cn f) -->
                 Sassign_compatible pc h v (FIELDSIGNATURE.type (snd f)) -->
                 wps_next pc s0 (SHUpdate h (SAMDynamicField (svalue2sloc loc) f) v, os', l)) in
	  wps_checkNull pc loc 1 wp h l s0

        | Putstatic f, (h, v :: os', l) =>
          (SisStatic pc f -->
           Sassign_compatible pc h v (FIELDSIGNATURE.type (snd f)) -->
           wps_next pc s0 (SHUpdate h (SAMStaticField f) v, os', l))

        | Swap, (h, v1 :: v2 :: os', l) =>
          (wps_next pc s0 (h, v2 :: v1 :: os', l))
 
        | Vaload k, (h, i :: loc :: os', l) =>
          ((SisNull pc loc --> wps_NullPointerException pc h l s0)
          & (SNeg (SisNull pc loc) -->
              (Forall length : BtInt,
              (Forall t : BtType,
               (HTypeArray pc h loc length t --> 
                (*Scompat_ArrayKind_type pc k t -->*)
                wps_bound i (SVnum length)
                  (let v := SHGet h (SAMArrayElement (svalue2sloc loc) i) in
                    (*Scompat_ArrayKind_value pc k v -->*)
                    wps_next pc s0 (h, SConv4Stack v :: os', l))
                pc h l s0)))))

        | Vastore k, (h, val :: i :: loc :: os', l) =>
          ((SisNull pc loc --> wps_NullPointerException pc h l s0)
          & (SNeg (SisNull pc loc) -->
              (Forall length : BtInt,
              (Forall t : BtType,
               (HTypeArray pc h loc length t -->
                (*Scompat_ArrayKind_type pc k t -->
                Scompat_ArrayKind_value pc k val -->*)
                wps_bound i (SVnum length)
                 (wps_jvmExc (Sassign_compatible pc h val t)
                   (wps_next pc s0
                     (SHUpdate h (SAMArrayElement (svalue2sloc loc) i) (SConv4Array val t), os', l))
                     pc ArrayStoreException h l s0)
                 pc h l s0)))))
                     
        | Vload k x, (h, os, l) =>
(*          Forall v : BtValue,*)
            (*Scompat_ValKind_value pc k (SLvGet l x) -->*)
            wps_next pc s0 (h, (SLvGet l x) :: os, l)

        | Vstore k x, (h, v :: os', l) =>
          ((*Scompat_ValKind_value pc k v --> *)wps_next pc s0 (h, os', mkupdate l x v))

        | Athrow, (h, loc :: os', l) =>
          ((SisNull pc loc --> wps_NullPointerException pc h l s0)
          & (SNeg (SisNull pc loc) --> 
          (Sassign_compatible pc h loc (ReferenceType (InterfaceType javaLangThrowable))) -->
              wps_Exception_unknown pc loc h l s0))
         
        | Vreturn k, (h, val :: os, l) =>
          match METHODSIGNATURE.result (METHOD.signature m) with
          | Some t => Sassign_compatible pc h val t -->
                      Scompat_ValKind_value pc k val -->
                      SPost pc None m s0 (h, SRVNormalSome val)
          | _ => STrue pc
          end

        | Return, (h, os, l) =>
          match METHODSIGNATURE.result (METHOD.signature m) with
          | None => SPost pc None m s0 (h, @SRVNormalNone SValue)
          | _ => STrue pc
          end

        | Goto o, _ => wps_label (OFFSET.jump pc o) s0 s

        | If_acmp cmp o, (h, val2 :: val1 :: os', l) =>
          (wps_cond (SCompRef pc cmp val1 val2)
            (wps_label (OFFSET.jump pc o) s0 (h, os', l))
            (wps_next pc s0 (h, os', l)))
        | If_icmp cmp o, (h, i2 :: i1 :: os', l) =>
          (wps_cond (SCompInt pc cmp i1 i2)
            (wps_label (OFFSET.jump pc o) s0 (h, os', l))
            (wps_next pc s0 (h, os', l)))
        | If0 cmp o, (h, i :: os', l) =>
          (wps_cond (SCompInt pc cmp i (SVnum (SIVar (Int.const 0))))
            (wps_label (OFFSET.jump pc o) s0 (h, os', l))
            (wps_next pc s0 (h, os', l)))
        | Ifnull cmp o, (h, loc :: os', l) =>
          (wps_cond (SCompRef pc cmp loc Null)
            (wps_label (OFFSET.jump pc o) s0 (h, os', l))
            (wps_next pc s0 (h, os', l)))

        | Invokestatic msig, (h, os, l) => 
          match findMethod p msig with
          | None => SFalse pc
          | Some m' =>
            if METHOD.isNative m' then SFalse pc
            else 
             match globalLookup p msig with
             | None => SFalse pc
             | Some (pairT pre post) =>
               let n := length (METHODSIGNATURE.parameters (snd msig)) in
  	       let (l', os') := init_args n os in
  	       (*pre (l', h) /\ *)
               match METHODSIGNATURE.result (snd msig) with
               | None =>
                 Forall rh : BtHeap,
                 (((SPre pc None msig (l',h) --> SPost pc None msig (l',h) (rh, @SRVNormalNone SValue)) -->
                  wps_next pc s0 (rh, os', l))
                 & (Forall loc : BtLocation,
                    (Forall cn : BtClassName,
                     (HTypeObject pc rh (SVRef loc) cn -->
                      (SPre pc None msig (l', h) --> SPost pc None msig (l', h) (rh, SRVException loc)) -->
                      wps_Exception_unknown pc loc rh l s0))))

               | Some _ =>
                 Forall rh : BtHeap,
                 ((Forall v : BtValue,
                   ((SPre pc None msig (l', h) --> SPost pc None msig (l', h) (rh, SRVNormalSome v)) -->
                    wps_next pc s0 (rh, v :: os', l)))
                 & (Forall loc : BtLocation,
                    (Forall cn : BtClassName,
                     (HTypeObject pc rh (SVRef loc) cn -->
                      (SPre pc None msig (l', h) --> SPost pc None msig (l', h) (rh, SRVException loc)) -->
                      wps_Exception_unknown pc loc rh l s0))))
                end
             end
          end 

        | Invokeinterface _, _ => SFalse pc
        | Invokespecial _, _ => SFalse pc
        | Invokevirtual msig, (h, os, l) =>
          match findMethod p msig with
          | None => SFalse pc
          | Some m' =>
            if METHOD.isNative m' then SFalse pc
            else 
             match globalLookup p msig with
             | None => SFalse pc
             | Some (pairT pre post) =>
               let n := length (METHODSIGNATURE.parameters (snd msig)) in
               match pop _ n os with 
               v :: _ => 
               let wp := 
                  (let (l', os') := init_args (n+1) os in
                   match METHODSIGNATURE.result (snd msig) with
                   | None =>
                     Forall rh : BtHeap,
                      (((SPre pc None msig (l',h) --> SPost pc None msig (l', h) (rh, @SRVNormalNone SValue)) -->
                         wps_next pc s0 (rh, os', l))
                       & (Forall loc : BtLocation,
                            (* (Forall cn : BtClassName,
                            (HTypeObject pc rh (SVRef loc) cn)) --> *)
                             (SPre pc None msig (l', h) --> SPost pc None msig (l', h) (rh, SRVException loc)) -->
                             wps_Exception_unknown pc loc rh l s0))
                   | Some (PrimitiveType _) => 
                     Forall rh : BtHeap,
                     (Forall i : BtInt,
                      (SPre pc None msig (l',h) --> SPost pc None msig (l', h) (rh, (SRVNormalSome (SVnum (SnI i))))) -->
                         wps_next pc s0 (rh, (SVnum (SnI i)) :: os', l))
                       & (Forall loc : BtLocation,
                          (* Forall cn : BtClassName,
                            HTypeObject pc rh (SVRef loc) cn --> *)
                            (SPre pc None msig (l', h) --> SPost pc None msig (l', h) (rh, SRVException loc)) -->
                             wps_Exception_unknown pc loc rh l s0)
                   | Some (ReferenceType _) => 
                     Forall rh : BtHeap,
                     (Forall lr : BtLocation,
                      (SPre pc None msig (l',h) --> SPost pc None msig (l', h) (rh, (SRVNormalSome (SVRef lr)))) -->
                         wps_next pc s0 (rh, (SVRef lr) :: os', l))
                       & (Forall loc : BtLocation,
                          (* Forall cn : BtClassName,
                            HTypeObject pc rh (SVRef loc) cn --> *)
                            (SPre pc None msig (l', h) --> SPost pc None msig (l', h) (rh, SRVException loc)) -->
                             wps_Exception_unknown pc loc rh l s0)
                   end) in
	       wps_checkNull pc v n wp h l s0 
               | _ => STrue pc
               end
             end
          end 
        | Lookupswitch _ _ , _=> SFalse pc
        | Tableswitch _ _ _ _, _ => SFalse pc
 
        | _, _ => STrue pc
        end
      | None => SFalse pc
      end.

    Lemma instrAt_B : 
      forall pc ins, instructionAt m pc = Some ins -> 
        BYTECODEMETHOD.instructionAt m pc = Some ins.
    Proof.
     intros pc ins; unfold instructionAt.
     rewrite isValidAm.(vam_Bytecode);trivial.
    Qed.

    Ltac instrAtB :=
      match goal with
      | [H: instructionAt _ _ = _ |- _ ] =>
        try rewrite (instrAt_B _ _ H)
      | _ => idtac
      end.
     


(*    Ltac wps_imp := 
      match goal with
      | [H1:wps_bound ?i ?length _ _ _ _ _, 
        H2:(0 <= Int.toZ ?i < Int.toZ ?length) |- _] =>
        let HH := fresh "H" in
        assert (HH:= wp_bound_true _ _ _ _ _ _ _ H1 H2);clear H1;wps_imp
      | [H1:wps_bound ?i ?length _ _ _ _ _, 
         H2:~(0 <= Int.toZ ?i < Int.toZ ?length) |- _] =>
         let HH := fresh "H" in
         assert (HH:= wp_bound_false _ _ _ _ _ _ _ H1 H2);clear H1;wps_imp
      | [H1:wps_jvmExc ?cond _ _ _ _ _ _, H2 : ?cond|- _] =>
         let HH := fresh "H" in
         assert (HH:= wp_jvmExc_true _ _ _ _ _ _ _ H1 H2);clear H1;wps_imp
      | [H1:wps_jvmExc ?cond _ _ _ _ _ _,H2:~?cond |- _] =>
         let HH := fresh "H" in
         assert (HH:=wp_jvmExc_false _ _ _ _ _ _ _ H1 H2);clear H1;wps_imp
      | [H1:wps_cond ?cond _ _, H2 : ?cond |- _] =>
         let HH := fresh "H" in
         assert (HH:=wp_cond_true _ _ _ H1 H2);clear H1;wps_imp
      | [H1:wps_cond ?cond _ _, H2 : ~?cond |- _] =>
         let HH := fresh "H" in
         assert (HH:=wp_cond_false _ _ _ H1 H2);clear H1;wps_imp
      | [H1:interp_p (wps_jvmExc ?cond _ _ _ _ _ _),H2:~?cond |- _] =>
         let HH := fresh "H" in
         assert (HH:=wp_jvmExc_false _ _ _ _ _ _ _ H1 H2);clear H1;wps_imp
      (*| [H1:wps_next ?pc _ (?h, ?os, ?l), 
         H2:next _ ?pc = Some ?pc' |- wps_label ?pc' _ (?h1, ?os1, ?l1)] =>
         apply (wp_next_wp_label _ _ _ _ H1 H2)
      | [H1:wps_next ?pc _ _, H2: next _ ?pc = Some _ |- _] =>
         let HH := fresh "H" in
         assert (HH := wp_next_wp_label _ _ _ _ H1 H2);clear H1; wps_imp   *)
      | [H1:Heap.new _ _ (Heap.LocationObject (javaLang, ?e))=Some (?loc2, _),
         H2:UnCaughtException _ _ (?pc, _, ?loc2),
         H3:wps_jvmException ?pc ?e _ _ _ |-am_post _ _ (_, Exception ?loc2)]=>
          apply (wp_jvmException_uncaught _ _ _ _ _ _ _ H1 H2 H3)
      | [H1:Dom.lookup_handlers _ _ ?pc (javaLang,?e) ?pc',
         H2:Heap.new _ _ (Heap.LocationObject (javaLang,?e))= _,
         H3:interp_p (wps_jvmException ?pc ?e _ _ _) |- interp_p (wps_label ?pc' _ _)] =>
         apply (wp_jvmException_caught _ _ _ _ _ _ _ _ H1 H2 H3)
      | [H1:Dom.lookup_handlers _ _ ?pc (javaLang, NullPointerException) ?pc',
         H2:Heap.new _ _ (Heap.LocationObject(_, NullPointerException)) = _,
         H3:wps_NullPointerException ?pc _ _ _ |- wps_label ?pc' _ _] =>
         apply (wp_NullPointerException_caught _ _ _ _ _ _ _ H1 H2 H3)
      | [H1:Heap.new _ _ (Heap.LocationObject (_, NullPointerException)) = 
               Some (?loc2, _),
         H2:UnCaughtException _ _ (?pc, _, ?loc2),
         H3:wps_NullPointerException ?pc _ _ _ |- 
            am_post _ _ (_, Exception ?loc2) ] =>
         apply (wp_NullPointerException_uncaught _ _ _ _ _ _ H1 H2 H3)
      | [H1:Dom.lookup_handlers _ _ ?pc _ ?pc', 
         H2:wps_Exception ?pc _ _ _ _ _ |- wps_label ?pc' _ _] =>
         apply (wp_Exception_caught _ _ _ _ _ _ _ H1 H2)
      | [H:Ref _ <> Null -> ?H2 |- _] =>
        assert H2; [ apply H; unfold interp_value; simpl; 
                     intro; discriminate | clear H ]; wps_imp
      | [H:Null = Null -> ?H2 |- _] =>
        assert H2; [ apply H; trivial | clear H ]; wps_imp
      | [H:Ref _ = Null -> _ |- _] => clear H;wps_imp
      | [H1:?a, H2:?a->?b |- _ ] => assert b; [ apply H2; trivial | idtac ]; clear H2; wps_imp
      (*| [H1:assign_compatible _ _ _ _,
         H2:assign_compatible _ _ _ _ -> ?HH,
         H3:~assign_compatible _ _ _ _ -> _ |- _ ] =>
         clear H3; assert HH; [ apply H2; simpl; auto | clear H1 ]; wps_imp
      | [H1:isStatic ?p ?f,
         H2:isStatic ?p ?f -> ?HH |- _ ] =>
         assert HH; [ apply H2; trivial | clear H2 ]; wps_imp*)
      (*| [ H1 :next _ ?pc = Some ?pc',
          H2:_ -> interp_p (wps_next ?pc ?s0 (?h,?os,?l)) |- interp_p (wps_label ?pc' ?s0 (?h1,?os1,?l1)) ] => 
          apply (wps_l_inv pc' s0 h h1 os os1 l l1); auto; 
          let HH := fresh in 
          assert (HH : interp_p (wps_next pc s0 (h,os,l)));
          [ apply H2 | apply (wp_next_wp_label _ _ _ _ HH H1); auto ]*)
      (*| [ H1 :next _ ?pc = Some ?pc',
          H2:interp_p (wps_next ?pc ?s0 (?h,?os,?l)) |- interp_p (wps_label ?pc' ?s0 (?h1,?os1,?l1)) ] => 
          apply (wps_l_inv pc' s0 h h1 os os1 l l1); simpl; [ auto | auto | auto | apply (wp_next_wp_label _ _ _ _ H2 H1); auto ]
      | [ H1 :next _ ?pc = Some ?pc',
          H2:interp_p (wps_next ?pc ?s0 (?h,?os,?l)) |- interp_p (wps_label ?pc' ?s0 (?h1,?os1,?l1)) ] => 
          apply (wps_l_inv pc' s0 h h1 os os1 l l1); auto; apply (wp_next_wp_label _ _ _ _ H2 H1); auto*)
      | [ |- interp_operandstack _ = interp_operandstack _ ] => simpl; auto
      | [ |- interp_heap _ = interp_heap _ ] => simpl; auto
      | [ |- interp_localvar _ = interp_localvar _ ] => simpl; auto
      | [ H: _ -> interp_p (wps_next ?pc ?s0 (?h,?os,?l)) |- interp_p (wps_next ?pc ?s0 (?h1,?os1,?l1)) ] => apply (wps_next_inv pc s0 h h1 os os1 l l1); eauto
      | [ H: interp_p (wps_next ?pc ?s0 (?h,?os,?l)) |- interp_p (wps_next ?pc ?s0 (?h1,?os1,?l1)) ] => apply (wps_next_inv pc s0 h h1 os os1 l l1); eauto
      | [ H: SFalse _ |- _ ] => simpl in H; elim H
      | [ H : _ /\ _ |- _ ] => elim H; intros; clear H; wps_imp
      | _ => trivial; try (eapply wp_next_wp_label;eauto;fail)
      end.

     Ltac sympl_hyps :=
     match goal with
     | [ H : _ /\ _ |- _ ] => elim H; intros; clear H
     | _ => idtac
     end.
*)

(*
    Lemma wp_intra : 
      forall s0 pc h os l pc' h' os' l', 
        NormalStep p m (pc,(h,os,l)) (pc',(h',os',l')) ->
        wps_i pc s0 (SHVar h, SSVar os, SLvVar l) ->
        wps_label pc' s0 (SHVar h', SSVar os',SLvVar l').
*)

(*
    Lemma wp_jvmExc_correct : 
      forall pc s0 h os l e loc2 h2 pc',
        JVMExceptionStep p m (pc, (h, os, l)) e ->
        Heap.new h p (Heap.LocationObject (javaLang, e)) = Some (loc2, h2) ->
        Dom.lookup_handlers p m pc (javaLang, e) pc' ->
        Heap.typeof h2 loc2 = Some (Heap.LocationObject (javaLang,e)) ->
        wps_i pc s0 (SHVar h, SSVar os, SLvVar l) ->
        wps_label pc' s0 (SHVar h2, SVRef (SLVar loc2) :: nil, SLvVar l).
    Proof with wps_imp.
*)
  
(*
    Lemma wp_exc : forall pc pc' h os l h2 loc2 e s0,
      ExceptionStep p m (pc, (h, os, l)) (h2, loc2) ->
      Dom.lookup_handlers p (BYTECODEMETHOD.exceptionHandlers m) pc e pc' ->
      Heap.typeof h2 loc2 = Some (Heap.LocationObject e) ->
      wps_i pc s0 (SHVar h, SSVar os, SLvVar l) ->
      wps_label pc' s0 (SHVar h2, SVRef (SLVar loc2) :: nil, SLvVar l).
    Proof.
*)

(*
    Lemma wp_jvmExc_uncaught :
      forall pc s0 h s l e loc2 h2,
        JVMExceptionStep p m (pc, (h,s,l)) e ->
        Heap.new h p (Heap.LocationObject (javaLang, e)) = Some (loc2, h2) ->
        UnCaughtException p m (pc, h2, loc2) ->
        wps_i pc s0 (SHVar h, SSVar s, SLvVar l) -> 
        SPost pc None m s0 (SHVar h2, SRVException (SLVar loc2)).
    Proof with wps_imp.
*)

(*
    Lemma wp_exc_uncaught :
      forall pc s0 h os l h2 loc2,
        ExceptionStep p m (pc, (h,os,l)) (h2, loc2) ->
        UnCaughtException p m (pc, h2, loc2) ->
        wps_i pc s0 (SHVar h,SSVar os,SLvVar l) ->
        SPost pc None m s0 (SHVar h2, SRVException (SLVar loc2)).
    Proof with wps_imp.
*)

(*
    Lemma wp_return_post : 
      forall s0 pc h os l h1 rv,  
        exec_return p m (pc, (h,os,l)) (h1,rv) ->
        wps_i pc s0 (SHVar h,SSVar os,SLvVar l) ->
        SPost pc None m s0 (SHVar h1, SRVVar rv).
    Proof with wps_imp.
*)

   End ____WPS_INSTR____. 

   Section ____WPS_LABEL____.

    Variable wps_i : (PC -> SAssertion) -> PC -> SAssertion.

    Fixpoint wps_l (n:nat) (pc:PC) {struct n}: SAssertion :=
      match assertLookup m.(am_local) pc with
      | Some _ => SAssert pc None m pc
      | None => 
        match n with
        | O => fun s0 s => SFalse pc
        | S n' => wps_i (wps_l n') pc
        end
      end.

   End ____WPS_LABEL____.

   Definition wps_label := wps_l wps_i.

   Definition wps_instr n := wps_i (wps_label n).


   Parameter valid_assert_pc : PC.
   
   Fixpoint valid_assert (lms:list (PC**((InitState -> LocalState -> list Prop) **nat))) : SWp :=
     match lms with
     | cons (pair pc (pair _ n)) lms' =>
       (Forall l0 : BtLocalVar,
       Forall h0 : BtHeap,
       Forall l : BtLocalVar,
       Forall h : BtHeap,
       Forall s : BtOperandStack,
            (SAssert pc (Some (wps_instr n pc (l0,h0) (h,s,l))) m pc (l0,h0) (h,s,l)))
         & valid_assert lms'
     | nil =>
       Forall l : BtLocalVar,
       Forall h : BtHeap,
        (SPre valid_assert_pc (Some (wps_label m.(am_needed) (BYTECODEMETHOD.firstAddress m) (l,h) (h,nil,l)))
              m (l,h))
     end.
   
   Lemma valid_assert_spec : 
     forall lms, valid_assert lms ->
     forall pc A n, In (pc,,(A,,n)) lms -> 
     forall l0 h0 h os l, SAssert pc (Some (wps_instr n pc (SLvVar l0,SHVar h0) (SHVar h,SSVar os,SLvVar l))) m pc (SLvVar l0,SHVar h0) (SHVar h,SSVar os,SLvVar l).
   Proof.
    induction lms;intros.
    inversion H0.
    simpl in H0.
    destruct H0.
      subst.
      simpl in H.
      destruct H.
      assert (H1 := H l0 h0 l h os); clear H.
      apply H1.
    eapply IHlms; eauto.
    simpl in H.
    case_eq a;intros; subst.
    case_eq p1;intros;subst.
    simpl in H.
    destruct H.
    trivial.
   Qed.

   Lemma valid_assert_pre :
     forall lms, valid_assert lms ->
     forall l h, SPre valid_assert_pc 
       (Some (wps_label m.(am_needed) 
            (BYTECODEMETHOD.firstAddress m) (SLvVar l,SHVar h) (SHVar h,nil,SLvVar l)))
            m (SLvVar l,SHVar h).
   Proof.
    induction lms. simpl. intros. auto.
    intros.
    destruct a as (pc,(A,n)).
    simpl in H.
    eapply IHlms.
    destruct H.
    trivial.

   Qed.


  End ____WPS_METH____.

  Definition annotMeth ms m n pre post lms :=
    match METHOD.body m with
    | Some bm =>
      Some 
        (AnnoMeth n
            ms m bm (BYTECODEMETHOD.exceptionHandlers bm)
            pre post lms)
    | None => None 
    end.

  Parameter certifiedMethod_pc : PC.

  Definition certifiedMethod ms m (anno:nat ** GlobalMethSpec ** LocMethSpec) :=
    match anno with
    | (pair (pair n (pair pre post)) lms) => 
      match annotMeth ms m n pre post lms with
      | Some am => valid_assert am (PCM.elements (fst am.(am_local)))
      | None => SFalse certifiedMethod_pc
      end
    end.

(*  
   Lemma certifiedMethod_spec : 
    forall (ms:MethodSignature) 
           (m:Method) (n:nat) pre post lms, 
             certifiedMethod ms m ((n,,(pre,,post)),,lms) ->
      exists am,
        valid_am am /\ m = am /\ am.(am_pre) = pre /\
        am.(am_post) = post /\ am.(am_local) = lms /\
        valid_assert am (PCM.elements (fst am.(am_local))).
  Proof.
*)

  Parameter allcertMeth_pc : PC.

  Fixpoint allCertMeth 
    (mst: list (MethodSignature**(nat ** GlobalMethSpec ** LocMethSpec))) : SWp :=
    match mst with
    | cons (msig,, anno) mst' => 
         match findMethod p msig with
         | None => SFalse allcertMeth_pc
         | Some m => (certifiedMethod msig m anno) & (allCertMeth mst')
         end
    | nil => STrue allcertMeth_pc
    end.

  Lemma allCertMeth_spec : 
    forall mst, allCertMeth mst ->  
    forall msig anno, In (msig,,anno) mst -> 
    exists m, findMethod p msig = Some m /\ certifiedMethod msig m anno.
  Proof.
   induction mst;simpl;intros;elimFalse.
   destruct H0.
   subst. generalize H;clear H;case_eq (findMethod p msig);intros;elimFalse.
   destruct H0;exists m;auto.
   elim H0.
   destruct a. generalize H;clear H;case_eq (findMethod p m);intros;elimFalse.
   destruct H1;auto.
   elim H1.
  Qed. 


  Definition certifiedProg := allCertMeth (MM.elements p.(ap_mst)).

  Lemma certifiedProg_spec : 
    certifiedProg ->
    forall msig m pre post, 
      findMethod p msig = Some m ->
      globalLookup p msig = Some (pre,,post) ->
      exists n, exists lms, certifiedMethod msig m ((n,,(pre,,post)),,lms).
  Proof.
   intros Hcertif msig m pre post Hfind; unfold globalLookup.
   case_eq (MM.get p.(ap_mst) msig).
   intros ((n,(pre',post')),lms) H1 H2;inversion_mine H2.
   exists n;exists lms. 
   destruct (allCertMeth_spec _ Hcertif _ _ (MM.get_some_in_elements _ _ _ _ H1)) as (m',(U1,U2)).
   rewrite Hfind in U1. injection U1;intros;clear U1;subst;trivial.
   intros;discriminate.
  Qed.
  
  Section ____WPS_PROOF____.
   Hypothesis IsCertProp : certifiedProg.

   Lemma interp_assume_interp_assert : forall c xs,
     interp_assume xs c -> interp_assert xs -> c.
   Proof.
     induction xs.
     trivial.
     intros.
     simpl in H0.
     case_eq xs.
     intros; subst.
     simpl in H.
     auto.
     intros.
     subst.
     replace (interp_assume (a :: P :: l) c) with (a -> interp_assume (P :: l) c) in H by trivial.
     destruct H0; auto.
   Qed.

   Lemma wp_label_wp_instr :        
     forall am, valid_am am ->  
       valid_assert am (PCM.elements (fst am.(am_local))) ->
     forall n l0 h0 pc h os l, wps_label am n pc (SLvVar l0,SHVar h0) (SHVar h,SSVar os,SLvVar l) ->
     exists n', wps_instr am n' pc (SLvVar l0,SHVar h0) (SHVar h,SSVar os,SLvVar l).
   Proof.
    intros am Hv Hva n l0 h0 pc h os l;destruct n;simpl; unfold assertLookup.
    case_eq (PCM.get (fst am.(am_local)) pc);intros;elimFalse.
    destruct p0 as (A,n).
    assert (HH := PCM.get_some_in_elements _ _ _ _ H).
    exists n. 
    generalize H0; clear H0.
    simpl.
    apply interp_assume_interp_assert.
    replace (interp_assume
         (localLookup p (am_meth_sign am) pc (l0,,h0) ((h,, interp_operandstack (SSVar os)),, l))
         (interp_p (wps_instr am n pc (SLvVar l0,, SHVar h0) ((SHVar h,, SSVar os),, SLvVar l)))) with
    (interp_p (SAssert pc (Some (wps_instr am n pc (SLvVar l0,SHVar h0) (SHVar h,SSVar os,SLvVar l))) am pc (SLvVar l0,SHVar h0) (SHVar h,SSVar os,SLvVar l))) by trivial.
    eapply valid_assert_spec;eauto.
    elim H0.
    case_eq (PCM.get (fst am.(am_local)) pc);intros.
    destruct p0 as (A,n').
    assert (HH := PCM.get_some_in_elements _ _ _ _ H).
    exists n'. 
    generalize H0; clear H0.
    simpl.
    apply interp_assume_interp_assert.
    replace (interp_assume
         (localLookup p (am_meth_sign am) pc (l0,,h0) ((h,, interp_operandstack (SSVar os)),, l))
         (interp_p (wps_instr am n' pc (SLvVar l0,, SHVar h0) ((SHVar h,, SSVar os),, SLvVar l)))) with
    (interp_p (SAssert pc (Some (wps_instr am n' pc (SLvVar l0,SHVar h0) (SHVar h,SSVar os,SLvVar l))) am pc (SLvVar l0,SHVar h0) (SHVar h,SSVar os,SLvVar l))) by trivial.
    eapply valid_assert_spec;eauto.
    exists n;exact H0.
   Qed.


   Hypothesis BST : 
         forall cn msig pre post, globalLookup p (cn,msig) = Some (pre,, post) ->
         forall cl m', lookup p cn msig (cl, m') -> 
         exists pre', exists post', 
             globalLookup p (cl, METHOD.signature m') = Some (pre',, post') /\
             forall s0 t, (interp_assert (pre' s0) -> interp_assert (post' s0 t)) -> 
                          (interp_assert (pre s0) -> interp_assert (post s0 t)).

  Lemma lookup_findMethod : 
     forall cn mid cl m , lookup p cn mid (cl,m) ->  
             findMethod p (cl,METHOD.signature m) = Some m.
  Proof.
      assert (forall cn mid M,
     lookup p cn mid M -> forall cl m, M =(cl, m) -> findMethod p (cl,METHOD.signature m) = Some m).
     induction 1;simpl;intros.
     injection H0;intros;subst.
     inversion_mine H. simpl in H1.
     destruct (PROG.class p cl);try discriminate.
     rewrite <-  (CLASS.method_signature_prop _ _ _ H1);trivial.
     exact (IHlookup _ _ H2).
     intros; eapply H;eauto.
  Qed.

  Definition slocalstate (s : LocalState) := match s with (h,os,l) => (@SHVar SValue h, SSVar os, @SLvVar SValue l) end.
  Definition sinitstate (s : InitState) := match s with (l,h) => (@SLvVar SValue l, @SHVar SValue h) end.
  Definition sreturnstate (s : ReturnState) := match s with (h,rv) => (@SHVar SValue h, @SRVVar SValue rv) end.

(*
   Lemma wp_label_correct : 
     forall m s sr, IntraStepStar p m s sr ->
     forall am, valid_am am -> m = am ->  
       valid_assert am (PCM.elements (fst am.(am_local))) ->
     forall n l0 h0, 
       wps_label am n (fst s) (sinitstate (l0,h0)) (slocalstate (snd s)) ->
       match sr with
       | inl (pc',s') => exists n', wps_label am n' pc' (sinitstate (l0,h0)) (slocalstate s')
       | inr r => interp_assert (am.(am_post) (l0,h0) r)
       end.
*)

  End ____WPS_PROOF____.

 End ____WPS____.


End SWp.

