Add LoadPath "../Bicolano".
Require Export ZArith.
Require Export Misc.
Require Export Tactic.
Require Export BigStep.
Require Export MapSignatures.
Require Export BaseLogic.

Module Wp (BL:RULES).
 Export BL V5 Dom Prog.

(** A deplacer *)
 Lemma isRef_notassig_Ref : 
     forall p h v t,
       isReference v -> 
        ~ assign_compatible p h v (ReferenceType t) ->
        exists loc, v = Ref loc.
 Proof.
  intros p0 h v t H;inversion_mine H;intros HH;eauto.   
  elim HH. constructor. 
 Qed.    
 
 Lemma lookup_unique :  
      forall p exl pc e pc' pc'',
            Dom.lookup_handlers p exl pc e pc' -> 
            Dom.lookup_handlers p exl pc e pc'' ->
            pc' = pc''.
 Proof.
  induction 1;intros HH;inversion_mine HH;simpl_eq;auto.
  elim (H2 H). elim (H H6).
 Qed. 

(** Fin a deplacer *)

 Definition GlobalMethSpec := BL.MethPre ** BL.MethPost.
 Definition LocMethSpec := PCM.t (BL.Assertion ** nat).
 
 Definition MethSpecTab := MM.t (nat**GlobalMethSpec**LocMethSpec).

 Definition assertLookup (lms:LocMethSpec) pc := 
   match PCM.get lms pc with
   | Some (pairT A _) => Some A
   | _ => None 
   end.

 Definition globalLookup (mst:MethSpecTab) msig := 
   match MM.get mst msig with
   | Some (pairT (pairT _ gms) _) => Some gms
   | None => None 
   end.

 (* Return the list of exception handler corresponding to [pc] *)
 Fixpoint handler_range (exl:list ExceptionHandler) (pc:PC) 
                               {struct exl} : list ExceptionHandler :=
   match exl with
   | nil => nil
   | ex :: exl => 
     if EXCEPTIONHANDLER.isPCinRange ex pc then
       ex :: handler_range exl pc 
     else handler_range exl pc
   end.

 Record annotatedProg : Type := AnnoProg {
   ap_prog     :> Prog.Program;
   ap_subclass : ClassName -> ClassName -> bool;
      (* A decidable version of subclass_name relation of p *)
   ap_mst      :> MethSpecTab
 }.

 Record annotatedMethod : Type := AnnoMeth {
   am_needed : nat;
   am_meth     :> Method;
   am_bytecode :> BytecodeMethod;
   am_excHand  :> list ExceptionHandler;
   am_pre      : BL.MethPre;
   am_post     : BL.MethPost;
   am_local    :> LocMethSpec
 }.

 (** Some decidables predicates *)

 Section ____WP____.

  Variable p:annotatedProg.
  
  Hypothesis ap_subclass_correct :
    forall cn cn', 
      if p.(ap_subclass) cn cn' then subclass_name p cn cn' 
      else ~subclass_name p cn cn'.

  Lemma ap_subclass_complete_t : 
    forall cn cn', subclass_name p cn cn' -> p.(ap_subclass) cn cn' = true.
  Proof.
   intros cn cn' H;assert (HH:= ap_subclass_correct cn cn');
    destruct (p.(ap_subclass) cn cn');trivial.  elim (HH H).
  Qed.
  
  Fixpoint lookup_handlers 
    (exl:list ExceptionHandler) (pc:PC) (cl:ClassName) {struct exl} : 
                                                               option PC :=
    match exl with
    | nil => None
    | ex :: exl' => 
      if EXCEPTIONHANDLER.isPCinRange ex pc then
	match EXCEPTIONHANDLER.catchType ex with
        | None => Some (EXCEPTIONHANDLER.handler ex)
        | Some cl' =>
	  if p.(ap_subclass) cl cl' then Some (EXCEPTIONHANDLER.handler ex)
	  else lookup_handlers exl' pc cl
        end
      else lookup_handlers exl' pc cl
    end.

  Section ____WP_METH____.
  
   Variable m:annotatedMethod.
   Record valid_am : Prop := mkVAM {
     vam_Bytecode : METHOD.body m = Some m.(am_bytecode);   
     vam_Handler : m.(am_excHand) = BYTECODEMETHOD.exceptionHandlers m
   }.
   Hypothesis isValidAm : valid_am.
 
   Lemma lookup_handler_correct :
     forall pc e,
       match lookup_handlers m pc e with
       | Some pc' => Dom.lookup_handlers p m pc e pc'
       | None => forall pc', ~Dom.lookup_handlers p m pc e pc'
       end.
   Proof.
    intros pc e;generalize m.(am_excHand).
    induction l;simpl;intros.
    intro H;inversion H.
    case_eq (EXCEPTIONHANDLER.isPCinRange a pc);intros Heq.
    case_eq (EXCEPTIONHANDLER.catchType a);intros.
    generalize (ap_subclass_correct e c);destruct (ap_subclass p e c);intros.
    econstructor;eauto. econstructor 2;eauto.
    destruct (lookup_handlers l pc e).
    econstructor 2;eauto.
    intro HH;inversion_mine HH;simpl_eq. elim (H0 H3).
    intros pc' HH;inversion_mine HH.
    inversion_mine H6;simpl_eq. elim (H0 H3). elim (IHl _ H7).
    econstructor;eauto. econstructor;eauto.
    destruct (lookup_handlers l pc e).
    econstructor 2;eauto.
    intro HH;inversion_mine HH;simpl_eq.
    intros pc' HH;inversion_mine HH.
    inversion_mine H4;simpl_eq. elim (IHl _ H5).
   Qed.
  
   Lemma lookup_complete_t : 
     forall pc e pc', Dom.lookup_handlers p m pc e pc' -> 
       lookup_handlers m pc e = Some pc'.
   Proof.
    intros;generalize (lookup_handler_correct pc e);
     destruct (lookup_handlers m pc e).
    intros Heq;rewrite (lookup_unique  _ _ _ _ _ _ H Heq);trivial.
    intros HH;elim (HH _ H).
   Qed.

   Section ____WP_INSTR____.
      
    Variable wp_label : PC -> BL.Assertion.

    Definition wp_next pc  :=
      match BYTECODEMETHOD.nextAddress m pc with
      | None => fun s0 s => True
      | Some pc' => wp_label pc'
      end.

    Lemma wp_next_wp_label : 
      forall pc pc' s0 s, 
        wp_next  pc s0 s ->
        next m pc = Some pc' ->
        wp_label pc' s0 s.
    Proof.
      unfold wp_next, next;intros pc pc' s0 s.
      rewrite isValidAm.(vam_Bytecode);intros H Heq;rewrite Heq in H;trivial.
    Qed.

   Definition wp_Exception pc loc cn h l s0 :=
      match lookup_handlers m pc cn with
      | None => (* UncaughtException *)
          m.(am_post) s0 (h, Dom.Exception loc)
      | Some pc' =>
 	  wp_label pc' s0 (h, Ref loc::OperandStack.empty,l)
      end.

  Lemma wp_Exception_caught :
      forall pc cn pc' loc2 h2 l s0,
        Dom.lookup_handlers p m pc cn pc' ->
        wp_Exception pc loc2 cn h2 l s0 ->
        wp_label pc' s0 (h2, Ref loc2 :: OperandStack.empty, l).
    Proof.
     unfold  wp_Exception;intros pc cn pc' loc2 h2 l s0 H1.
     rewrite (lookup_complete_t _ _ _ H1);auto.
    Qed.
 
   Lemma wp_Exception_uncaught :
      forall pc cn loc2 h2 l s0,
        Heap.typeof h2 loc2 = Some (Heap.LocationObject cn) ->
        UnCaughtException p m (pc, h2, loc2) ->
        wp_Exception pc loc2 cn h2 l s0 ->
        am_post m s0 (h2, Exception loc2).
    Proof.
     unfold wp_Exception;intros pc cn loc2 h2 h l s0 H1 H2.
     assert (UU:=lookup_handler_correct pc cn). 
     destruct (lookup_handlers m pc cn).
     inversion_mine H1.
     destruct isValidAm;simpl_eq.
     rewrite <- vam_Handler0 in H7;elim (H7 _ UU).
     trivial.
    Qed.

   Definition wp_jvmException pc e h l s0 :=  
       forall h' loc,
          Heap.new h p (Heap.LocationObject (javaLang,e)) = Some (loc,h') ->
          wp_Exception pc loc (javaLang,e) h' l s0.

    Lemma wp_jvmException_caught :
      forall pc e pc' loc2 h2 h l s0,
        Dom.lookup_handlers p m pc (javaLang, e) pc' ->
        Heap.new h p (Heap.LocationObject (javaLang, e)) = Some (loc2, h2) ->
        wp_jvmException pc e h l s0 ->
        wp_label pc' s0 (h2, Ref loc2 :: OperandStack.empty, l).
    Proof. unfold  wp_jvmException;intros;eapply  wp_Exception_caught;eauto. Qed.

    Lemma wp_jvmException_uncaught :
      forall pc e loc2 h2 h l s0,
        Heap.new h p (Heap.LocationObject (javaLang, e)) = Some (loc2, h2) ->
        UnCaughtException p m (pc, h2, loc2) ->
        wp_jvmException pc e h l s0 ->
        am_post m s0 (h2, Exception loc2).
    Proof. 
     unfold wp_jvmException;intros;eapply wp_Exception_uncaught;eauto. 
     assert (VV := Heap.new_typeof _ _ _ _ _ H);simpl_eq;trivial.
    Qed.

    Definition wp_NullPointerException pc :=
      wp_jvmException pc NullPointerException.

    Lemma wp_NullPointerException_caught :
      forall pc pc' loc2 h2 h l s0,
        Dom.lookup_handlers p m pc (javaLang, NullPointerException) pc' ->
        Heap.new h p (Heap.LocationObject (javaLang, NullPointerException)) =
           Some (loc2, h2) ->
        wp_NullPointerException pc h l s0 ->
        wp_label pc' s0 (h2, Ref loc2 :: OperandStack.empty, l).
    Proof.
     unfold wp_NullPointerException;intros;eapply wp_jvmException_caught;eauto.
    Qed.

    Lemma wp_NullPointerException_uncaught :
      forall pc  loc2 h2 h l s0,
        Heap.new h p (Heap.LocationObject (javaLang, NullPointerException)) =
           Some (loc2, h2) ->
       UnCaughtException p m (pc, h2, loc2) ->
       wp_NullPointerException pc h l s0 ->
       am_post m s0 (h2, Exception loc2).
    Proof.
     unfold wp_NullPointerException;intros;
      eapply wp_jvmException_uncaught;eauto.
    Qed.

    Definition wp_cond (cond wpT wpF:Prop) := (cond -> wpT) /\ (~cond -> wpF).

    Lemma wp_cond_true : 
      forall cond wpT wpF, wp_cond cond wpT wpF -> cond -> wpT.
    Proof. intros cond wpT wpF (H1,H2);auto. Qed.

    Lemma wp_cond_false : 
      forall cond wpT wpF, wp_cond cond wpT wpF -> ~cond -> wpF.
    Proof. intros cond wpT wpF (H1,H2);auto. Qed.
 
    Definition wp_jvmExc cond wpT pc e h l s0 :=
      wp_cond cond wpT (wp_jvmException pc e h l s0).

    Lemma wp_jvmExc_true : 
      forall cond wpT pc e h l s0, 
       wp_jvmExc cond wpT pc e h l s0 -> cond -> wpT.
    Proof.
     intros cond wpT pc w h l s0 H1 H2;apply (wp_cond_true _ _ _ H1 H2). 
    Qed.

    Lemma wp_jvmExc_false :
      forall cond wpT pc e h l s0,
        wp_jvmExc cond wpT pc e h l s0 -> ~cond ->
        wp_jvmException pc e h l s0.
    Proof.
     intros cond wpT pc w h l s0 H1 H2;apply (wp_cond_false _ _ _ H1 H2).   
    Qed.
 
    Definition wp_bound i length wpIn pc :=
      wp_jvmExc (0 <= Int.toZ i < Int.toZ length)
         wpIn pc ArrayIndexOutOfBoundsException.
 
    Lemma wp_bound_true :
      forall i length wpIn pc h l s0,
        wp_bound  i length wpIn pc h l s0 -> 
        (0 <= Int.toZ i < Int.toZ length) -> wpIn.
    Proof. 
     intros i length wpIn pc h l s0 H1 H2;
      exact (wp_jvmExc_true _ _ _ _ _ _ _ H1 H2). 
    Qed.

    Lemma wp_bound_false :
      forall i length wpIn pc h l s0,
        wp_bound  i length wpIn pc h l s0 -> 
        ~(0 <= Int.toZ i < Int.toZ length) -> 
        wp_jvmException pc ArrayIndexOutOfBoundsException h l s0.
    Proof. 
     intros i length wpIn pc h l s0 H1 H2;
      exact (wp_jvmExc_false _ _ _ _ _ _ _ H1 H2). 
    Qed.

    Fixpoint pop (A:Set) (n:nat) (l:list A) {struct n}: list A := 
      match n, l with
      | O, _ => l
      | S n, _ :: l => pop A n l
      | S n, nil => nil
      end.

    Lemma pop_append : 
      forall A n l1 l2, length l1 = n  -> pop A n (l1 ++ l2) = l2.
    Proof.
     induction n;simpl;intros.
     destruct l1;try discriminate;trivial.
     destruct l1;try discriminate;simpl.
     injection H;intros;auto.
    Qed.

    Definition init_args n os := (stack2localvar os n, pop value n os).

    Lemma init_args_correct :
      forall n args os l' os', 
        init_args n (args ++ os) = (l', os') ->
        length args = n  ->
        stack2localvar (args ++ os) (length args) = l' /\ os' = os.
    Proof.
     unfold init_args;simpl;intros.
     injection H;intros;subst.
     rewrite pop_append;auto.
    Qed.

    Definition wp_i pc s0 s :=
      match BYTECODEMETHOD.instructionAt m pc with
      | Some ins =>
(** Basic instruction *)
        match ins, s with
        | Aconst_null, (h, os, l) => wp_next pc s0 (h, Null :: os, l)

        | Arraylength, (h, Null :: os, l) => wp_NullPointerException pc h l s0
        | Arraylength, (h, Ref loc :: os, l) =>
          forall length tp, 
            Heap.typeof h loc = Some (Heap.LocationArray length tp) ->
            wp_next pc s0 (h, Num (I length) :: os, l)

        | Checkcast t, (h, val :: os, l)=>
          wp_jvmExc (assign_compatible p h val (ReferenceType t))
             (wp_next pc s0 (h, val :: os, l))
             pc ClassCastException h l s0

        | Const t z, (h, os, l) =>
          match t with
          | BYTE => -2^7  <= z < 2^7
          | SHORT=> -2^15 <= z < 2^15 
          | INT  => -2^31 <= z < 2^31 
          | _ => False
          end ->
          wp_next pc s0 (h, Num (I (Int.const z)) :: os, l)

        | Dup, (h, v :: os, l) => wp_next pc s0 (h, v :: v :: os, l)
        | Dup_x1, (h, v1 :: v2 :: os, l) => 
          wp_next pc s0 (h, v1 :: v2 :: v1 :: os, l)
        | Dup_x2, (h, v1 :: v2 :: v3 :: os, l) => 
          wp_next pc s0 (h, v1 :: v2 :: v3 :: v1 :: os, l)
        | Dup2, (h, v1 :: v2 :: os, l) =>
          wp_next pc s0 (h, v1 :: v2 :: v1 :: v2 :: os,l)
        | Dup2_x1, (h, v1 :: v2 :: v3 :: os, l)=>
          wp_next pc  s0 (h, v1 :: v2 :: v3 :: v1 :: v2 :: os,l)
        | Dup2_x2, (h, v1 :: v2 :: v3 :: v4 :: os, l)=>
          wp_next pc  s0 (h, v1 :: v2 :: v3 :: v4 :: v1 :: v2 :: os, l)
   
        | Getfield f, (h, Null :: os, l) => wp_NullPointerException pc h l s0
        | Getfield f, (h, Ref loc :: os, l) =>
          forall v,
            Heap.get h (Heap.DynamicField loc f) = Some v ->
            wp_next pc  s0 (h, v :: os, l)

        | Getstatic f, (h, os, l)=>
          forall v, 
            isStatic p f ->
  	  Heap.get h (Heap.StaticField f) = Some v ->
            wp_next pc  s0 (h, v :: os, l)

        | I2b, (h, Num (I i) :: os, l) =>
          wp_next pc  s0 (h, Num (I (b2i (i2b i))) :: os, l)
        | I2s, (h, Num (I i) :: os, l) =>
          wp_next pc  s0 (h, Num (I (s2i (i2s i))) :: os, l)

        | Ibinop DivInt, (h, Num (I i2) :: Num (I i1) :: os, l) =>
          wp_jvmExc (Int.toZ i2 <> 0)
            (wp_next pc s0 (h, Num (I (SemBinopInt DivInt i1 i2)) :: os, l))
            pc ArithmeticException h l s0
        | Ibinop RemInt, (h, Num (I i2) :: Num (I i1) :: os, l) =>
          wp_jvmExc (Int.toZ i2 <> 0)
            (wp_next pc s0 (h, Num (I (SemBinopInt RemInt i1 i2)) :: os, l))
            pc ArithmeticException h l s0
        | Ibinop op, (h, Num (I i2) :: Num (I i1) :: os, l) =>
          wp_next pc  s0 (h, Num (I (SemBinopInt op i1 i2)) :: os, l)

        | Iinc x z, (h, os, l) =>
          forall i, -2^7 <= z < 2^7 -> 
            METHOD.valid_var m x ->
            LocalVar.get l x = Some (Num (I i)) ->
            wp_next pc s0 
              (h, os, LocalVar.update l x (Num (I (Int.add i (Int.const z)))))

        | Ineg, (h, Num (I i) :: os, l) =>
          wp_next pc s0 (h, Num (I (Int.neg i)) :: os, l)

(** WARNING : No Null case *)
        | Instanceof t, (h, Ref loc :: os, l)=>  
          wp_cond (assign_compatible p h (Ref loc) (ReferenceType t))
            (wp_next pc s0 (h, Num (I (Int.const 1)) :: os, l))
            (wp_next pc s0 (h, Num (I (Int.const 0)) :: os, l))
        | Instanceof t, (h, Null :: os, l) =>
            wp_next pc s0 (h, Num (I (Int.const 0)) :: os, l)
           
        | New c, (h, os, l) =>
          forall loc h',
            Heap.new h p (Heap.LocationObject c) = Some (pair loc h') ->
            wp_next pc s0 (h', Ref loc :: os, l)

        | Newarray t, (h, Num (I i) :: os, l) =>
          wp_jvmExc (0 <= Int.toZ i)
            (forall loc h',
               Heap.new h p (Heap.LocationArray i t) = 
                 Some (pair loc h') ->
               wp_next pc s0 (h', Ref loc::os, l))
            pc NegativeArraySizeException h l s0

        | Nop, s => wp_next pc s0 s

        | Pop, (h, v :: os, l) => wp_next pc s0 (h,os,l)
        | Pop2, (h, v1 :: v2 :: os, l) => wp_next pc s0 (h, os, l)

        | Putfield f, (h, v :: Null :: os, l) =>
          wp_NullPointerException pc h l s0
        | Putfield f, (h, v :: Ref loc :: os, l) =>
          forall cn,
            Heap.typeof h loc = Some (Heap.LocationObject cn) -> 
            defined_field p cn f ->
            assign_compatible p h v (FIELDSIGNATURE.type (snd f)) ->
            wp_next pc s0 (Heap.update h (Heap.DynamicField loc f) v, os, l)

        | Putstatic f, (h, v :: os, l) =>
          isStatic p f ->
          assign_compatible p h v (FIELDSIGNATURE.type (snd f)) ->
          wp_next pc s0 (Heap.update h (Heap.StaticField f) v, os, l)

        | Swap, (h, v1 :: v2 :: os, l) => wp_next pc s0 (h, v2 :: v1 :: os, l) 
 
        | Vaload k, (h, Num (I i) :: Null :: os, l) =>      
           wp_NullPointerException pc h l s0

        | Vaload k, (h, Num (I i) :: (Ref loc) :: os, l) =>  
           forall length t,
              Heap.typeof h loc = Some (Heap.LocationArray length t) ->   
              compat_ArrayKind_type k t ->
              wp_bound i length        
              (forall v,
                 Heap.get h (Heap.ArrayElement loc i) = Some v -> 
                 compat_ArrayKind_value k v ->
                 wp_next pc s0 (h, conv_for_stack v :: os, l))
              pc h l s0
        | Vastore k, (h, val :: Num (I i) :: Null :: os, l) =>   
           wp_NullPointerException pc h l s0

        | Vastore k, (h, val :: Num (I i) :: Ref loc :: os, l) =>
           forall length t,
              Heap.typeof h loc = Some (Heap.LocationArray length t) ->
              compat_ArrayKind_type k t ->    
              compat_ArrayKind_value k val ->
              wp_bound i length    
               (wp_jvmExc (assign_compatible p h val t)
                  (wp_next pc s0 
                     (Heap.update h (Heap.ArrayElement loc i)  (conv_for_array val t), os, l))
                     pc ArrayStoreException h l s0)
               pc h l s0
        
        |  Vload k x, (h, os, l)=>
            METHOD.valid_var m x ->
            match LocalVar.get l x with
            | Some val =>
               compat_ValKind_value k val -> 
               wp_next pc s0 (h, val :: os, l)
            | _ => True
            end
        | Vstore k x, (h, v :: os, l) => 
           METHOD.valid_var m x ->
           compat_ValKind_value k v ->
           wp_next pc s0 (h, os, LocalVar.update l x v)

(** Athrow *)
        | Athrow, (h, Null :: os, l) => 
          wp_NullPointerException pc h l s0
        | Athrow, (h, Ref loc :: os, l) =>
  	  forall cn,
            Heap.typeof h loc = Some (Heap.LocationObject cn) ->
            p.(ap_subclass) cn javaLangThrowable = true ->
            wp_Exception pc loc cn h l s0
         
(** Return *)
        | Vreturn k, (h, val :: os, l) =>
          match METHODSIGNATURE.result (METHOD.signature m) with
          | Some t => 
            assign_compatible p h val t ->
            compat_ValKind_value k val ->
            m.(am_post) s0 (h, Dom.Normal (Some val))
          | _ => True
          end
       | Return, (h, os, l)  =>
          match  METHODSIGNATURE.result (METHOD.signature m) with
          | None => m.(am_post) s0 (h, Normal None)
          | _ => True
          end

(** Goto *)
        | Goto o, _ => wp_label (OFFSET.jump pc o) s0 s

(** Conditional *)
        | If_acmp cmp o, (h, val2 :: val1 :: os, l) => 
          wp_cond (SemCompRef cmp val1 val2)
           (wp_label (OFFSET.jump pc o) s0 (h, os, l))
           (wp_next pc s0 (h, os, l))
        | If_icmp cmp o, (h,  Num(I i2) :: Num(I i1) :: os, l) =>
          wp_cond (SemCompInt cmp (Int.toZ i1) (Int.toZ i2))
           (wp_label (OFFSET.jump pc o) s0 (h, os, l))
           (wp_next pc s0 (h, os, l))
        | If0 cmp o, (h, Num(I i) :: os,l) => 
          wp_cond (SemCompInt cmp (Int.toZ i) 0)
           (wp_label (OFFSET.jump pc o) s0 (h, os, l))
           (wp_next pc s0 (h, os, l))
        | Ifnull cmp o, (h, loc :: os, l) =>
          wp_cond (SemCompRef cmp loc Null)
           (wp_label (OFFSET.jump pc o) s0 (h, os, l))
           (wp_next pc s0 (h, os, l))

  (** Invokestatic *)
        | Invokestatic msig, (h, os, l)=> 
          match findMethod p msig with
          | None => False
          | Some m' =>
            if METHOD.isNative m' then False
            else 
             match globalLookup p msig with
             | None => False
             | Some (pairT pre post) =>
               let n := length (METHODSIGNATURE.parameters (snd msig)) in
  	       let (l', os') := init_args n os in
  	       (*pre (l', h) /\ *)
  	       forall rh rv, 
  	         match rv with
  	         | Normal ov =>
  	           (pre (l', h) -> post (l',h) (rh, rv)) -> 
                   wp_next pc s0 (rh, cons_option ov os', l)
                 | Exception loc =>
                   forall cn, 
                     Heap.typeof rh loc = Some (Heap.LocationObject cn) ->
                     (pre (l', h) -> post (l',h) (rh, rv)) ->
                     (*p.(ap_subclass) cn javaLangThrowable = true ->*)
                     wp_Exception pc loc cn rh l s0
                 end
             end
          end 

(** Not Defined instruction *)  
        | Invokeinterface _, _ => False
        | Invokespecial _, _ => False
        | Invokevirtual msig, (h,os,l) =>
          match findMethod p msig with
          | None => False
          | Some m' =>
            if METHOD.isNative m' then False
            else 
             match globalLookup p msig with
             | None => False
             | Some (pairT pre post) =>
               let n := length (METHODSIGNATURE.parameters (snd msig)) in
               match pop _ n os with
               | Ref _ :: _ =>  
  	          let (l', os') := init_args (n+1) os in
  	          (*pre (l', h) /\ *)
  	          forall rh rv, 
  	            match rv with
  	            | Normal ov =>
  	               (pre (l', h) -> post (l',h) (rh, rv)) -> 
                       wp_next pc s0 (rh, cons_option ov os', l)
                    | Exception loc =>
                       forall cn, 
                          Heap.typeof rh loc = Some (Heap.LocationObject cn) ->
                         (pre (l', h) -> post (l',h) (rh, rv)) ->
                         (*p.(ap_subclass) cn javaLangThrowable = true ->*)
                         wp_Exception pc loc cn rh l s0
                    end
               | Null :: _ => 
                  wp_NullPointerException pc h l s0
               | _ => True
                end
             end
          end 
        | Lookupswitch _ _, _ => False
        | Tableswitch _ _ _ _, _ => False

        | _, _ => True 
        end
      | None => False 
      end.

    Ltac wp_imp := 
      match goal with
      | [H1:wp_bound ?i ?length _ _ _ _ _, 
        H2:(0 <= Int.toZ ?i < Int.toZ ?length) |- _] =>
        let HH := fresh "H" in
        assert (HH:= wp_bound_true _ _ _ _ _ _ _ H1 H2);clear H1;wp_imp
      | [H1:wp_bound ?i ?length _ _ _ _ _, 
         H2:~(0 <= Int.toZ ?i < Int.toZ ?length) |- _] =>
         let HH := fresh "H" in
         assert (HH:= wp_bound_false _ _ _ _ _ _ _ H1 H2);clear H1;wp_imp
      | [H1:wp_jvmExc ?cond _ _ _ _ _ _, H2 : ?cond|- _] =>
         let HH := fresh "H" in
         assert (HH:= wp_jvmExc_true _ _ _ _ _ _ _ H1 H2);clear H1;wp_imp
      | [H1:wp_jvmExc ?cond _ _ _ _ _ _,H2:~?cond |- _] =>
         let HH := fresh "H" in
         assert (HH:=wp_jvmExc_false _ _ _ _ _ _ _ H1 H2);clear H1;wp_imp
      | [H1:wp_cond ?cond _ _, H2 : ?cond |- _] =>
         let HH := fresh "H" in
         assert (HH:=wp_cond_true _ _ _ H1 H2);clear H1;wp_imp
      | [H1:wp_cond ?cond _ _, H2 : ~?cond |- _] =>
         let HH := fresh "H" in
         assert (HH:=wp_cond_false _ _ _ H1 H2);clear H1;wp_imp
      | [H1:wp_jvmExc ?cond _ _ _ _ _ _,H2:~?cond |- _] =>
         let HH := fresh "H" in
         assert (HH:=wp_jvmExc_false _ _ _ _ _ _ _ H1 H2);clear H1;wp_imp
      | [H1:wp_next ?pc _ _, 
         H2:next _ ?pc = Some ?pc' |- wp_label ?pc' _ _] =>
         apply (wp_next_wp_label _ _ _ _ H1 H2)
      | [H1:wp_next ?pc _ _, H2: next _ ?pc = Some _ |- _] =>
         let HH := fresh "H" in
         assert (HH := wp_next_wp_label _ _ _ _ H1 H2);clear H1; wp_imp   
      | [H1:Heap.new _ _ (Heap.LocationObject (javaLang, ?e))=Some (?loc2, _),
         H2:UnCaughtException _ _ (?pc, _, ?loc2),
         H3:wp_jvmException ?pc ?e _ _ _ |-am_post _ _ (_, Exception ?loc2)]=>
          apply (wp_jvmException_uncaught _ _ _ _ _ _ _ H1 H2 H3)
      | [H1:Dom.lookup_handlers _ _ ?pc (javaLang,?e) ?pc',
         H2:Heap.new _ _ (Heap.LocationObject (javaLang,?e))= _,
         H3:wp_jvmException ?pc ?e _ _ _ |- wp_label ?pc' _ _] =>
         apply (wp_jvmException_caught _ _ _ _ _ _ _ _ H1 H2 H3)
      | [H1:Dom.lookup_handlers _ _ ?pc (javaLang, NullPointerException) ?pc',
         H2:Heap.new _ _ (Heap.LocationObject(_, NullPointerException)) = _,
         H3:wp_NullPointerException ?pc _ _ _ |- wp_label ?pc' _ _] =>
         apply (wp_NullPointerException_caught _ _ _ _ _ _ _ H1 H2 H3)
      | [H1:Heap.new _ _ (Heap.LocationObject (_, NullPointerException)) = 
               Some (?loc2, _),
         H2:UnCaughtException _ _ (?pc, _, ?loc2),
         H3:wp_NullPointerException ?pc _ _ _ |- 
            am_post _ _ (_, Exception ?loc2) ] =>
         apply (wp_NullPointerException_uncaught _ _ _ _ _ _ H1 H2 H3)
      | [H1:Dom.lookup_handlers _ _ ?pc _ ?pc', 
         H2:wp_Exception ?pc _ _ _ _ _ |- wp_label ?pc' _ _] =>
         apply (wp_Exception_caught _ _ _ _ _ _ _ H1 H2)
      | _ => trivial; try (eapply wp_next_wp_label;eauto;fail)
      end.

    Lemma instrAt_B : 
      forall pc ins, instructionAt m pc = Some ins -> 
        BYTECODEMETHOD.instructionAt m pc = Some ins.
    Proof.
     intros pc ins; unfold instructionAt.
     rewrite isValidAm.(vam_Bytecode);trivial.
    Qed.

    Ltac instrAtB :=
      match goal with
      | [H: instructionAt _ _ = _ |- _ ] =>
        try rewrite (instrAt_B _ _ H)
      | _ => idtac
      end.
 
    Lemma wp_intra : 
      forall s0 pc s pc' s', 
        NormalStep p m (pc,s) (pc',s') ->
        wp_i pc s0 s ->
        wp_label pc' s0 s'.
    Proof with wp_imp.
     intros s0 pc s pc' s' Hel;inversion_mine Hel;
        unfold wp_i;instrAtB;simpl;subst;simpl;intros;elimFalse ...
     assert  (wp_next pc s0 (h, Num (I (Int.const z)) :: s1, l)) ...
     apply H;destruct H6 as [(Heq, HH) | [(Heq, HH) | (Heq, HH) ]]; 
       rewrite Heq;exact HH.
     destruct op ...
     assert (UU := H6 (or_introl _ (refl_equal DivInt))) ...
     assert (UU := H6 (or_intror _ (refl_equal RemInt))) ...
     destruct H7. 
     destruct(isRef_notassig_Ref _ _ _ _ H6 H0) as (loc, Heq);rewrite Heq in H;
      rewrite Heq in H0...
     subst ...
     rewrite H4 in H;rewrite H5 in H;elim H.
     assert (UU := H _ _ H5 H7) ...
     assert (UU := H _ _ H5 H9 H10) ...
     rewrite H7 in H ...
    Qed.

    Lemma wp_jvmExc_correct : 
      forall pc s0 h os l e loc2 h2 pc',
        JVMExceptionStep p m (pc, (h, os, l)) e ->
        Heap.new h p (Heap.LocationObject (javaLang, e)) = Some (loc2, h2) ->
        Dom.lookup_handlers p m pc (javaLang, e) pc' ->
        Heap.typeof h2 loc2 = Some (Heap.LocationObject (javaLang,e)) ->
        wp_i pc s0 (h, os, l) ->
        wp_label pc' s0 (h2, Ref loc2 :: OperandStack.empty, l).
    Proof with wp_imp.
     intros pc s0 h os l e loc2 h2 pc' Hel;inversion_mine Hel;
      unfold wp_i;instrAtB;simpl;subst;simpl;intros;elimFalse;simpl_eq...
     assert (~ (Int.toZ i2 <> 0));[omega|destruct H6;subst;wp_imp].
     generalize H2;clear H2.
     case_eq (findMethod p mid);[intros m' Heq | intros Heq].
     case_eq (METHOD.isNative m');intros Heq1.     intro HF;elim HF.
     case_eq (globalLookup p mid);[intros (pre,post) Heq2 | intros Heq2 HF;elim HF].
     rewrite pop_append;trivial.     intros...     intros HF;elim HF.
     assert (UU := H2 _ _ H6 H7) ...
     assert (UU := H2 _ _ H5 H8 H9) ...
     assert (UU := H2 _ _ H4 H9 H10) ...
    Qed.

    Lemma wp_exc : forall pc pc' h os l h2 loc2 e s0,
      ExceptionStep p m (pc, (h, os, l)) (h2, loc2) ->
      Dom.lookup_handlers p (BYTECODEMETHOD.exceptionHandlers m) pc e pc' ->
      Heap.typeof h2 loc2 = Some (Heap.LocationObject e) ->
      wp_i pc s0 (h, os, l) ->
      wp_label pc' s0 (h2, Ref loc2 :: OperandStack.empty, l).
    Proof.
     intros pc pc' h os l h2 loc2 e s0 Hel;inversion_mine Hel.
     unfold wp_i;instrAtB;simpl;subst;simpl;intros;elimFalse;simpl_eq. 
     assert (UU:= H1 _ H7  (ap_subclass_complete_t _ _ H8)). 
     eapply wp_Exception_caught;eauto.
     rewrite isValidAm.(vam_Handler);trivial.
     unfold wp_i;instrAtB;simpl;subst;simpl;intros;elimFalse;simpl_eq. 
     simplIf;elimFalse.
     intros;assert (HH:= Heap.new_typeof _ _ _ _ _  H7);simpl_eq.
     eapply wp_jvmExc_correct;eauto.
     rewrite  isValidAm.(vam_Handler);trivial.
    Qed.

    Lemma wp_jvmExc_uncaught :
      forall pc s0 h s l e loc2 h2,
        JVMExceptionStep p m (pc, (h,s,l)) e ->
        Heap.new h p (Heap.LocationObject (javaLang, e)) = Some (loc2, h2) ->
        UnCaughtException p m (pc, h2, loc2) ->
        wp_i pc s0 (h, s, l) -> 
        am_post m s0 (h2, Exception loc2).
    Proof with wp_imp.
     intros pc s0 h os l e loc2 h2 Hel;inversion_mine Hel;
      unfold wp_i;instrAtB;simpl;subst;simpl;intros;elimFalse;simpl_eq...
     assert (~(Int.toZ i2 <> 0)). omega. destruct H6;subst ...
     generalize H1;clear H1.
     case_eq (findMethod p mid);[intros m' Heq | intros Heq].
     case_eq (METHOD.isNative m');intros Heq1.     intro HF;elim HF.
     case_eq (globalLookup p mid);[intros (pre,post) Heq2 | intros Heq2 HF;elim HF].
     rewrite pop_append;trivial.     intros...     intros HF;elim HF.
     assert (UU:= H1 _ _ H6 H7) ...
     assert (UU:= H1 _ _ H5 H8 H9) ...
     assert (UU:= H1 _ _ H4 H9 H10) ...
    Qed.
     
    Lemma wp_exc_uncaught :
      forall pc s0 s h2 loc2,
        ExceptionStep p m (pc, s) (h2, loc2) ->
        UnCaughtException p m (pc, h2, loc2) ->
        wp_i pc s0 s ->
        m.(am_post) s0 (h2, Exception loc2).
    Proof.
     intros pc s0 s h2 loc2 Hel;inversion_mine Hel.
     unfold wp_i;simpl;instrAtB;simpl;intros;simplIf;intros.
     assert (HH := H0 _ H5 (ap_subclass_complete_t _ _ H6));clear H0.
     eapply wp_Exception_uncaught;eauto.
     unfold wp_i;simpl;instrAtB;simpl;intros;simplIf;elimFalse.
     intros;eapply wp_jvmExc_uncaught;eauto.
    Qed.

    Lemma wp_return_post : 
      forall s0 s r,  
        exec_return p m s r ->
        wp_i  (fst s) s0 (snd s) ->
        m.(am_post) s0 r.
    Proof with wp_imp.
     intros s0 s r Hr;inversion_mine Hr ...
     inversion_mine H;
        try (unfold wp_i;simpl;instrAtB;simpl;intros;simplIf;auto;fail).
     intros;eapply wp_exc_uncaught;eauto.
    Qed.

   End ____WP_INSTR____.  

   Section ____WP_LABEL____.

    Variable wp_i : (PC -> BL.Assertion) -> PC -> BL.Assertion.

    Fixpoint wp_l (n:nat) (pc:PC) {struct n}: BL.Assertion :=
      match assertLookup m.(am_local) pc with
      | Some a => a
      | None => 
        match n with
        | O => fun s0 s => False
        | S n' => wp_i (wp_l n') pc
        end
      end.

   End ____WP_LABEL____.

   Definition wp_label := wp_l wp_i.

   Definition wp_instr n := wp_i (wp_label n).
   
   Fixpoint valid_assert (lms:list (PC**(BL.Assertion**nat))) : Prop :=
     match lms with
     | cons (pair pc (pair A n)) lms' =>
        (forall s0 s, A s0 s -> wp_instr n pc s0 s) /\
        valid_assert lms'
     | nil =>
       forall l h, m.(am_pre) (l,h) -> 
          wp_label m.(am_needed) 
             (BYTECODEMETHOD.firstAddress m) (l,h) (h,OperandStack.empty,l)
     end.
   
   Lemma valid_assert_spec : 
     forall lms, valid_assert lms ->
     forall pc A n, In (pc,,(A,,n)) lms -> 
     forall s0 s, A s0 s -> wp_instr n pc s0 s.
   Proof.
    induction lms;simpl;intros;elimFalse.
    destruct H0.
    subst;destruct H;auto.
    destruct a as (pc',(A',n'));destruct H;eauto.
   Qed.

   Lemma valid_assert_pre :
     forall lms, valid_assert lms ->
     forall l h, m.(am_pre) (l,h) -> 
       wp_label m.(am_needed) 
            (BYTECODEMETHOD.firstAddress m) (l,h) (h,OperandStack.empty,l).
   Proof.
    induction lms;simpl;intros;auto.
    destruct a as (pc,(A,n));destruct H;auto.
   Qed.

  End ____WP_METH____.

  Definition annotMeth m n pre post lms :=
    match METHOD.body m with
    | Some bm =>
      Some 
        (AnnoMeth n
            m bm (BYTECODEMETHOD.exceptionHandlers bm)
            pre post lms)
    | None => None 
    end.

  Definition certifiedMethod m (anno:nat ** GlobalMethSpec ** LocMethSpec) :=
    match anno with
    | (pair (pair n (pair pre post)) lms) => 
      match annotMeth m n pre post lms with
      | Some am => valid_assert am (PCM.elements am.(am_local))
      | None => False
      end
    end.

  Lemma certifiedMethod_spec : 
    forall m n pre post lms, certifiedMethod m ((n,,(pre,,post)),,lms) ->
      exists am,
        valid_am am /\ m = am /\ am.(am_pre) = pre /\
        am.(am_post) = post /\ am.(am_local) = lms /\
        valid_assert am (PCM.elements am.(am_local)).
  Proof.
   intros m n pre post lms;unfold certifiedMethod;simpl.
   unfold annotMeth;case_eq (METHOD.body m);simpl;intros;elimFalse.
   exists (AnnoMeth n m b (BYTECODEMETHOD.exceptionHandlers b) pre post lms);
    repeat split;auto.
  Qed.

  Fixpoint allCertMeth 
    (mst: list (MethodSignature**(nat ** GlobalMethSpec ** LocMethSpec))) : Prop :=
    match mst with
    | cons (msig,, anno) mst' => 
         match findMethod p msig with
         | None => False
         | Some m => certifiedMethod m anno /\ allCertMeth mst'
         end
    | nil => True
    end.

  Lemma allCertMeth_spec : 
    forall mst, allCertMeth mst ->  
    forall msig anno, In (msig,,anno) mst -> 
    exists m, findMethod p msig = Some m /\ certifiedMethod m anno.
  Proof.
   induction mst;simpl;intros;elimFalse.
   destruct H0.
   subst. generalize H;clear H;case_eq (findMethod p msig);intros;elimFalse.
   destruct H0;exists m;auto.
   destruct a. generalize H;clear H;case_eq (findMethod p m);intros;elimFalse.
   destruct H1;auto.
  Qed. 
     
  Definition certifiedProg := allCertMeth (MM.elements p.(ap_mst)).
  
  Lemma certifiedProg_spec : 
    certifiedProg ->
    forall msig m pre post, 
      findMethod p msig = Some m ->
      globalLookup p msig = Some (pre,,post) ->
      exists n, exists lms, certifiedMethod m ((n,,(pre,,post)),,lms).
  Proof.
   intros Hcertif msig m pre post Hfind; unfold globalLookup.
   case_eq (MM.get p.(ap_mst) msig).
   intros ((n,(pre',post')),lms) H1 H2;inversion_mine H2.
   exists n;exists lms. 
   destruct (allCertMeth_spec _ Hcertif _ _ (MM.get_some_in_elements _ _ _ _ H1)) as (m',(U1,U2)).
   rewrite Hfind in U1. injection U1;intros;clear U1;subst;trivial.
   intros;discriminate.
  Qed.

  Section ____WP_PROOF____.
   Hypothesis IsCertProp : certifiedProg.

   Lemma wp_label_wp_instr :        
     forall am, valid_am am ->  
       valid_assert am (PCM.elements am.(am_local)) ->
     forall n s0 pc s, wp_label am n pc s0 s ->
     exists n', wp_instr am n' pc s0 s.
   Proof.
    intros am Hv Hva n s0 pc s;destruct n;simpl; unfold assertLookup.
    case_eq (PCM.get am.(am_local) pc);intros;elimFalse.
    destruct p0 as (A,n).
    assert (HH := PCM.get_some_in_elements _ _ _ _ H).
    exists n;eapply valid_assert_spec;eauto.
    case_eq (PCM.get am.(am_local) pc);intros.
    destruct p0 as (A,n').
    assert (HH := PCM.get_some_in_elements _ _ _ _ H).
    exists n';eapply valid_assert_spec;eauto.
    exists n;exact H0.
   Qed.

   Hypothesis BST : 
         forall cn msig pre post, globalLookup p (cn,msig) = Some (pre,, post) ->
         forall cl m', lookup p cn msig (cl, m') -> 
         exists pre', exists post', 
             globalLookup p (cl, METHOD.signature m') = Some (pre',, post') /\
             forall s0 t, (pre' s0 -> post' s0 t) -> (pre s0 -> post s0 t).

  Lemma lookup_findMethod : 
     forall cn mid cl m , lookup p cn mid (cl,m) ->  
             findMethod p (cl,METHOD.signature m) = Some m.
  Proof.
      assert (forall cn mid M,
     lookup p cn mid M -> forall cl m, M =(cl, m) -> findMethod p (cl,METHOD.signature m) = Some m).
     induction 1;simpl;intros.
     injection H0;intros;subst.
     inversion_mine H. simpl in H1.
     destruct (PROG.class p cl);try discriminate.
     rewrite <-  (CLASS.method_signature_prop _ _ _ H1);trivial.
     exact (IHlookup _ _ H2).
     intros; eapply H;eauto.
  Qed.

   Lemma wp_label_correct : 
     forall m s sr, IntraStepStar p m s sr ->
     forall am, valid_am am -> m = am ->  
       valid_assert am (PCM.elements am.(am_local)) ->
     forall n s0, 
       wp_label am n (fst s) s0 (snd s) ->
       match sr with
       | inl (pc',s') => exists n', wp_label am n' pc' s0 s'
       | inr r => am.(am_post) s0 r
       end.
   Proof.
    induction 1 using IntraStepStar_ind;intros;subst.
    destruct s;simpl in H2;exists n;trivial.
    destruct (wp_label_wp_instr _ H0 H2 _ _ _ _ H3) as (n',H4).
    unfold wp_instr in H4;eapply wp_return_post; eauto.
    destruct (wp_label_wp_instr _ H1 H3 _ _ _ _ H4) as (n',H5).
    destruct s as (pc,s);destruct s' as (pc',s').
    inversion_mine H.
    assert (UU := wp_intra _ H1 _ _ _ _ _ _ H2 H5).
    eapply IHIntraStepStar;eauto.
    inversion_mine H10.
    assert (VV := vam_Bytecode _ H1);simpl_eq.
    assert (UU:= wp_exc _ H1 _ _ _ _ _ _ _ _ _ _ H8 H13 H12 H5).
    eapply IHIntraStepStar;eauto.

    (* Call *) 
    destruct (wp_label_wp_instr _ H1 H3 _ _ _ _ H4) as (n',H5).
    inversion_mine H.
    inversion_mine H6.
    unfold wp_instr, wp_i in H5;simpl in H5;
     rewrite (instrAt_B _ H1 _ _ H14) in H5;simplIf.
    generalize H5;clear H5;case_eq (globalLookup p mid).
    intros (pre',post') Heq.
     destruct(certifiedProg_spec IsCertProp mid m' pre' post' H16 Heq)as(N,(lms', UU)).
    case_eq  (init_args (length (METHODSIGNATURE.parameters (snd mid))) (args ++ os));
     intros l' os' Heq2.
    destruct (certifiedMethod_spec _ _ _ _ _ UU) as (am',HH);intuition;subst.
    assert (WW := H5  h2 (Exception loc));clear H5;simpl in WW.
    assert (VV:= IHIntraStepStar _ H (refl_equal _) H11);clear IHIntraStepStar.
    destruct (init_args_correct _ _ _ _ _ Heq2 H18); subst. 
    assert  (U := valid_assert_pre _ _ H11).
    assert (Y:= vam_Bytecode _ H); simpl_eq;simpl in VV.
    rewrite H18 in VV.
    generalize H12;intros Hun;inversion_mine Hun.
    eapply wp_Exception_uncaught;eauto.

    intros UU UUU;elim UUU.

    unfold wp_instr, wp_i in H5;simpl in H5;
     rewrite (instrAt_B _ H1 _ _ H14) in H5;simplIf;elimFalse.

(* Invokevirtual *)
      generalize H5;clear H5.
     case_eq (findMethod p (cn, mid));[intros m'' Heq | intros Heq].
     case_eq (METHOD.isNative m'');intros Heq1.     intro HF;elim HF.
     case_eq (globalLookup p (cn,mid)); [intros (pre,post) Heq2 | intros Heq2 HF;elim HF].
     rewrite pop_append;trivial. 
     inversion H12;subst.
     intros UU;assert (V:= UU h2 (Exception loc) _ H9);clear UU;simpl in V.
     apply (wp_Exception_uncaught _ H1 (wp_label am n') _ _ _ _ l1 s0 H9 H12).
     apply V;clear V.
     destruct (BST _ _ _ _ Heq2 _ _ H16) as (pre',(post',(U1,U2))).
     intros Hpre;apply U2;[clear Hpre U2 Heq2 pre post| trivial].
     destruct(certifiedProg_spec IsCertProp (cl,METHOD.signature m') m' pre' post' (lookup_findMethod _ _ _ _ H16) U1)as(N,(lms', UU)).
     destruct (certifiedMethod_spec _ _ _ _ _ UU) as (am',HH);intuition;subst.
     assert (VV:= IHIntraStepStar _ H2 (refl_equal _) H20);clear IHIntraStepStar.
     simpl in VV. eapply VV with (n:= am_needed am');clear VV.
    assert (Y:= vam_Bytecode _ H2); simpl_eq;trivial.
    rewrite H18;rewrite plus_comm;simpl. rewrite plus_comm in H;simpl in H.
    apply valid_assert_pre with (PCM.elements (am_local am'));auto.
    intros Hf;elim Hf.  

(* *)   
    destruct (wp_label_wp_instr _ H2 H4 _ _ _ _ H5) as (n',H5').
    inversion_mine H.
    inversion_mine H6.
    unfold wp_instr, wp_i in H5';simpl in H5';
     rewrite (instrAt_B _ H2 _ _ H14) in H5';simplIf.
    generalize H5';clear H5';case_eq (globalLookup p mid).
    intros (pre',post') Heq;
     destruct(certifiedProg_spec IsCertProp mid m' pre' post' H16 Heq)as(N,(lms', UU)).
    case_eq  (init_args (length (METHODSIGNATURE.parameters (snd mid))) (args ++ os));
     intros l' os' Heq2.
    destruct (certifiedMethod_spec _ _ _ _ _ UU) as (am',HH);intuition;subst.
    assert (VV:= IHIntraStepStar1 _ H (refl_equal _) H11);clear IHIntraStepStar1.
    destruct (init_args_correct _ _ _ _ _ Heq2 H18);subst.
    assert  (U := valid_assert_pre _ _ H11).
    assert (Y:= vam_Bytecode _ H); simpl_eq;simpl in VV.
    rewrite H18 in VV.
    generalize H12;intros Hun;inversion_mine Hun.
    assert (WW := H5'  h2 (Normal ov));clear H5';simpl in WW.
    eapply IHIntraStepStar2 with (n:=n');eauto.
    rewrite <- H18 in WW;rewrite pop_append in WW;trivial.
    rewrite <- H18 in VV.
    simpl. eapply wp_next_wp_label with (m:= am);eauto.

    intros UU UUU;elim UUU.

(* invokevirtual *)
    unfold wp_instr, wp_i in H5';simpl in H5';
     rewrite (instrAt_B _ H2 _ _ H14) in H5';simplIf.
    generalize H5';clear H5 H5'.
     case_eq (findMethod p (cn, mid));[intros m'' Heq | intros Heq].
     case_eq (METHOD.isNative m'');intros Heq1.     intro HF;elim HF.
     case_eq (globalLookup p (cn,mid)); [intros (pre,post) Heq2 | intros Heq2 HF;elim HF].
     rewrite pop_append;trivial. 
     intros UU;assert (V:= UU h2 (Normal ov) );clear UU;simpl in V.
    assert (pre
       (stack2localvar (args ++ Ref loc :: os)
          (length (METHODSIGNATURE.parameters mid) + 1), h1) ->
     post
       (stack2localvar (args ++ Ref loc :: os)
          (length (METHODSIGNATURE.parameters mid) + 1), h1) (h2, Normal ov)).
     destruct (BST _ _ _ _ Heq2 _ _ H16) as (pre',(post',(U1,U2))).
     intros Hpre;apply U2;[clear Hpre U2 | trivial].
     destruct(certifiedProg_spec IsCertProp (cl,METHOD.signature m') m' pre' post' (lookup_findMethod _ _ _ _ H16) U1)as(N,(lms', UU)).
     destruct (certifiedMethod_spec _ _ _ _ _ UU) as (am',HH);intuition;subst.
     assert (VV:= IHIntraStepStar1 _ H3 (refl_equal _) H11);clear IHIntraStepStar1.
     simpl in VV. eapply VV with (n:= am_needed am');clear VV.
    assert (Y:= vam_Bytecode _ H3); simpl_eq;trivial.
    rewrite H18;rewrite plus_comm;simpl. rewrite plus_comm in H;simpl in H.
    apply valid_assert_pre with (PCM.elements (am_local am'));auto.
    assert (W := V H);clear V.
    eapply IHIntraStepStar2 with (n:=n');eauto.
    eapply wp_next_wp_label with (m:= am);eauto.
    assert (os = (pop value (length (METHODSIGNATURE.parameters mid) + 1)
           (args ++ Ref loc :: os))).
   change (Ref loc :: os) with ((Ref loc::nil)++os).   rewrite ass_app.
   rewrite pop_append;trivial.
   rewrite <- H18. generalize args.
   induction args0;simpl;trivial. rewrite IHargs0;trivial.
   rewrite H3;trivial.
   intros Hf;elim Hf.  
(* *)
    inversion_mine H6.
    unfold wp_instr, wp_i in H5';simpl in H5';
     rewrite (instrAt_B _ H2 _ _ H14) in H5';simplIf.
    generalize H5';clear H5';case_eq (globalLookup p mid).
    intros (pre',post') Heq;
     destruct(certifiedProg_spec IsCertProp mid m' pre' post' H16 Heq)as(N,(lms', UU)).
    case_eq  (init_args (length (METHODSIGNATURE.parameters (snd mid))) (args ++ os));
     intros l' os' Heq2.
    destruct (certifiedMethod_spec _ _ _ _ _ UU) as (am',HH);intuition;subst.
    assert (VV:= IHIntraStepStar1 _ H (refl_equal _) H11);clear IHIntraStepStar1.
    destruct (init_args_correct _ _ _ _ _ Heq2 H18);subst.
    assert  (U := valid_assert_pre _ _ H11).
    assert (Y:= vam_Bytecode _ H); simpl_eq;simpl in VV.
    rewrite H18 in VV.
    generalize H12;intros Hun;inversion_mine Hun.
    assert (WW := H5'  h2 (Exception loc));clear H5';simpl in WW.
    eapply IHIntraStepStar2 with (n:=n');eauto.
    rewrite <- H18 in WW; rewrite <- H18 in VV.
    simpl. eapply wp_Exception_caught;eauto.
    rewrite (vam_Handler _ H2).
    assert (Y:= vam_Bytecode _ H2); simpl_eq;trivial.

    intros UU UUU;elim UUU.

    unfold wp_instr, wp_i in H5';simpl in H5';
    rewrite (instrAt_B _ H2 _ _ H14) in H5';simplIf;elimFalse.
 (* invokevirtual *)
    generalize H5';clear H5 H5'.
     case_eq (findMethod p (cn, mid));[intros m'' Heq | intros Heq].
     case_eq (METHOD.isNative m'');intros Heq1.     intro HF;elim HF.
     case_eq (globalLookup p (cn,mid)); [intros (pre,post) Heq2 | intros Heq2 HF;elim HF].
     rewrite pop_append;trivial. 
    generalize H12;intros Hun;inversion_mine Hun.
    intros UU;assert (V:= UU h2 (Exception loc) _ H10);clear UU;simpl in V.
    assert (pre
       (stack2localvar (args ++ Ref loc0 :: os)
          (length (METHODSIGNATURE.parameters mid) + 1), h1) ->
     post
       (stack2localvar (args ++ Ref loc0 :: os)
          (length (METHODSIGNATURE.parameters mid) + 1), h1)
       (h2, Exception loc)).
     destruct (BST _ _ _ _ Heq2 _ _ H16) as (pre',(post',(U1,U2))).
     intros Hpre;apply U2;[clear Hpre U2 | trivial].
     destruct(certifiedProg_spec IsCertProp (cl,METHOD.signature m') m' pre' post' (lookup_findMethod _ _ _ _ H16) U1)as(N,(lms', UU)).
     destruct (certifiedMethod_spec _ _ _ _ _ UU) as (am',HH);intuition;subst.
     assert (VV:= IHIntraStepStar1 _ H3 (refl_equal _) H20);clear IHIntraStepStar1.
     simpl in VV. eapply VV with (n:= am_needed am');clear VV.
    assert (Y:= vam_Bytecode _ H3); simpl_eq;trivial.
    rewrite H18;rewrite plus_comm;simpl. rewrite plus_comm in H;simpl in H.
    apply valid_assert_pre with (PCM.elements (am_local am'));auto.
    assert (W := V H);clear V.
    eapply IHIntraStepStar2 with (n:=n');eauto.
    simpl;eapply wp_Exception_caught;eauto.
    rewrite (vam_Handler _ H2).
    assert (Y:= vam_Bytecode _ H2); simpl_eq;trivial.
    intros Hf;elim Hf.
 Qed.
 
  End ____WP_PROOF____.

 End ____WP____.

End Wp.



